# ConflictResolutionUtils module provides API to handle settings that might conflict and cannot be handled by PolicyEnforcer and need to be handled in the feature-module context.
# Currently used by AV for actionsForDetectedThreats and for ASR settings. For example you can see ASR 'AttackSurfaceReductionRules' or AV 'ThreatSeverityDefaultAction'
# These settings are mergeable and can induce conflicts when different values are sent from different policies.
# The settings are in format of: "subSetting1=value1|subSetting2=value2|..." (In ThreatSeverityDefaultAction for example - "4=1|1=8|5=3")
# For each setting the calling (feature) module will hold 2 structs:
#   1. requestedValues - A Hashtable mapping between different subSettings and their data containing: @(value, @(groups), @(HResult, ErrorStr)). 
#       value: Represents the requested value to be set
#       groups: Represents the policies indexes where this subSettings were requested
#       firstErrorDetails (HResult, ErrorStr): An array containing the first erorr encountered for this subSetting
#   This struct needs to be passed as a parameter to functions in this module in 2 different ways:
#   - when the parameter's name is 'requestedValues': pass the entire struct.
#   - when the parameter's name is 'requestedValuesForSetting': pass only the specifeic setting's entry. For example: '$requestedValues[$settingsNamesMapping.$nameSuffix]'
# 2. conflictingGroups - An array containing groups (policies) for which a conflict was detected.
#   Both structs, the hashtable and the array will be updated using the 'StoreSettingToRequested' function in this module that should be called during 'Set' in the feature-module.

$script:versionId = "1.0001"

LogInfo (-join("ConflictResolutionUtils setup (", $script:versionId, ")"))

if (-not ([System.Management.Automation.PSTypeName]'utilsInfoIndex').Type)
{
    Add-Type -TypeDefinition @"
        public enum utilsInfoIndex {
            subValue,
            groups,
            settingFirstErrorInfo
        }
"@
}

# CalcConflictingGroupsDifference - calulate the groups we need to add from $newConflictingGroups to $totalConflictingGroups.
# arguments:
#$newConflictingGroups - array of ints, representing group indices that are to be added to the totalConflictingGroups array. Duplicate indices possible. 
#totalConflictingGroups - array of ints, representing the currently known group indices of conflicting policies.
# returned value: array of ints that represent the group indices that differ between the 2 arguments and are missing in $totalConflictingGroups.
function CalcConflictingGroupsDifference([int[]] $newConflictingGroups, [int[]] $totalConflictingGroups)
{
    $addedConflictingGroups = @()
    foreach($conflictingGroupId in $newConflictingGroups)
    {
        if(($conflictingGroupId -notin $totalConflictingGroups) -and ($conflictingGroupId -notin $addedConflictingGroups))
        {
            $addedConflictingGroups += $conflictingGroupId
        }
    }
    return $addedConflictingGroups
}

# ParseSettingInfo - Returns a map containing pairs of {subSetting, value}
# Used for settings we receive in format of "subSetting1=value1|subSetting2=value2|..."
function ParseSettingInfo([string] $value)
{
    $resInfo = @{} #subSetting = subValue
    $settingArray = $value.Split("|") | where {$_}
    
    foreach($subValue in $settingArray) 
    {
        $singleValueInfo = $subValue.Split("=")
        try {
            if($singleValueInfo.Count -ne 2)
            {
                LogWarning (-join("ParseSettingInfo: Error parsing value. subValue is not in format of subSetting=subValue. subValue:",$subValue))
            }
            elseif($resInfo.ContainsKey($singleValueInfo[0]))
            {
                LogWarning (-join("ParseSettingInfo: Error parsing value. Value contains duplicate subSetting:",$singleValueInfo[0]))   
            }
            else {
                $resInfo.Add($singleValueInfo[0],$singleValueInfo[1])
            }
        } 
        catch {
            LogWarning (-join("ParseSettingInfo: Error adding subValue:", $subValue, ". HResult:",$_.Exception.HResult,", error message:",$_.Exception.Message))
        }
    }

    return $resInfo
}


 # GetValuesToRemove - Returns an array of values (only subSetting) to remove from the value string.
 # The value string can be in subSetting1=value1|subSetting2=value2|... format or in subSetting1|subSetting2|subSetting3... format.
function GetValuesToRemove([string] $value)
{
    $removeValuesArr = $value.Split("|")
    if($value -match "=") #setting is in format of subSetting1=value1|subSetting2=value2|...
    {
        $cleanRemovedValuesArr = @()
        foreach($removedValue in $removeValuesArr)
        {
            $cleanRemovedValuesArr += ($removedValue.Split("="))[0]
        }
        $removeValuesArr = $cleanRemovedValuesArr
    }
    
    return $removeValuesArr
}

# GetGroupIdFromValue - returns the groups (policies) indices that is related to the groups of the setting received in value.
# function is called AFTER setting was Set and the value was inserted to the $requestedValues array (with the relevant groupsIDs)
function GetGroupIdFromValue([string] $value, [int] $groupCounter, [hashtable] $requestedValuesForSetting)
{
    $valuesArr = $value.Split("|") | where {$_}
    $possibleGroups = @(1..($groupCounter))
    $valueGroups = @()
    write-host (-join("GetGroupIdFromValue called with value:", $value))
    foreach($singleValue in @($valuesArr))
    {
        if($singleValue -match "=")
        {
            $singleValue = ($singleValue.Split("="))[0]
        }
        $singleValueGroups = ($requestedValuesForSetting[$singleValue][([utilsInfoIndex]::groups)])
        $newGroups = $possibleGroups | where {$singleValueGroups -contains $_}
        $valueGroups += CalcConflictingGroupsDifference $newGroups $valueGroups
    }

    return $valueGroups
}



# GetFirstErrorInfoByGroup - returns the 1st error that is related to this group of values.
# Error info format: @(HReslt,errStr)
# If no error stored in requestedValues for this group, return an empty array.
function GetFirstErrorInfoByGroup([int] $groupId, [hashtable] $requestedValuesForSetting)
{
    foreach($singleValueInfo in $requestedValuesForSetting.Values)
   {
       if((@($singleValueInfo[[utilsInfoIndex]::groups]) -contains $groupId) -and (($singleValueInfo[([utilsInfoIndex]::settingFirstErrorInfo)]).Count -gt 0))
       {
           return @($singleValueInfo[[utilsInfoIndex]::settingFirstErrorInfo])
       }
   }

   return @()
}

#StoreSettingToRequested - Stores requested settings in the requestedValues struct in case there is no conflict, or in conflictingGroups in case of conflict. 
#Params are:
# requestedSettingsInfoToAdd - The requested values to add in format of pairs (subSetting, value).
# currentGroup - The current group (policy) index the values are comming from.
# requestedValues - The entire requestedValues struct for this setting. Used in order to check for conflicts, and in case of no conflict adding the new setting to this struct.
# conflictingSettingGroups - The conflicting groups struct for this setting. In case we discover a conflict here, adding the new group to the struct.
function StoreSettingToRequested([hashtable] $requestedSettingsInfoToAdd, [int] $currentGroup, [hashtable] $requestedValues, [ref] $conflictingSettingGroups)
{
    foreach($subSetting in $requestedSettingsInfoToAdd.keys)
    {
        #sub setting is already requested
        if($requestedValues.ContainsKey($subSetting))
        {
            #in any case - add new groupID to this sub setting
            ($requestedValues[$subSetting][([utilsInfoIndex]::groups)]) += $currentGroup
            
            #if applied sub setting's value doesn't match on several policies, report conflict
            if($requestedSettingsInfoToAdd[$subSetting] -ne ($requestedValues[$subSetting][([utilsInfoIndex]::subValue)]))
            {
                $errorMsg = (-join("Conflict recognized when storing sub setting (",$subSetting,"). Value: ",($requestedValues[$subSetting][([utilsInfoIndex]::subValue)]),",",$requestedSettingsInfoToAdd[$subSetting]))
                LogInfo $errorMsg

                #mark all sub settings previously stored for application since we found a conflict and will not set them.
                $newConflictingGroups = $requestedValues[$subSetting][([utilsInfoIndex]::groups)] + $currentGroup 
                $conflictingSettingGroups.value += CalcConflictingGroupsDifference $newConflictingGroups $conflictingSettingGroups
                LogDebug (-join("Currently conflicting groups:",($newConflictingGroups -join("|")),". Total conflicting groups:",($conflictingSettingGroups.value -join("|"))))
            } 

        }
        else
        {
            #first time we see this sub setting- add new sub setting and value (no error, yet)
            LogDebug (-join("Adding new sub setting to be set: [",$subSetting,",",$requestedSettingsInfoToAdd[$subSetting],",",$currentGroup,"]"))
            $requestedValues.Add($subSetting,@($requestedSettingsInfoToAdd[$subSetting],@($currentGroup),@()))
        }
    }
}


# GetGroupFromSettings - function uses requestedValues data struct to 
# get the common group for all the subSettings recieved.
# if no such group is found, $null is returned
function GetGroupFromSettings([string[]] $settingsArr, [hashtable] $requestedValuesForSetting)
{
   $settingsInfo = $requestedValuesForSetting
    
   $groups = @()
   foreach($subSetting in $settingsArr)
   {
       $singleInfo = $settingsInfo[$subSetting]
       $subSettingGroups = $singleInfo[([utilsInfoIndex]::groups)]
       if($groups.Count -eq 0)
       {
           $groups = $subSettingGroups
       } 
       else 
       {
           $groups = ($groups | ?{$subSettingGroups -contains $_}) #leave only the group ids that exists in both
       }
   }

    # if the subSettings recieved have more than 1 group in common 
    # remove the groups that contain other subSettings as well
    if($groups.Count -gt 1) 
    {
        LogInfo (-join("GetGroupFromSettings common groups: ", ($groups -join "|")))
        foreach($singleSetting in $settingsInfo.Keys)
        {
            $singleSettingInfo = $settingsInfo[$singleSetting]
            $singleSettingGroups = $singleSettingInfo[([utilsInfoIndex]::groups)]
            $commonGroups = ($groups | ?{$singleSettingGroups -contains $_})

            if( ($null -ne $commonGroups) -and !($settingsArr -contains $singleSetting))
            {
                $groups = ($groups | ?{!($commonGroups -contains $_)}) 
                if($groups.Count -eq 1) 
                {
                    #if groups is filtered enough (1 group) - stop
                    break
                }
            }
        }
    }
    
    LogInfo (-join("GetGroupFromSettings identifying groups: ", ($groups -join "|")))
    
    #at this point we either have only 1 group or several groups with the same set of subSettings 
    # (could be if order is not the same so PolicyEnforcer does not recognize that they are the same), 
    # and then they must have the same result (conflict/no conflict).
    return $groups
}


# IsSubSettingOnlyFromGroupInConflict - return true if: (A) sub setting is in a group that has a conflict
# (not necessarily because of this sub setting, can also be due to other sub setting in same policy)
# AND (B) this sub setting is NOT a part of a policy which doesn't have conflicts
# In other words: this is a sub setting that should *not* be applied
function IsSubSettingOnlyFromGroupInConflict([string] $subSetting, [hashtable] $requestedValuesForSetting, [int[]] $conflictingGroups)
{
   $subSettingInfo = $requestedValuesForSetting[$subSetting]
   $subSettingGroups = $subSettingInfo[([utilsInfoIndex]::groups)]
   $groupsInConflict = ($subSettingGroups | ?{$conflictingGroups -contains $_})
   $groupsNotInConflict = ($subSettingGroups | ?{!($conflictingGroups -contains $_)})
   LogInfo (-join("Sub setting ",$subSetting," in following conflicting groups:",[string]$groupsInConflict,", and in non-conflicting groups:",[string]$groupsNotInConflict))
   return (($null -ne $groupsInConflict) -and ($null -eq $groupsNotInConflict))
}

# SIG # Begin signature block
# MIIntwYJKoZIhvcNAQcCoIInqDCCJ6QCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCDwqmofZIyIFiXP
# o24v4ndiYajLLY3LuoQe9hxU+cQdRaCCDZcwggYVMIID/aADAgECAhMzAAADcaWg
# nFyRHuruAAAAAANxMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMwNTExMTg0MTEzWhcNMjQwNTA4MTg0MTEzWjCBlDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjE+MDwGA1UEAxM1TWlj
# cm9zb2Z0IFdpbmRvd3MgRGVmZW5kZXIgQWR2YW5jZWQgVGhyZWF0IFByb3RlY3Rp
# b24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDQl+fOHFGkNfDHrcv3
# g8shIsUp58WmSbtNm4uu7Sw75Wgsgo5p2yYGGqiRWbu0eIKhAJ1bRp9Qa7vAjMEw
# d0wf9wlLjDNAITzDsokwoCl0rOgWxzYuGSusZq4e/liClIpY5fuxrpsEnsssR1o6
# jITkdJQ6ZkWsTFETxLLOEGxe9BkTkXZaBOHlhTt5QIq6UB6j8zNN/Vo4rFmq751X
# lQLA1jOnNi5NOS3/11TW16ZCo9XKG22zdNsJCUckT/QH2eeQNgpuyfIvP06o+RpP
# Uo7P4xdc4c74S3HKfRiNuqbs3+xvb22GtXIz3cKFDhILpNKN/SC8qLjizq0PxwL1
# lr1ZAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBggrBgEFBQcDAwYKKwYBBAGCN0wv
# ATAdBgNVHQ4EFgQUcR5wDY+Mh5A/I7scm54A8ODeSVcwRQYDVR0RBD4wPKQ6MDgx
# HjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEWMBQGA1UEBRMNNDUxODk0
# KzUwMTAwNTAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8E
# TTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9N
# aWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBR
# BggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0
# cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAw
# DQYJKoZIhvcNAQELBQADggIBAKYhzPfo4pWE2MYg2W1VnHJ7NZ/nwEpGYEz7Qhqc
# bPj2sJFFhnUzYHOAyXz4vP0+ERhFw9lrqLvoyj8s6xQU11QPQVItopahH65USGWh
# UcyrXFBVSx7MG3Dwi+NT6rxq9MjY3KLB1hZPF7hgtOjWgtjXgeJOsy3IjPiRkDwX
# 5b5Pg6o8Xpyt84mpvAPsDKO7yJGJq8V528deIYh+0DcELvUZTtl4Kz8ElT4BxF6i
# XxXkuqpCWO/rwEa6HrLNV2udND/AImVio09bbZXOWBcb9uTTNezYX5fZvinIX7hl
# UHHIDaVUqqwTFtac8q7BLE02T0P7NAeBtayujIQFzGNuwLN302UmUWAyc4LynqZa
# h2OmZWcMODUCuHIMP3pk54XjWt0YYEWPFAbA7T7j8wjby/vPXdG93LybID/yjGDb
# U1t5xMoEpN5rvSUM1U42iJ1tF268/nE2/8nKqUpdJXH/MaXxYU3FnbshmIe1xfxF
# NjeMYN4jg+5gjz5SBzHAhT1X9QWXNOhAXxib3wcchlsxFYtSNihHLEJ2GuckRuPi
# 47VWi/3hTTnl88S8r3jsbWLatGbQxcbrsF4sglAvFD9SQSXLGUsE5wz/1p/Gb1Fh
# qfNBc6E4AWVj9+A1KezSnMBAw8gJk3fiRkdxfSfpLOkHvgA7qsOmQ7qwjs05kWZf
# 055CMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWlj
# cm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4
# MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3Y
# bqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUB
# FDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnbo
# MlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT
# +OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuy
# e4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEh
# NSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2
# z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3
# s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78Ic
# V9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E
# 11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5P
# M4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcV
# AQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAf
# BgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBL
# hklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggr
# BgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsG
# AQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwA
# ZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0G
# CSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDB
# ZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc
# 8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYq
# wooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu
# 5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWI
# UUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXh
# j38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yH
# PgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtI
# EJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA9Z74v2u3S5fi63V4Guzq
# N5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgR
# MiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQ
# zTGCGXYwghlyAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
# dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9y
# YXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEC
# EzMAAANxpaCcXJEe6u4AAAAAA3EwDQYJYIZIAWUDBAIBBQCgga4wGQYJKoZIhvcN
# AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUw
# LwYJKoZIhvcNAQkEMSIEIMEre0F9JtKN+tyT/CP02pGDT0aCClWgsFmhtGu7v93a
# MEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8AcwBvAGYAdKEagBhodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEAgrfCDrjsWnTs
# 6u0RJ8dKJ4TN+RAJNpgvJDvooC7JnFFvpUeAL+i5K5MuqFNvzKtNSb4j9Ol2rV00
# p28OZQJvVvs1fuUfiCEil8H+bjj9HnPd6RPUdgsFaEnJDJfWae/8FW3jW/WFRNGn
# bYwxOAKImrWbKmDYlJiGw3W7WDQ/HPoFWPjeQvjhnl7USe7EnT7S2p4o8DU8Tanc
# zD+pLSWMkOuO3V2ZZfgQg9+GSWvLwoD5XgVVXSeljCGARJIDTaj9Ctc1L+/iYQ1X
# kBRNOREMGnLY7mENwGAgej6+L6nB0zqbYyCaOPaGa3/HRsU8Uozk6Q2CxaMtFyIW
# JotL+DegqKGCFwAwghb8BgorBgEEAYI3AwMBMYIW7DCCFugGCSqGSIb3DQEHAqCC
# FtkwghbVAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFRBgsqhkiG9w0BCRABBKCCAUAE
# ggE8MIIBOAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCBRxWxf2fZN
# RHYTtJPTLdKJd4embn9qZJIJCr27xOWm8QIGZIsUPZkDGBMyMDIzMDcxMTE3NDcw
# MS4yMTlaMASAAgH0oIHQpIHNMIHKMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBPcGVyYXRpb25z
# MSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjpENkJELUUzRTctMTY4NTElMCMGA1UE
# AxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaCCEVcwggcMMIIE9KADAgEC
# AhMzAAABx/sAoEpb8ifcAAEAAAHHMA0GCSqGSIb3DQEBCwUAMHwxCzAJBgNVBAYT
# AlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYD
# VQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBU
# aW1lLVN0YW1wIFBDQSAyMDEwMB4XDTIyMTEwNDE5MDEzNVoXDTI0MDIwMjE5MDEz
# NVowgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
# EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAjBgNV
# BAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJjAkBgNVBAsTHVRoYWxl
# cyBUU1MgRVNOOkQ2QkQtRTNFNy0xNjg1MSUwIwYDVQQDExxNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBTZXJ2aWNlMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA
# r0LcVtnatNFMBrQTtG9P8ISAPyyGmxNfhEzaOVlt088pBUFAIasmN/eOijE6Ucaf
# 3c2bVnN/02ih0smSqYkm5P3ZwU7ZW202b6cPDJjXcrjJj0qfnuccBtE3WU0vZ8Ci
# QD7qrKxeF8YBNcS+PVtvsqhd5YW6AwhWqhjw1mYuLetF5b6aPif/3RzlyqG3SV7Q
# PiSJends7gG435Rsy1HJ4XnqztOJR41I0j3EQ05JMF5QNRi7kT6vXTT+MHVj27FV
# Q7bef/U+2EAbFj2X2AOWbvglYaYnM3m/I/OWDHUgGw8KIdsDh3W1eusnF2D7oenG
# gtahs+S1G5Uolf5ESg/9Z+38rhQwLgokY5k6p8k5arYWtszdJK6JiIRl843H74k7
# +QqlT2LbAQPq8ivQv0gdclW2aJun1KrW+v52R3vAHCOtbUmxvD1eNGHqGqLagtlq
# 9UFXKXuXnqXJqruCYmfwdFMD0UP6ii1lFdeKL87PdjdAwyCiVcCEoLnvDzyvjNjx
# tkTdz6R4yF1N/X4PSQH4FlgslyBIXggaSlPtvPuxAtuac/ITj4k0IRShGiYLBM2D
# w6oesLOoxe07OUPO+qXXOcJMVHhE0MlhhnxfN2B1JWFPWwQ6ooWiqAOQDqzcDx+7
# 9shxA1Cx0K70eOBplMog27gYoLpBv7nRz4tHqoTyvA0CAwEAAaOCATYwggEyMB0G
# A1UdDgQWBBQFUNLdHD7BAF/VU/X/eEHLiUSSIDAfBgNVHSMEGDAWgBSfpxVdAF5i
# XYP05dJlpxtTNRnpcjBfBgNVHR8EWDBWMFSgUqBQhk5odHRwOi8vd3d3Lm1pY3Jv
# c29mdC5jb20vcGtpb3BzL2NybC9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENB
# JTIwMjAxMCgxKS5jcmwwbAYIKwYBBQUHAQEEYDBeMFwGCCsGAQUFBzAChlBodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY3Jvc29mdCUyMFRp
# bWUtU3RhbXAlMjBQQ0ElMjAyMDEwKDEpLmNydDAMBgNVHRMBAf8EAjAAMBMGA1Ud
# JQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBCwUAA4ICAQDQy5c8ogP0y8xAsLVc
# a07wWy1mT+nqYgAFnz2972kNO+KJ7AE4f+SVbvOnkeeuOPq3xc+6TS8g3FuKKYEw
# YqvnRHxX58tjlscZsZeKnu7fGNUlpNT9bOQFHWALURuoXp8TLHhxj3PEq9jzFYBP
# 2YNMLol70ojY1qpze3nMMJfpdurdBBpaOLlJmRNTLhxd+RJGJQbY1XAcx6p/Figw
# qBasSDUxp+0yFPEBB9uBE3KILAtq6fczGp4EMeon6YmkyCGAtXMKDFQQgdP/ITe7
# VghAVbPTVlP3hY1dFgc+t8YK2obFSFVKslkASATDHulCMht+WrIsukclEUP9DaMm
# pq7S0RLODMicI6PtqqGOhdnaRltA0d+Wf+0tPt9SUVtrPJyO7WMPKbykCRXzmHK0
# 6zr0kn1YiUYNXCsOgaHF5ImO2ZwQ54UE1I55jjUdldyjy/UPJgxRm9NyXeO7adYr
# 8K8f6Q2nPF0vWqFG7ewwaAl5ClKerzshfhB8zujVR0d1Ra7Z01lnXYhWuPqVZayF
# l7JHr6i6huhpU6BQ6/VgY0cBiksX4mNM+ISY81T1RYt7fWATNu/zkjINczipzbfg
# 5S+3fCAo8gVB6+6A5L0vBg39dsFITv6MWJuQ8ZZy7fwlFBZE4d5IFbRudakNwKGd
# yLGM2otaNq7wm3ku7x41UGAmkDCCB3EwggVZoAMCAQICEzMAAAAVxedrngKbSZkA
# AAAAABUwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRl
# IEF1dGhvcml0eSAyMDEwMB4XDTIxMDkzMDE4MjIyNVoXDTMwMDkzMDE4MzIyNVow
# fDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMd
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwggIiMA0GCSqGSIb3DQEBAQUA
# A4ICDwAwggIKAoICAQDk4aZM57RyIQt5osvXJHm9DtWC0/3unAcH0qlsTnXIyjVX
# 9gF/bErg4r25PhdgM/9cT8dm95VTcVrifkpa/rg2Z4VGIwy1jRPPdzLAEBjoYH1q
# UoNEt6aORmsHFPPFdvWGUNzBRMhxXFExN6AKOG6N7dcP2CZTfDlhAnrEqv1yaa8d
# q6z2Nr41JmTamDu6GnszrYBbfowQHJ1S/rboYiXcag/PXfT+jlPP1uyFVk3v3byN
# pOORj7I5LFGc6XBpDco2LXCOMcg1KL3jtIckw+DJj361VI/c+gVVmG1oO5pGve2k
# rnopN6zL64NF50ZuyjLVwIYwXE8s4mKyzbnijYjklqwBSru+cakXW2dg3viSkR4d
# Pf0gz3N9QZpGdc3EXzTdEonW/aUgfX782Z5F37ZyL9t9X4C626p+Nuw2TPYrbqgS
# Uei/BQOj0XOmTTd0lBw0gg/wEPK3Rxjtp+iZfD9M269ewvPV2HM9Q07BMzlMjgK8
# QmguEOqEUUbi0b1qGFphAXPKZ6Je1yh2AuIzGHLXpyDwwvoSCtdjbwzJNmSLW6Cm
# gyFdXzB0kZSU2LlQ+QuJYfM2BjUYhEfb3BvR/bLUHMVr9lxSUV0S2yW6r1AFemzF
# ER1y7435UsSFF5PAPBXbGjfHCBUYP3irRbb1Hode2o+eFnJpxq57t7c+auIurQID
# AQABo4IB3TCCAdkwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEEAYI3FQIEFgQU
# KqdS/mTEmr6CkTxGNSnPEP8vBO4wHQYDVR0OBBYEFJ+nFV0AXmJdg/Tl0mWnG1M1
# GelyMFwGA1UdIARVMFMwUQYMKwYBBAGCN0yDfQEBMEEwPwYIKwYBBQUHAgEWM2h0
# dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvRG9jcy9SZXBvc2l0b3J5Lmh0
# bTATBgNVHSUEDDAKBggrBgEFBQcDCDAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMA
# QTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBTV9lbL
# j+iiXGJo0T2UkFvXzpoYxDBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1p
# Y3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXRfMjAxMC0w
# Ni0yMy5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIz
# LmNydDANBgkqhkiG9w0BAQsFAAOCAgEAnVV9/Cqt4SwfZwExJFvhnnJL/Klv6lwU
# tj5OR2R4sQaTlz0xM7U518JxNj/aZGx80HU5bbsPMeTCj/ts0aGUGCLu6WZnOlNN
# 3Zi6th542DYunKmCVgADsAW+iehp4LoJ7nvfam++Kctu2D9IdQHZGN5tggz1bSNU
# 5HhTdSRXud2f8449xvNo32X2pFaq95W2KFUn0CS9QKC/GbYSEhFdPSfgQJY4rPf5
# KYnDvBewVIVCs/wMnosZiefwC2qBwoEZQhlSdYo2wh3DYXMuLGt7bj8sCXgU6ZGy
# qVvfSaN0DLzskYDSPeZKPmY7T7uG+jIa2Zb0j/aRAfbOxnT99kxybxCrdTDFNLB6
# 2FD+CljdQDzHVG2dY3RILLFORy3BFARxv2T5JL5zbcqOCb2zAVdJVGTZc9d/HltE
# AY5aGZFrDZ+kKNxnGSgkujhLmm77IVRrakURR6nxt67I6IleT53S0Ex2tVdUCbFp
# AUR+fKFhbHP+CrvsQWY9af3LwUFJfn6Tvsv4O+S3Fb+0zj6lMVGEvL8CwYKiexcd
# FYmNcP7ntdAoGokLjzbaukz5m/8K6TT4JDVnK+ANuOaMmdbhIurwJ0I9JZTmdHRb
# atGePu1+oDEzfbzL6Xu/OHBE0ZDxyKs6ijoIYn/ZcGNTTY3ugm2lBRDBcQZqELQd
# VTNYs6FwZvKhggLOMIICNwIBATCB+KGB0KSBzTCByjELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMcTWljcm9zb2Z0IEFtZXJpY2Eg
# T3BlcmF0aW9uczEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046RDZCRC1FM0U3LTE2
# ODUxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2WiIwoBATAH
# BgUrDgMCGgMVAOIASP0JSbv5R23wxciQivHyckYooIGDMIGApH4wfDELMAkGA1UE
# BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAc
# BgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0
# IFRpbWUtU3RhbXAgUENBIDIwMTAwDQYJKoZIhvcNAQEFBQACBQDoV9ipMCIYDzIw
# MjMwNzExMjEzMzI5WhgPMjAyMzA3MTIyMTMzMjlaMHcwPQYKKwYBBAGEWQoEATEv
# MC0wCgIFAOhX2KkCAQAwCgIBAAICIrwCAf8wBwIBAAICEZMwCgIFAOhZKikCAQAw
# NgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAqAKMAgCAQACAwehIKEKMAgC
# AQACAwGGoDANBgkqhkiG9w0BAQUFAAOBgQAmatB4OqorgIYPz0/r9OMv8re+xNTv
# GY8dL9Y+JqZlTrxbU1Nbbdl6MCJmsiMT+H3eZhrETPlCpb//dQGK10nBQZueDs1B
# bOv0hJciC5TS8QDBbiL8RZDjnP9FQSkso6Qmsgt3tnJNPynimEZT+ytU8VAyInVq
# evKSeyN5jsodmTGCBA0wggQJAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQI
# EwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3Nv
# ZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBD
# QSAyMDEwAhMzAAABx/sAoEpb8ifcAAEAAAHHMA0GCWCGSAFlAwQCAQUAoIIBSjAa
# BgkqhkiG9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEIPS5+S6l
# tueiC4NfGMksbrt2S3q9OGvlP7WazLQpIr9KMIH6BgsqhkiG9w0BCRACLzGB6jCB
# 5zCB5DCBvQQgR+fl2+JSskULOeVYLbeMgk7HdIbREmAsjwtcy6MJkskwgZgwgYCk
# fjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
# UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQD
# Ex1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAcf7AKBKW/In3AAB
# AAABxzAiBCBebdHw6iRwU3G6HNJtkCNvtWg2+UARRSKVVEs1Z9cpiDANBgkqhkiG
# 9w0BAQsFAASCAgCDda9v2BnTwJkGtaWgjCVv8dO7nbCSp7p1UXwC6o9EJQ90WuuU
# uHeDii60UPQTh70PKbiEUUTaLPOECqsKTEhI8nnLHnVDnhLqtoRuhf1DWDdfWOff
# MNN3PmWbJ+bWHJWHKxu68o/QPzAo7SEj/S6qlVHgdRog7dSm1b4PweZDjswwLONI
# kr/HqdfV1nmQBfia3tNQhkajBnTP16MheXoU72rw4KkkGMN2bNx/ZtQcfuVHzKre
# D3n7b0+x40xlqePVYcJ50dSF2i5EOk2ctORd6qAjLyhh34gN3/gMHpsffF3uO6k5
# opjScu1ZV0qJaGQ9i2hm25kD8b+YwN8xxmFgYxFTMa4XE1+1ujbz999BuEflvzcg
# d3sRiUf6xNT+ndU/pUMkYwlio38jNfnL+ohDxOEP4h2fBQw/t3wiFkSwrRY0CscK
# aPap0ZCcxRstJ3cuyfO0pshgWN9G340p9jrbQs/OLQ2GGAKgA4iBDR2jGibh6NjF
# elyIDu9ZeQtLd5jUAV2hKAk+ZoFCOSvkGUOITXsUfNrKHywdSnH2a1I9AWgQRssD
# +zjiLEf8WFi7Hly8w+2d357cIpGu5gVvoWl9XD0YjFPyndQMnpfXCCagaxWVSKWJ
# oriXXkdOXNPwDe127zkpZ6YSrW3gMjsHmRj7exh59LDwEwEzuqiyGmeztQ==
# SIG # End signature block
