$script:versionId = "1.0001"

$regKey = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows Advanced Threat Protection"

$SampleSharingRegValue = "AllowSampleCollection"

# EDR settings name prefix
$global:settingNamePrefix = "Windows:Device:WindowsAdvancedThreatProtection.Configuration."

$settingsNamesMapping = @{
    SampleSharing = "SampleSharing";
    ForceDefenderPassiveMode = "ForceDefenderPassiveMode";
}

# Map to name in registry
$settingsRegNamesMapping = @{
    SampleSharing = $SampleSharingRegValue;
    ForceDefenderPassiveMode = "ForceDefenderPassiveMode";
}

# Supported on Server SKU only
$isApplicableOnlyOnServer = @(
   $settingsNamesMapping.ForceDefenderPassiveMode 
)

function IsSupportedParam([string] $name)
{
    if (!($name.StartsWith($global:settingNamePrefix)))
    {
        return $false
    }

    $nameSuffix = $name.Replace($global:settingNamePrefix, "")

    return $settingsNamesMapping.ContainsKey($nameSuffix)
}

function Setup() 
{
    LogInfo (-join("EDR setup (", $script:versionId, ")"))
    return
}

function ResetToDefault([string] $name, [string] $value, [ref] $errorCode, [ref] $errorString)
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $nameSuffix = $name.Replace($global:settingNamePrefix,"")
    LogMessage ([verbosityLevel]::verbosity_info) (-join("EDR remove setting ",$nameSuffix,",",$value))

    try 
    {
        if ($settingsNamesMapping.ContainsKey($nameSuffix))
        {
            Remove-ItemProperty -Path $regKey -Name $settingsRegNamesMapping.$nameSuffix
        }
        else 
        {
            $errorString.Value = (-join("EDR:ResetToDefault Setting is not supported ",$nameSuffix))
            $errorCode.Value = [HResult]::E_NOINTERFACE.value__
            LogMessage ([verbosityLevel]::verbosity_warn) $errorString.Value
        }  
          
    } catch {
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
        LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR::ResetToDefault failure for ",$nameSuffix,". HResult:",$errorCode," error message:",$_.Exception.Message))
    }
}

function Set([string] $name, [string] $value, [ref] $errorCode, [ref] $errorString)
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $nameSuffix = $name.Replace($global:settingNamePrefix,"")
    LogMessage ([verbosityLevel]::verbosity_info) (-join("EDR Set setting ",$nameSuffix,",",$value))

    try 
    {
        if ($settingsNamesMapping.ContainsKey($nameSuffix)) {
            Set-ItemProperty -Path $regKey -Name $settingsRegNamesMapping.$nameSuffix -Value ([System.Convert]::ToInt32($value, 10)) -Type DWord
        }
        else
        {
            $errorString.Value = (-join("EDR:Set Setting is not supported ", $nameSuffix)) 
            $errorCode.Value = [HResult]::E_NOINTERFACE.value__
            LogMessage ([verbosityLevel]::verbosity_warn) $errorString.Value
        }       
    }
    catch
    {
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
        LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR::Set failure for ",$nameSuffix,", value:",$value,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
    }
}

function VerifyAssignment([string] $name, [string] $value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $errorString)
{    
    $errorCode.Value = [HResult]::S_OK.value__ #error code is returned by Set
    $errorString.Value = ""
    $nameSuffix = $name.Replace($global:settingNamePrefix,"")
    LogMessage ([verbosityLevel]::verbosity_info) (-join("EDR VerAssign ",$nameSuffix,",",$value))

    try
    {
        if ($settingsNamesMapping.ContainsKey($nameSuffix))
        {
            $currentRegItem = Get-ItemProperty $regkey $settingsRegNamesMapping.$nameSuffix -ErrorAction SilentlyContinue
            if (($null -eq $currentRegItem) -or ($currentRegItem.Length -eq 0))
            {
                LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR:VerAssign returns false for ",$nameSuffix," no such registry value: ",(Join-Path $regkey $settingsRegNamesMapping.$nameSuffix)))
                $enforceSucceeded.Value = $false
                
            } 
            else
            {   
                $current = $currentRegItem.($settingsRegNamesMapping.$nameSuffix) 
                $expected = ([System.Convert]::ToInt32($value, 10))
                if ($current -eq $expected) 
                { 
                    $enforceSucceeded.Value = $true 
                } 
                else 
                { 
                    LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR:VerAssign returns false for ",$nameSuffix," (current:",$current,", expected:",$expected,")"))
                    $enforceSucceeded.Value = $false 
                }
            }
            return [HResult]::S_OK
        }
        else 
        {
            LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR:VerAssign Setting is not supported ", $nameSuffix))
            $enforceSucceeded.Value = $false
            return [HResult]::E_NOINTERFACE
        }
    }  
    catch
    {
        $errorCode.Value = $_.Exception.HResult
        $errorString.Value = $_.Exception.Message
        LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR::VerAssign failure for ",$nameSuffix,", value:",$value,". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
        $enforceSucceeded.Value = $false
        return [HResult]::E_UNEXPECTED
    }
}

function VerifyRemoval([string] $name, [string] $value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $errorString)
{
    $errorCode.Value = [HResult]::S_OK.value__ #error code is returned by ResetToDefault
    $errorString.Value = ""
    $nameSuffix = $name.Replace($global:settingNamePrefix,"")
    LogMessage ([verbosityLevel]::verbosity_info) (-join("EDR VerRemov ",$nameSuffix,",",$value))

    if ($settingsNamesMapping.ContainsKey($nameSuffix))
    {
        $currentRegItem = Get-ItemProperty $regkey $settingsRegNamesMapping.$nameSuffix -ErrorAction SilentlyContinue
        if (($null -eq $currentRegItem) -or ($currentRegItem.Length -eq 0))
        {
            $enforceSucceeded.Value = $true
        } 
        else
        {      
            $enforceSucceeded.Value = $false      
            LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR:VerRemov returns false for ",$nameSuffix))
        } 
        return [HResult]::S_OK
    }  
    else 
    {
        LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR:VerRemov Setting is not supported ", $name.Replace($global:settingNamePrefix,"")))
        return [HResult]::E_NOINTERFACE
    }
}

 function IsServer()
{
    foreach($serverStr in $global:ServersStrs)
    {
        if($global:OSVersion -match $serverStr)
        {
            return $true
        }
    }
    return $false
 }

function IsApplicable([string] $name, [string] $value)
{
    $nameSuffix = $name.Replace($global:settingNamePrefix, "")
    if ($settingsNamesMapping.ContainsKey($nameSuffix))
    {
        if(!(IsServer) -and ($settingsNamesMapping.$nameSuffix -in $isApplicableOnlyOnServer))
        {
            LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR:IsApplicable Setting is not supported on servers ", $name.Replace($global:settingNamePrefix,"")))
            return $false
        }
        return $true
    } 
    else 
    {
        LogMessage ([verbosityLevel]::verbosity_warn) (-join("EDR:IsApplicable Setting is not supported ", $name.Replace($global:settingNamePrefix,"")))
        return $false
    }
}

function CompletionActions() {}


function IsMergeSupported{
    param (
        [string]$name
    )
    return $false
}

#RetrieveTestInfo - returns module info for the TestRunner. Returning here any variable and struct needed for running the tests for this module
function RetrieveTestInfo() {
    return $regKey
}
# SIG # Begin signature block
# MIIntwYJKoZIhvcNAQcCoIInqDCCJ6QCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCApWq9gyapYKtNS
# ufH2OyPab7JLpA0w1G8ty7fHg0dbOqCCDZcwggYVMIID/aADAgECAhMzAAADcaWg
# nFyRHuruAAAAAANxMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMwNTExMTg0MTEzWhcNMjQwNTA4MTg0MTEzWjCBlDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjE+MDwGA1UEAxM1TWlj
# cm9zb2Z0IFdpbmRvd3MgRGVmZW5kZXIgQWR2YW5jZWQgVGhyZWF0IFByb3RlY3Rp
# b24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDQl+fOHFGkNfDHrcv3
# g8shIsUp58WmSbtNm4uu7Sw75Wgsgo5p2yYGGqiRWbu0eIKhAJ1bRp9Qa7vAjMEw
# d0wf9wlLjDNAITzDsokwoCl0rOgWxzYuGSusZq4e/liClIpY5fuxrpsEnsssR1o6
# jITkdJQ6ZkWsTFETxLLOEGxe9BkTkXZaBOHlhTt5QIq6UB6j8zNN/Vo4rFmq751X
# lQLA1jOnNi5NOS3/11TW16ZCo9XKG22zdNsJCUckT/QH2eeQNgpuyfIvP06o+RpP
# Uo7P4xdc4c74S3HKfRiNuqbs3+xvb22GtXIz3cKFDhILpNKN/SC8qLjizq0PxwL1
# lr1ZAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBggrBgEFBQcDAwYKKwYBBAGCN0wv
# ATAdBgNVHQ4EFgQUcR5wDY+Mh5A/I7scm54A8ODeSVcwRQYDVR0RBD4wPKQ6MDgx
# HjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEWMBQGA1UEBRMNNDUxODk0
# KzUwMTAwNTAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8E
# TTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9N
# aWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBR
# BggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0
# cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAw
# DQYJKoZIhvcNAQELBQADggIBAKYhzPfo4pWE2MYg2W1VnHJ7NZ/nwEpGYEz7Qhqc
# bPj2sJFFhnUzYHOAyXz4vP0+ERhFw9lrqLvoyj8s6xQU11QPQVItopahH65USGWh
# UcyrXFBVSx7MG3Dwi+NT6rxq9MjY3KLB1hZPF7hgtOjWgtjXgeJOsy3IjPiRkDwX
# 5b5Pg6o8Xpyt84mpvAPsDKO7yJGJq8V528deIYh+0DcELvUZTtl4Kz8ElT4BxF6i
# XxXkuqpCWO/rwEa6HrLNV2udND/AImVio09bbZXOWBcb9uTTNezYX5fZvinIX7hl
# UHHIDaVUqqwTFtac8q7BLE02T0P7NAeBtayujIQFzGNuwLN302UmUWAyc4LynqZa
# h2OmZWcMODUCuHIMP3pk54XjWt0YYEWPFAbA7T7j8wjby/vPXdG93LybID/yjGDb
# U1t5xMoEpN5rvSUM1U42iJ1tF268/nE2/8nKqUpdJXH/MaXxYU3FnbshmIe1xfxF
# NjeMYN4jg+5gjz5SBzHAhT1X9QWXNOhAXxib3wcchlsxFYtSNihHLEJ2GuckRuPi
# 47VWi/3hTTnl88S8r3jsbWLatGbQxcbrsF4sglAvFD9SQSXLGUsE5wz/1p/Gb1Fh
# qfNBc6E4AWVj9+A1KezSnMBAw8gJk3fiRkdxfSfpLOkHvgA7qsOmQ7qwjs05kWZf
# 055CMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWlj
# cm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4
# MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3Y
# bqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUB
# FDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnbo
# MlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT
# +OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuy
# e4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEh
# NSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2
# z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3
# s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78Ic
# V9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E
# 11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5P
# M4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcV
# AQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAf
# BgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBL
# hklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggr
# BgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsG
# AQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwA
# ZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0G
# CSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDB
# ZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc
# 8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYq
# wooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu
# 5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWI
# UUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXh
# j38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yH
# PgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtI
# EJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA9Z74v2u3S5fi63V4Guzq
# N5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgR
# MiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQ
# zTGCGXYwghlyAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
# dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9y
# YXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEC
# EzMAAANxpaCcXJEe6u4AAAAAA3EwDQYJYIZIAWUDBAIBBQCgga4wGQYJKoZIhvcN
# AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUw
# LwYJKoZIhvcNAQkEMSIEIEY/L6DbidSjN1SjfjselJNfkwFx664jH6V0d6a3+5TQ
# MEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8AcwBvAGYAdKEagBhodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEArx7Pu8HmowuR
# kf1qbda2xLqqnnoqq4racg9VZJ3MFZuemuSuN6C+o7kmuxSvRHkWt5HN0w0kKhgT
# 3eAX3OXiUAl7YEEw4gM5iIhuH/mzavrrvZ+w5VuMLVia/zLQHAYZZQC2Z2U4+W0l
# ZfwNJxGJRrszYW00TmN7XJ2W5wuM0hT4D+fpk/G+iMesHytEV/1zGHk1NGNR428A
# jGLPg05FRojGUIizC2iAEStJN9lbCkxwBcgnomn+kZoxwpe6R8b7nLOcpsvBrxqe
# A4EPyf0qiaLGdrghpeL7q+ItKaBDKF41LaNXIz/6e1TnNec5VZszRqiGfJosplf8
# LyWUdwBuFKGCFwAwghb8BgorBgEEAYI3AwMBMYIW7DCCFugGCSqGSIb3DQEHAqCC
# FtkwghbVAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFRBgsqhkiG9w0BCRABBKCCAUAE
# ggE8MIIBOAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCDd3E4ZAIdi
# 9A7vH559DSXtB/9+TJYGas16SZ9UQm5zPQIGZIts0TV1GBMyMDIzMDcxMTE3NDcw
# MC40ODNaMASAAgH0oIHQpIHNMIHKMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBPcGVyYXRpb25z
# MSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjoyMjY0LUUzM0UtNzgwQzElMCMGA1UE
# AxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaCCEVcwggcMMIIE9KADAgEC
# AhMzAAABwT6gg5zgCa/FAAEAAAHBMA0GCSqGSIb3DQEBCwUAMHwxCzAJBgNVBAYT
# AlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYD
# VQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBU
# aW1lLVN0YW1wIFBDQSAyMDEwMB4XDTIyMTEwNDE5MDEyN1oXDTI0MDIwMjE5MDEy
# N1owgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
# EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAjBgNV
# BAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJjAkBgNVBAsTHVRoYWxl
# cyBUU1MgRVNOOjIyNjQtRTMzRS03ODBDMSUwIwYDVQQDExxNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBTZXJ2aWNlMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA
# 5LHXMydw2hUC4pJU0I5uPJnMeRm8LKC4xaIDu3Fxx3IpZ/We2qXLj4NOmow/WPFe
# Y4vaT4/S4T9xoDsFGg5wEJM6OLZVfa7BUNu0tDt4rkl7QBYNHzz6pcr9bwaq2qm7
# x6P9yi5W0Y8sjoj+QTgtmmXoxCoNXhJ1oG6GbqADQXDZkTcDjIAiteE6TxrhBpIb
# 7e6upifTGZNfcChPfuzHq61FSIwJ0XCxcaR1BwAlSKhb/NUOuQGPr9Zzd6OnIcA+
# RctxwKgfOKB9aWEEHlt0jhKKgpEBvcJnMMP+WaTwmMhob1e+hoCEFx/nI0YHupi6
# 082kFdNFraE72msOYQrwrUyWCeSmN202LZDpTzxZVty6QrBOk+f+BErsR+M5evkK
# uUTWVJHI3vtNgb6K5+gk6EuQw0ocsDdspiPp+qlxBaW50yUbr6wnfzYjJh7QkPcf
# BIZbJAhWQHaV0uS3T7OkObdCssCRMWH7VWUAeSbemuUqOXCR7rdpFTfY/SXKO9lC
# IQBAQSh+wzwh5Zv1b+jT2zWwVl82By3YHmST8b8CKnRXSCjLtgoyy7ERLwkbzPIk
# CfBXcyVneC1w2/wUnqPiAjK0wQfztfXFfoMQr8YUcLHnAtek8OVNPuRIV6bcERbF
# 6rtFXmnjjD4ZwVxIZ/HM4cjeVGsEwkFA9XTzqX9W1P8CAwEAAaOCATYwggEyMB0G
# A1UdDgQWBBRfr2MJ6x7yE+gP5uX9xWGTwpRC+jAfBgNVHSMEGDAWgBSfpxVdAF5i
# XYP05dJlpxtTNRnpcjBfBgNVHR8EWDBWMFSgUqBQhk5odHRwOi8vd3d3Lm1pY3Jv
# c29mdC5jb20vcGtpb3BzL2NybC9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENB
# JTIwMjAxMCgxKS5jcmwwbAYIKwYBBQUHAQEEYDBeMFwGCCsGAQUFBzAChlBodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY3Jvc29mdCUyMFRp
# bWUtU3RhbXAlMjBQQ0ElMjAyMDEwKDEpLmNydDAMBgNVHRMBAf8EAjAAMBMGA1Ud
# JQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBCwUAA4ICAQBfuiaBgsecHvM90RZi
# TDlfHblL09r9X+5q9ckuMR0Bs1Sr5B2MALhT5Y0R3ggLufRX6RQQbSc7WxRXIMr5
# tFEgB5zy/7Yg81Cn2dhTf1GzjCb7/n3wtJSGtr2QwHsa1ehYWdMfi+ETLoEX1G79
# VPFrs0t6Giwpr74tv+CLE3s6m10VOwe80wP4yuT3eiFfqRV8poUFSdL2wclgQKoS
# wbCpbJlNC/ESaDQbbQFli9uO5j2f/G7S4TMG/gyyxvMQ5QJui9Fw2s7qklmozQoX
# 2Ah4aKubKe9/VZveiETNYl1AZPj0kj1g51VNyWjvHw+Hz1xZekWIpfMXQEi0wrGd
# WeiW4i8l92rY3ZbdHsErFYqzh6FRFOeXgazNsfkLmwy+TK17mA7CTEUzaAWMq5+f
# 9K4Y/3mhB4r6UristkWpdkPWEo8b9tbkdKSY00E+FS5DUtjgAdCaRBNaBu8cFYCb
# Erh9roWDxc+Isv8yMQAUDuEwXSy0ExnIAlcVIrhzL40OsG2ca5R5BgAevGP1Hj9e
# j4l/y+Sh0HVcN9N6LmPDmI/MaU2rEZ7Y+jRfCZ1d+l5DESdLXIxDTysYXkT+3VM/
# 1zh6y2s0Zsb/3vPaGnp2zejwf2YlGWl1XpChNZTelF5eOCCfSzUUn3qHe7IyyDKh
# ahgbnKpmwcEkMVBs+RHbVkNWqDCCB3EwggVZoAMCAQICEzMAAAAVxedrngKbSZkA
# AAAAABUwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRl
# IEF1dGhvcml0eSAyMDEwMB4XDTIxMDkzMDE4MjIyNVoXDTMwMDkzMDE4MzIyNVow
# fDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMd
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwggIiMA0GCSqGSIb3DQEBAQUA
# A4ICDwAwggIKAoICAQDk4aZM57RyIQt5osvXJHm9DtWC0/3unAcH0qlsTnXIyjVX
# 9gF/bErg4r25PhdgM/9cT8dm95VTcVrifkpa/rg2Z4VGIwy1jRPPdzLAEBjoYH1q
# UoNEt6aORmsHFPPFdvWGUNzBRMhxXFExN6AKOG6N7dcP2CZTfDlhAnrEqv1yaa8d
# q6z2Nr41JmTamDu6GnszrYBbfowQHJ1S/rboYiXcag/PXfT+jlPP1uyFVk3v3byN
# pOORj7I5LFGc6XBpDco2LXCOMcg1KL3jtIckw+DJj361VI/c+gVVmG1oO5pGve2k
# rnopN6zL64NF50ZuyjLVwIYwXE8s4mKyzbnijYjklqwBSru+cakXW2dg3viSkR4d
# Pf0gz3N9QZpGdc3EXzTdEonW/aUgfX782Z5F37ZyL9t9X4C626p+Nuw2TPYrbqgS
# Uei/BQOj0XOmTTd0lBw0gg/wEPK3Rxjtp+iZfD9M269ewvPV2HM9Q07BMzlMjgK8
# QmguEOqEUUbi0b1qGFphAXPKZ6Je1yh2AuIzGHLXpyDwwvoSCtdjbwzJNmSLW6Cm
# gyFdXzB0kZSU2LlQ+QuJYfM2BjUYhEfb3BvR/bLUHMVr9lxSUV0S2yW6r1AFemzF
# ER1y7435UsSFF5PAPBXbGjfHCBUYP3irRbb1Hode2o+eFnJpxq57t7c+auIurQID
# AQABo4IB3TCCAdkwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEEAYI3FQIEFgQU
# KqdS/mTEmr6CkTxGNSnPEP8vBO4wHQYDVR0OBBYEFJ+nFV0AXmJdg/Tl0mWnG1M1
# GelyMFwGA1UdIARVMFMwUQYMKwYBBAGCN0yDfQEBMEEwPwYIKwYBBQUHAgEWM2h0
# dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvRG9jcy9SZXBvc2l0b3J5Lmh0
# bTATBgNVHSUEDDAKBggrBgEFBQcDCDAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMA
# QTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBTV9lbL
# j+iiXGJo0T2UkFvXzpoYxDBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1p
# Y3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXRfMjAxMC0w
# Ni0yMy5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIz
# LmNydDANBgkqhkiG9w0BAQsFAAOCAgEAnVV9/Cqt4SwfZwExJFvhnnJL/Klv6lwU
# tj5OR2R4sQaTlz0xM7U518JxNj/aZGx80HU5bbsPMeTCj/ts0aGUGCLu6WZnOlNN
# 3Zi6th542DYunKmCVgADsAW+iehp4LoJ7nvfam++Kctu2D9IdQHZGN5tggz1bSNU
# 5HhTdSRXud2f8449xvNo32X2pFaq95W2KFUn0CS9QKC/GbYSEhFdPSfgQJY4rPf5
# KYnDvBewVIVCs/wMnosZiefwC2qBwoEZQhlSdYo2wh3DYXMuLGt7bj8sCXgU6ZGy
# qVvfSaN0DLzskYDSPeZKPmY7T7uG+jIa2Zb0j/aRAfbOxnT99kxybxCrdTDFNLB6
# 2FD+CljdQDzHVG2dY3RILLFORy3BFARxv2T5JL5zbcqOCb2zAVdJVGTZc9d/HltE
# AY5aGZFrDZ+kKNxnGSgkujhLmm77IVRrakURR6nxt67I6IleT53S0Ex2tVdUCbFp
# AUR+fKFhbHP+CrvsQWY9af3LwUFJfn6Tvsv4O+S3Fb+0zj6lMVGEvL8CwYKiexcd
# FYmNcP7ntdAoGokLjzbaukz5m/8K6TT4JDVnK+ANuOaMmdbhIurwJ0I9JZTmdHRb
# atGePu1+oDEzfbzL6Xu/OHBE0ZDxyKs6ijoIYn/ZcGNTTY3ugm2lBRDBcQZqELQd
# VTNYs6FwZvKhggLOMIICNwIBATCB+KGB0KSBzTCByjELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMcTWljcm9zb2Z0IEFtZXJpY2Eg
# T3BlcmF0aW9uczEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046MjI2NC1FMzNFLTc4
# MEMxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2WiIwoBATAH
# BgUrDgMCGgMVAESKOtSK7RVVK+Si+aqFd0YSY+VPoIGDMIGApH4wfDELMAkGA1UE
# BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAc
# BgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0
# IFRpbWUtU3RhbXAgUENBIDIwMTAwDQYJKoZIhvcNAQEFBQACBQDoV4jpMCIYDzIw
# MjMwNzExMTU1MzEzWhgPMjAyMzA3MTIxNTUzMTNaMHcwPQYKKwYBBAGEWQoEATEv
# MC0wCgIFAOhXiOkCAQAwCgIBAAICDwcCAf8wBwIBAAICEh4wCgIFAOhY2mkCAQAw
# NgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAqAKMAgCAQACAwehIKEKMAgC
# AQACAwGGoDANBgkqhkiG9w0BAQUFAAOBgQBYd266tvAZlgwqysx5Pir83XBLxNQx
# YwvYYsVZ4qeI0tOedZUd3GpTv2C9QcqR6EYnFoN4APfFasu+fZWSKcF732gNWDka
# RAqMBzo8SOolKYZiXvKxxVp8KHtcELjq7X3IbjZCbKAa/Z+mQrtuO3/t7bpwueDJ
# PcjqqxkZ8Mu7cTGCBA0wggQJAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQI
# EwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3Nv
# ZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBD
# QSAyMDEwAhMzAAABwT6gg5zgCa/FAAEAAAHBMA0GCWCGSAFlAwQCAQUAoIIBSjAa
# BgkqhkiG9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEILcya3Dw
# 03sMiHxbgSOAMsSbGECiJ6GzKAv8moByqljdMIH6BgsqhkiG9w0BCRACLzGB6jCB
# 5zCB5DCBvQQgCrkg6tgYHeSgIsN3opR2z7EExWA0YkirkvVYSTBgdtQwgZgwgYCk
# fjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
# UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQD
# Ex1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAcE+oIOc4AmvxQAB
# AAABwTAiBCAUNpm33LInpKsOwdWidx2Ge/N1xjDz7VXRKBrMryzclTANBgkqhkiG
# 9w0BAQsFAASCAgBupyN/DARb7GeVkpKzNxQ8XmlU8EC97x+OQiC2U4TCGiGFNyi6
# VvROu0rUmTFzv51EbE0XN+LQGXXNuB5ykRVSTbLjTBs5PaNTODsIMI6JNWHMNBYD
# OCQTJKTdxsHuKZsxPELcSB0Uk1ANtTmIb0k81BVdibSPfscjJuApNhm/abLSj29Q
# ZbPADUGzCLHyY7tdbjoc0QEp4F9axbZlDT04JeTds4FLxecNgZxjJXMF76w953rx
# Yq/GGGVAGdvkOn5iKtkW4HSmISmnEhz4W6jIrOTyNOlkmZTuDerCg0IjM2YOZdY6
# OHjRJkx4oruWjczKSePqpWaLjlRgJeVJrBYuqQ9+nHsoH7VzKxiRn5iWh0jtVT5g
# D99ydbWOlFNmBN72og7EEsP/hDcV8xYT2enfwt3o3sKifHx8zpQ0LDytYSDYGtZQ
# zAWPQIPiYRGB5X4KtiffcMG4ZFXmRWk/cwmsoFtxphZXbRCqIv21AnKExGF6wgsl
# dNU3h4Bk3Oqd8LDzwe2N/CHFa4gtIsC4AygXs0pgtQvICbZDT0mFEuPMGt6e8lMR
# 03XZ+iSRpdTYb1T+dVQtC4nAc/3T3YSo6HS/m94X7+mlWxRjc6qsmjbFuwA86eHc
# oUEHOtr3n0J05LR/ONfTNXDDWdmz6O7UhSXH1DJhNjJt6lgMzqnrouCSAA==
# SIG # End signature block
