####################################################### globals & definitions: #######################################################
$script:versionId = "1.0006"

$regKey = "Registry::HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\SharedAccess\Parameters\FirewallPolicy"
$componentPrefix = "Windows:None:Firewall.MdmStore."
$delimiter = '.'
$timeLimitForSettingAction = 50 #in milliseconds. actions that take longer than that are logged.

# rulesToRmove: Hash-Map that holds the rules (names) that will be removed on CompletionActions(). 
# If the rule was not entirely removed, but it is removed for a single setting Remove (because there is no ResetToDefault for FW on powershell, so we remove an entire rule and create a new rule by the same name without that setting), 
# the setting will be also added to $rulesToSet to be re-added on CompletionActions(). 
# Struct maps rule name to a map of removal result: errCode and errString. example: {my_2_rule {[HResult]::S_OK ""}}
$rulesToRmove = @{} 

# rulesToSet: Hash-Map that holds the rules (and their settings) that will be set on CompletionActions(). 
# Rules added to it in SaveSettingForNewRule(). Maps rule name to a map of setting full path and value. 
# example: {'my_rule_name' {'Windows:None:Firewall.MdmStore.FirewallRules.neomi_2_rule.App.ServiceName' 'my_srv_name'}}
$rulesToSet = @{}
# rulesToSetSplatHash: Hash-Map that hold the rules and their "splat" format, to be used in cmdlet 
#TBD - check option to use $rulesToSet for theis purpose as well: to setup the rule with all it's settings in one go.'
$script:rulesToSetSplatHash = @{} 

# rulesChanged: an array that holds rules that were updated or added (not removed). Used to avoid re-setting unnecessary FW rules as it is an expensive action.
$script:changedRules = @()

# rulesToSetErrors: Hash-Map that holds all the errors that happened in setting of rules (ONLY rules settings. Other settings issue their errors in Set()/RestToDefault(), rules settings issue warnings in VerifyAssignment()/VerifyRemoval() so the result must be saved on CompletionActions()). 
# Maps setting full name to a map of errCode and errStr. 
# example: {'Windows:None:Firewall.MdmStore.FirewallRules.my_2_rule.App.ServiceName' {[HResult]::E_INVALIDARG 'Invalid service name'}}
$rulesToSetErrors = @{} 

# structs to track actions that were made (to avoid unnecessary and expensive cmdlet actions in VerifyAssignment)
$script:anyProfileValueSet = @()
$script:anyGlobalValueSet = $false

# structs to hold current status on device. Used right after completion actions to get values once and avoid expensive repeating calls to cmdlets.
$script:fwRulesCurrentStatus = @{} #will contain output of Get-NetFirewallRule of created rules (and only them)
$script:fwSettingsCurrentStatus = $null #will contain output of Get-NetFirewallSetting
$script:fwProfileCurrentStatus = @{} #will contain key-value pairs: ([$settingScope], [output of Get-NetFirewallProfile for that scope (private/public/domain)])


Add-Type -TypeDefinition @"
   public enum settingNameSegment
   {
      Module,
      MdmStore,
      Scope,
      RuleName
   }
"@

Add-Type -TypeDefinition @"
   public enum settingScope
   {
      GlobalSetting,
      PrivateProfileSetting,
      PublicProfileSetting,
      DomainProfileSetting,
      RuleSetting
   }
"@

Add-Type -TypeDefinition @"
   public enum settingAction
   {
      Set,
      Get
   }
"@

# maps setting name (as sent in policy) to it's FW name (for use in cmdlets)
$settingsNamesMapping = @{
    #global settings:
    DisableStatefulFtp = "EnableStatefulFtp"; # values: true/false
    SaIdleTime  = "MaxSAIdleTimeSeconds";
    PresharedKeyEncoding  = "KeyEncoding"; 
    Exemptions = "Exemptions"; #TBD - get actuall name
    CertValidationLevel = "CertValidationLevel"; #TBD - get actuall name
    RequireFullAuthSupport = "RequireFullAuthSupport"; #TBD - get actuall name
    EnablePacketQueue = "EnablePacketQueuing"; # values: 0 = Disabled; 1 = Queue Inbound; 2 = Queue Outbound; 3 = Quene Inbound and Queue Outbound -> (3) is not supported according to documenation.
    
    #profile settings:
    DisableUnicastResponsesToMulticastBroadcast = "AllowUnicastResponseToMulticast"; #values: GPOBooloean
    DefaultInboundAction = "DefaultInboundAction";
    DefaultOutboundAction  = "DefaultOutboundAction";
    DisableInboundNotifications = "NotifyOnListen";
    EnableFirewall = "Enabled";
    AllowLocalPolicyMerge = "AllowLocalFirewallRules"; #values: GPOBooloean
    AuthAppsAllowUserPrefMerge = "AllowUserApps"; #values: GPOBooloean
    GlobalPortsAllowUserPrefMerge = "AllowUserPorts"; #values: GPOBooloean
    AllowLocalIpsecPolicyMerge = "AllowLocalIPsecRules";
    DisableStealthModeIpsecSecuredPacketExemption = "EnableStealthModeForIPsec";
    Shielded = "AllowInboundRules";
    
    #rule settings:
    Name = "Name";
    Description = "Description";
    Direction  = "Direction";
    'Action.Type' = "Action";
    Enabled = "Enabled";
    Profiles = "Profile";
    'App.PackageFamilyName' = "Package";
    'App.FilePath' = "Program";
    'App.ServiceName' = "Service";
    'App.Fqbn' = "PolicyStore";
    Protocol = "Protocol";
    InterfaceTypes = "InterfaceType";
    LocalUserAuthorizedList = "LocalUser";
    LocalAddressRanges = "LocalAddress";
    RemoteAddressRanges = "RemoteAddress";
    LocalPortRanges = "LocalPort";
    RemotePortRanges = "remotePort";
    EdgeTraversal = "EdgeTraversalPolicy"
}

$settingsNameToRegEntryMapping = @{
    #global settings:
    DisableStatefulFtp = "DisableStatefulFtp";
    EnablePacketQueue = "EnablePacketQueue";
    PresharedKeyEncoding = "PresharedKeyEncoding";
    SaIdleTime = "SAIdlTime";
    Exemptions = "IPSecExempt";
    CertValidationLevel = "StrongCRLCheck";
    RequireFullAuthSupport = "IPsecOpportunisticallyMatchAuthSetPerKM";

    # profile scope:
    DisableUnicastResponsesToMulticastBroadcast = "DisableUnicastResponsesToMulticastBroadcast";
    DefaultInboundAction = "DefaultInboundAction";
    DefaultOutboundAction = "DefaultOutboundAction";
    AllowLocalPolicyMerge = "AllowLocalPolicyMerge";
    DisableInboundNotifications = "DisableNotifications";
    EnableFirewall = "EnableFirewall";
    AuthAppsAllowUserPrefMerge = "AllowUserPrefMerge";
    GlobalPortsAllowUserPrefMerge = "AllowUserPrefMerge";
    AllowLocalIpsecPolicyMerge = "AllowLocalIPsecPolicyMerge";
    DisableStealthModeIpsecSecuredPacketExemption = "DisableStealthModeIPsecSecuredPacketExemption";
    Shielded = "DoNotAllowExceptions"
}

$settingsNameToAdditionsForRegEntryPathMapping = @{
    # profile scope:
    AuthAppsAllowUserPrefMerge  = "AuthorizedApplications";
    GlobalPortsAllowUserPrefMerge  = "GloballyOpenPorts"
}

$scopeToRegEntryMapping = @{  
    [settingScope]::GlobalSetting = "";
    [settingScope]::PublicProfileSetting = "PublicProfile";
    [settingScope]::DomainProfileSetting = "DomainProfile";
    [settingScope]::PrivateProfileSetting = "StandardProfile"; #as name - "Private"
    [settingScope]::RuleSetting = "FirewallRules"
}

$script:profileScopeValues = 
    [settingScope]::PrivateProfileSetting,
    [settingScope]::PublicProfileSetting,
    [settingScope]::DomainProfileSetting

#global setting related values
$PacketQueuingValuesMapping = @{
    "0" = "None";
    "1" = "Inbound";
    "2" = "Forward"
}

#profile setting related values
$DefaultActionValuesMapping = @{ 
    "0" = "Allow";
    "1" = "Block"
}

#rule setting related values
$ActionTypeValuesMapping = @{ 
    "0" = "Block";
    "1" = "Allow"
}

#supported FW rule profile types:
$script:RuleProfileTypesInfoArr = @{"Domain" = 0x001; "Private" = 0x002; "Public" = 0x004;}
$script:RuleProfileTypeAny = 0x7FFFFFFF;
$script:RuleProfileTypesSupportedCombinations = 0
foreach ($ruleProfileValueOption in $script:RuleProfileTypesInfoArr.Values)
{
    $script:RuleProfileTypesSupportedCombinations =  ($script:RuleProfileTypesSupportedCombinations -bor $ruleProfileValueOption)
}
LogDebug "RuleProfileTypesSupportedCombinations: $script:RuleProfileTypesSupportedCombinations"

######### settings by setting scope (profile/global/rule):

$GlobalSettings = 
    $settingsNamesMapping.DisableStatefulFtp,
    $settingsNamesMapping.SaIdleTime,
    $settingsNamesMapping.PresharedKeyEncoding,
    $settingsNamesMapping.Exemptions,
    $settingsNamesMapping.CertValidationLevel,
    $settingsNamesMapping.RequireFullAuthSupport,
    $settingsNamesMapping.EnablePacketQueue

$RuleSettings = 
    $settingsNamesMapping.Description,
    $settingsNamesMapping.Name,
    $settingsNamesMapping.Direction,
    $settingsNamesMapping.'Action.Type',
    $settingsNamesMapping.Enabled,
    $settingsNamesMapping.Profiles,
    $settingsNamesMapping.'App.PackageFamilyName',
    $settingsNamesMapping.'App.FilePath',
    $settingsNamesMapping.'App.ServiceName',
    $settingsNamesMapping.Protocol,
    $settingsNamesMapping.InterfaceTypes,
    $settingsNamesMapping.LocalUserAuthorizedList,
    $settingsNamesMapping.LocalAddressRanges,
    $settingsNamesMapping.RemoteAddressRanges,
    $settingsNamesMapping.LocalPortRanges,
    $settingsNamesMapping.RemotePortRanges,
    $settingsNamesMapping.EdgeTraversal,
    $settingsNamesMapping.'Ap.Fqbn'

$ProfileSettings = 
    $settingsNamesMapping.DisableUnicastResponsesToMulticastBroadcast,
    $settingsNamesMapping.DefaultInboundAction,
    $settingsNamesMapping.DefaultOutboundAction,
    $settingsNamesMapping.DisableInboundNotifications,
    $settingsNamesMapping.EnableFirewall,
    $settingsNamesMapping.AllowLocalPolicyMerge,
    $settingsNamesMapping.AuthAppsAllowUserPrefMerge,
    $settingsNamesMapping.GlobalPortsAllowUserPrefMerge,
    $settingsNamesMapping.AllowLocalIpsecPolicyMerge,
    $settingsNamesMapping.DisableStealthModeIpsecSecuredPacketExemption,
    $settingsNamesMapping.Shielded

######### settings by commonly handled allowed values:
$GPOBooleanValues = 
    #global
    $settingsNamesMapping.DisableStatefulFtp,
    $settingsNamesMapping.RequireFullAuthSupport,
    #profile
    $settingsNamesMapping.DisableUnicastResponsesToMulticastBroadcast,
    $settingsNamesMapping.AllowLocalPolicyMerge,
    $settingsNamesMapping.DisableInboundNotifications,
    $settingsNamesMapping.EnableFirewall,
    $settingsNamesMapping.AuthAppsAllowUserPrefMerge,
    $settingsNamesMapping.GlobalPortsAllowUserPrefMerge,
    $settingsNamesMapping.AllowLocalIpsecPolicyMerge,
    $settingsNamesMapping.DisableStealthModeIpsecSecuredPacketExemption,
    $settingsNamesMapping.Enabled

#Enable in name vs Disable in FW cmd (and vice versa)
$ReversedGPOBooleanValues = 
    $settingsNamesMapping.DisableStatefulFtp,
    $settingsNamesMapping.DisableUnicastResponsesToMulticastBroadcast,
    $settingsNamesMapping.DisableInboundNotifications,
    $settingsNamesMapping.DisableStealthModeIpsecSecuredPacketExemption,
    $settingsNamesMapping.Shielded

$SimpleComparisonValues = 
    $settingsNamesMapping.Exemptions,
    $settingsNamesMapping.CertValidationLevel,
    $settingsNamesMapping.PresharedKeyEncoding,
    $settingsNamesMapping.Name,
    $settingsNamesMapping.Description,
    $settingsNamesMapping.Direction,
    $settingsNamesMapping.Enabled,
    $settingsNamesMapping.LocalPortRanges,
    $settingsNamesMapping.RemotePortRanges,
    $settingsNamesMapping.EdgeTraversal,
    $settingsNamesMapping.'App.Fqbn'

$RuleFilterSettings = 
    $settingsNamesMapping.'App.PackageFamilyName',
    $settingsNamesMapping.'App.FilePath',
    $settingsNamesMapping.LocalPortRanges,
    $settingsNamesMapping.RemotePortRanges,
    $settingsNamesMapping.Protocol,
    $settingsNamesMapping.LocalAddressRanges,
    $settingsNamesMapping.RemoteAddressRanges,
    $settingsNamesMapping.'App.ServiceName',
    $settingsNamesMapping.InterfaceTypes,
    $settingsNamesMapping.LocalUserAuthorizedList

$RuleApplicationFilterSettings = 
    $settingsNamesMapping.'App.PackageFamilyName',
    $settingsNamesMapping.'App.FilePath'

$RulePortFilterSettings = 
    $settingsNamesMapping.LocalPortRanges,
    $settingsNamesMapping.RemotePortRanges,
    $settingsNamesMapping.Protocol

$RulePortRanges = 
    $settingsNamesMapping.LocalPortRanges,
    $settingsNamesMapping.RemotePortRanges

$RuleAddressFilterSettings = 
    $settingsNamesMapping.LocalAddressRanges,
    $settingsNamesMapping.RemoteAddressRanges

$DependentRuleSettings = 
    $settingsNamesMapping.LocalPortRanges,
    $settingsNamesMapping.RemotePortRanges

#Map protocol names to numbers for VerifyAssignment
$ProtocolNamesMapping = @{
    "ICMPv4" = "1";
    "TCP" = "6";
    "UDP" = "17";
    "ICMPv6" = "58"
}

 ####################################################### API functions: #######################################################

function Setup()
{
    LogInfo (-join("Firewall setup (", $script:versionId, ")"))
    LogInfo "FW importing NetSecurity.psd1"
    try 
    {
        Import-Module (Join-Path (Get-CimInstance Win32_OperatingSystem -ErrorAction Stop).SystemDrive "Windows\system32\WindowsPowerShell\v1.0\Modules\NetSecurity\NetSecurity.psd1") #needed for GpoBoolean
    } 
    catch 
    {
        LogWarning (-join("FW::Exception on Setup. HResult: ", $_.Exception.HResult, ". Error message: ", $_.Exception.Message))
    }
}

#removes entry from Registry which makes value "NotConfigured".
function ResetToDefault([string] $name, [string] $value, [ref] $errorCode, [ref] $errorStr)
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorStr.Value = ""
    $loggingName = $name.Replace($componentPrefix,"")
    $startTimeMilli = GetCurrentTimeInMillis

    LogInfo (-join("FW Remove ",$loggingName,",",$value))
    try
    {
        #handle rule settings:
        $settingScope = GetSettingScopeFromFullPath($name)
        if($settingScope -eq [SettingScope]::RuleSetting)
        {
            #firewall rules are removed as a whole, not each setting at a time. Save rule name for later removal (on CompletionActions call).
            $ruleName = GetSettingRuleName($name)
            LogInfo (-join("FW saving rule(",$ruleName,") for removal"))
            SaveFWRuleForRemoval -ruleName $ruleName
            return
        }

        #handle global/profile settings - removed one setting at a time:
        $namePathArr = $name.Split($delimiter)
        $scopeEntry = $scopeToRegEntryMapping.$settingScope #get the reg path part relevant to the scope

        $cleanName = GetSuffixName($namePathArr) 

        $fullPath = $regKey #base reg path 

        if($scopeEntry -ne "") #add scope path to registry path
        {
            $fullPath = $fullPath+"\"+$scopeEntry
        }

        if($settingsNameToAdditionsForRegEntryPathMapping.ContainsKey($cleanName)) #handle settings which have additionl subfolders in their path
        {
            $fullPath = $fullPath+"\"+$settingsNameToAdditionsForRegEntryPathMapping.$cleanName
        }

        $regKey = $settingsNameToRegEntryMapping[$cleanName]
        LogInfo (-join("FW Remove removing registry key ",$fullPath,"\\",$regKey))

        Remove-ItemProperty $fullPath $regKey
    } 
    catch 
    {
        $errorCode.Value = $_.Exception.HResult
        $errorStr.Value = $_.Exception.Message
        LogWarning (-join("FW remove for ",$loggingName,", fullPath:",$fullPath,", regKey:",$regKey,"- HResult:",$errorCode," error message:",$_.Exception.Message))
    }

    $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join ("FW: Remove for ",$loggingName," took ", $totalTimeMilli," ms."))
    }
}

function Set([string] $name, [string] $value, [ref] $errorCode, [ref] $errorStr) 
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorStr.Value = ""
    $loggingName = $name.Replace($componentPrefix,"")
    $startTimeMilli = GetCurrentTimeInMillis

    $settingScope = GetSettingScopeFromFullPath -fullPath $name

    try
    {
        # rules are handled as a "whole rule" because we cannot remove only a single setting in them (they are saves as a long string in the registry, and the default value of each setting may vary between clients/servers).
        # As a consequence, we save the settings to apply and at CompletionActions we delete all rules that any of their setting is removed, and then create/recreate rules that their settings were applied. 
        # that way we know we have ResetToDefault all settings we needed to.
        if($settingScope -eq [SettingScope]::RuleSetting)
        {
            $ruleName = GetSettingRuleName -fullPath $name
            SaveSettingForNewRule -ruleName $ruleName -fullPath $name -value $value

            $loggingName = GetSuffixName($loggingName.Split($delimiter)) 
            LogInfo (-join("FW Set R(",$ruleName,") saved {",$loggingName,",",$value,"}"))
            return 
        } 
        elseif(($settingScope -in $script:profileScopeValues) -and !($settingScope -in $script:anyProfileValueSet))
        {
            $script:anyProfileValueSet += $settingScope
        }
        elseif(!$script:anyGlobalValueSet)
        {
            $script:anyGlobalValueSet = $true
        }

        LogInfo (-join("FW Set:",$loggingName,",",$value))
        #global/profile setting are handled one by one:
        SetValue -workPath $name -value $value -settingScope $settingScope -errorCode $errorCode -errorStr $errorStr
    } 
    catch 
    {
        $errorStr.Value = $_.Exception.Message
        $errorCode.Value = $_.Exception.HResult
        LogWarning (-join("FW: Set failure for ", $loggingName, ". HResult:", $_.Exception.HResult, " error message:", $_.Exception.Message))
    }

    $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("FW: Set for ",$loggingName," took ", $totalTimeMilli," ms."))
    }
    
}

#enforceSucceeded - indication if the Set worked
#errorCode/errorString - error code/String in case of error on specific setting
function VerifyAssignment([string] $name, [string] $value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $errorString) 
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""
    $startTimeMilli = GetCurrentTimeInMillis
    
    $loggingName = $name.Replace($componentPrefix,"")

    $workPath = $name 
    $settingScope = GetSettingScopeFromFullPath($name)
    $ruleName = $null
    try
    {
        if($settingScope -eq [SettingScope]::RuleSetting)
        {
            $ruleName = GetSettingRuleName -fullPath $name
            LogInfo (-join("FW VerAssign R:",$ruleName,",",(GetSuffixName($loggingName.Split($delimiter))),",",$value))

            $rule = $script:fwRulesCurrentStatus | Where {$_.Name -eq $ruleName}
           
            if($rule -eq $null)
            {
                #if rule doesn't exists - enforce failed, check what was the error (if possible)
                LogInfo (-join("FW VerAssign: no rule ",$ruleName))
                $enforceSucceeded.Value = $false 
                if($rulesToSetErrors.ContainsKey($name))
                {
                    $errorCode.Value = $rulesToSetErrors[$name]['errCode']
                    $errorString.Value = $rulesToSetErrors[$name]['errStr']
                }
                $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
                if($totalTimeMilli -gt $timeLimitForSettingAction)
                {
                    LogInfo (-join("FW: VerAssign for ",$loggingName," (",$ruleName,") took ", $totalTimeMilli," ms."))
                }
                return [HResult]::S_OK
            } 
            else
            {
                $workPath = GetRuleSettingWithoutRuleName -fullPath $name -ruleNameToRmove $ruleName
                $namePathArr = $workPath.Split($delimiter)
                $settingSuffixName = GetSuffixName($namePathArr)

                $currentValue, $expectedValue = GetCompareValuesForFWRuleSetting $ruleName $settingSuffixName $value

            }
        }
        else 
        { 
            LogInfo (-join("FW VerAssign ", $loggingName, ", value ", $value))

            $namePathArr = $workPath.Split($delimiter)
            $SettingSuffixName = GetSuffixName($namePathArr)
            $currentValue, $expectedValue = GetCompareValuesForFWGeneralSetting $SettingSuffixName $settingScope $value     
        }
        
        if($null -eq $currentValue)
        {
            $enforceSucceeded.Value = $false #currentValue should not be null in VerAssign (called after Set).
            $errorCode.Value = [HResult]::E_FAIL.value__
            $errorString.Value = -join("Current value for setting ",$loggingName, " is null.")
        }
        elseif($null -eq $expectedValue)
        {
            $msg = -join("FW: VerAssign failure. Setting ",$loggingName, " failed to get expectedValue.")
            LogWarning $msg
            $errorString.Value = $msg
            return [HResult]::E_Fail
        }
        else #Compare-Object can't handle null values so handeling it above
        {
            $enforceSucceeded.Value = ($null -eq (Compare-Object $currentValue $expectedValue)) #Using Compare-Object is not case-sensitive, and can also compare arrays (for RuleAddressFilterSettings)
        }
        
        if(!$enforceSucceeded.Value) 
        {
            #for fw rules, the error string and code are saved when set and should be returned now. for other settings, the error was already returned on Set.
            LogInfo (-join("FW VerAssign returns false for ", $loggingName, " (current:", ($currentValue -join ','), ", expected:", ($expectedValue -join ','), ")"))
            if($rulesToSetErrors.ContainsKey($name))
            {
                $errorCode.Value = $rulesToSetErrors[$name]['errCode']
                $errorString.Value = $rulesToSetErrors[$name]['errStr']
            }
        }
    } 
    catch 
    {
        $errorString.Value = $_.Exception.Message
        $errorCode.Value = $_.Exception.HResult
        LogWarning (-join("FW: VerAssign failure. Setting ",$loggingName," HR:",$_.Exception.HResult," msg:",$_.Exception.Message, " lineNumber:", $_.InvocationInfo.ScriptLineNumber))
        return [HResult]::E_Fail
    }

    $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("FW: VerAssign for ",$loggingName," took ", $totalTimeMilli," ms."))
    }
    return [HResult]::S_OK 
}

#enforceSucceeded - indication if the ResetToDefault worked
#errorCode/errorString - error code/String in case of error on specific setting
function VerifyRemoval([string] $name, [string] $value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $errorString) 
{
    $errorCode.Value = [HResult]::S_OK.value__
    $errorString.Value = ""

    $loggingName = $name.Replace($componentPrefix,"")
    LogInfo (-join("FW VerRemov:",$loggingName,",",$value))
    $startTimeMilli = GetCurrentTimeInMillis

    $settingScope = GetSettingScopeFromFullPath($name)
    try
    {
        #firewall rules handling: for FW rules, errors are saved on CompletionAction and issued here
        if($settingScope -eq [SettingScope]::RuleSetting)
        {
            #Get rule Name
            $ruleName = GetSettingRuleName -fullPath $name
            if($rulesToRmove.ContainsKey($ruleName))
            {
                #rule should have been removed (for full removal or specific setting removal) - check removal result that was saved on CompletionAction
                $removalRes = $rulesToRmove[$ruleName]['errCode']
                LogInfo (-join("FW VerRemov ",$loggingName," rule ",$ruleName," was removed. Removal res: ",$removalRes))
                if($removalRes -eq [HResult]::S_OK.value__)
                {
                    $enforceSucceeded.Value = $true
                } else {
                    $enforceSucceeded.Value = $false
                    $errorCode.Value = $removalRes
                    $errorString.Value = $rulesToRmove[$ruleName]['errStr']
                }
            } 
            else 
            {
                $msg = -join ("FW VerRemov ", $loggingName, " - attempt to remove rule ", $ruleName,"  was not done. Removal failed.")
                LogInfo $msg
                $enforceSucceeded.Value = $false
                $errorCode.Value = [HResult]::E_UNEXPECTED.value__
                $errorString.Value = $msg
            }
            
            $totalTimeMilli = (GetCurrentTimeInMillis) - $startTimeMilli
            if($totalTimeMilli -gt $timeLimitForSettingAction)
            {
                LogInfo (-join("FW: VerRemov for ",$loggingName," (",$ruleName,") took ", $totalTimeMilli," ms."))
            }
            return [HResult]::S_OK
        }

        #Global/Profile settings handling (error code, if needed, issued on Set already):
        $currentRegItem = $null;
        $namePathArr = $name.Split($delimiter)

        $scopeEntry = $scopeToRegEntryMapping.$settingScope
        $cleanName = GetSuffixName($namePathArr)
        $fullPath = $regKey

        if($scopeEntry -ne "")
        {
            $fullPath = $fullPath+"\"+$scopeEntry
        }

        if($settingsNameToAdditionsForRegEntryPathMapping.ContainsKey($cleanName)) #handle settings which have additionl subfolders in their path
        {
            $fullPath = $fullPath+"\"+$settingsNameToAdditionsForRegEntryPathMapping.$cleanName
        }

        $regKey = $settingsNameToRegEntryMapping[$cleanName]
        LogInfo (-join("FW VerRemov checking registry key ",$fullPath,"\\",$regKey))

        $currentRegItem = Get-ItemProperty $fullPath $regKey;

        if (($null -eq $currentRegItem) -or ($currentRegItem.Length -eq 0))
        {
            LogInfo (-join("FW VerRemov ",$loggingName," removal success."))
            $enforceSucceeded.Value = $true 
        } 
        else
        {        
            LogInfo (-join("FW VerRemov ",$loggingName," removal failure."))
            $enforceSucceeded.Value = $false    
        }
    } 
    catch 
    {
        $errorString.Value = $_.Exception.Message
        $errorCode.Value = $_.Exception.HResult 
        LogWarning (-join("FW: VerRemov failure. Setting ",$loggingName," HR:",$_.Exception.HResult," msg:",$_.Exception.Message))
        return [HResult]::E_Fail
    }

    $totalTimeMilli = ((GetCurrentTimeInMillis) - $startTimeMilli)
    if($totalTimeMilli -gt $timeLimitForSettingAction)
    {
        LogInfo (-join("FW: VerRemov for ",$loggingName," took ", $totalTimeMilli," ms."))
    }

    return [HResult]::S_OK
}

function IsApplicable([string] $name, [string] $value)
{   
    return $true
}

function IsMergeSupported{
    param (
        [string]$name
    )
    return $false
}

function IsSupportedParam([string] $name) 
{

    #validate module of setting:
    if(!$name.StartsWith($componentPrefix))
    {
        return $false;
    }

    $settingScopeAndName = $name.Replace($componentPrefix,"") #remove module prefix
    $settingScopeStr = $settingScopeAndName.Substring(0,$settingScopeAndName.IndexOf($delimiter))   
    $settingScope = GetSettingScope($settingScopeStr)

    $restOfNameStr = $settingScopeAndName.Replace($settingScopeStr+".","")

    switch($settingScope)
    {
        GlobalSetting 
        { 
            if(validateSettingCleanName($restOfNameStr))
            {
                return ($settingsNamesMapping.$restOfNameStr -in $GlobalSettings)
            }

            return $false;
        }
        {$_ -in $script:profileScopeValues} 
        { 
            if(validateSettingCleanName($restOfNameStr))
            {
                return ($settingsNamesMapping.$restOfNameStr -in $ProfileSettings)
            }

            return $false;
        }
        RuleSetting 
        { 
            return validateRuleSettingName($restOfNameStr);
        }
        default 
        { return $false; }
    }
}

function CompletionActions() #function assumes valid input (called after IsSupportedParam)
{

    LogInfo "FW: closing loose ends." 
    # TBD: activate log when verbosity level for scripts is tested and default is changed to info.
    # LogDebug (-join("FW: rulesToSetSplatHash:",(Out-String -InputObject $script:rulesToSetSplatHash)))

    $startTimeMilli = GetCurrentTimeInMillis

    # remove rules that were deleted or that a single setting in them was removed (we cannot un-configure a single setting.
    RemoveRulesIfRequested

    LogInfo (-join("FW: rules removal took ", ((GetCurrentTimeInMillis) - $startTimeMilli) ," ms."))

    # marking rules that were requested and differ in desired and current state.
    # this happens after removal because if a single rule setting was removed, which caused the rule to be removed,
    # we need to detect it and mark it as changed, for the rule to be re-created and the other settings to be re-aplied
    if($rulesToSet.keys.Count -gt 0)
    {
        markRulesThatChanged
    }

    # go over rules that settings were applied to them and change them or create them (if they were deleted in the prev loop).
    # Only rules that have been marked as changed will be Set/Created. Where there was no change in current configuration, we will not re-apply.
    $rulesApplicationStartTimeMilli = GetCurrentTimeInMillis
    LogInfo (-join("FW: setting ",($rulesToSet.keys).Count," rules."))

    AddRulesIfRequested

    LogInfo (-join("FW: Rules Application took ",((GetCurrentTimeInMillis) - $rulesApplicationStartTimeMilli)," ms"))

    SaveCurrentStatus
}

 ####################################################### utility functions: ####################################################### 

 # function removed all the rules that need to be removed (can be due to a complete policy removal,
 # a single setting value removal, or a whole rule removal).
 function RemoveRulesIfRequested()
{
    if($rulesToRmove.Count -gt 0)
    {
        LogInfo (-join("FW: removing ",$rulesToRmove.Count," rules."))

        $rulesToRmove.keys.Clone() | ForEach-Object{

            $removedRuleName = $_
            LogInfo (-join("CompletionActions removing rule ",$removedRuleName))
            try 
            {
                Remove-NetFirewallRule -Name $removedRuleName
                #check if removal succeeded:
                $removedRule = Get-NetFirewallRule -Name $removedRuleName #validate removal (later used in VerifyRemoval)
                if($removedRule -eq $null)
                {
                    #removal ok - save the error of VerifyRemoval
                    $rulesToRmove[$removedRuleName]['errCode'] = [HResult]::S_OK.value__
                }
            } 
            catch 
            {
                #exception on Remove-NetFirewallRule - save the error for VerifyRemoval
                $rulesToRmove[$removedRuleName]['errCode'] = [HResult]::E_UNEXPECTED.value__
                $rulesToRmove[$removedRuleName]['errStr'] = $_.Exception.Message

                LogWarning (-join("Firewall::CompletionActions error while removing rule (",$removedRuleName,"). HResult:",$_.Exception.Message," error message:",$_.Exception.Message))
                continue #move on to the next rule
            }
        }
    }
}

# function creates/sets a single rule with all its values (using cmdlets with a splat hash map)
function AddSingleRule([string]$ruleName)
{
    $singleSplatHash = $script:rulesToSetSplatHash.$ruleName

    # TBD: activate log when verbosity level for scripts is tested and default is changed to info.
    #LogDebug (-join("singleSplatHash:",(Out-String -InputObject $singleSplatHash)))

    $existingRule = $script:fwRulesCurrentStatus | Where {$_.Name -eq $ruleName}
     
    if($existingRule)
    {
        #if the rule already exists, don't try to redefine the Name value, it will fail the Set.
        $singleSplatHash.Remove('Name')
        $existingRule = Set-NetFirewallRule -ErrorAction Stop @singleSplatHash
    }
        
    if($null -eq $existingRule)
    {
        # rule did not exist before, we need to create it from scratch.
        New-NetFirewallRule -ErrorAction Stop @singleSplatHash | Out-Null
    }
}

# function sets error codes and strings for each setting, based on rule creation/set result (in $rulesToSetErrors struct).
# this is done to allow reporting for each setting.
function SetSettingsResult([Int64]$ruleErrorCode, [string] $ruleErrorString)
{
    foreach($setting in $rulesToSet.$ruleName.keys)
    {
             
        #$loggingName = $setting.Substring($setting.LastIndexOf($delimiter) + 1)
        $loggingName = GetSuffixName($setting.Split($delimiter))

        if($ruleErrorCode -ne [HResult]::S_OK.value__) 
        {
            #rule is not valid - report same rule creation error for setting
            $settingErrorCode = $ruleErrorCode
            $settingErrorMessage = $ruleErrorString

            if(!$rulesToSetErrors.ContainsKey($setting))
            {
                $ruleStatus = @{
                    'errCode' = $settingErrorCode
                    'errStr' = $settingErrorMessage
                }
                $rulesToSetErrors.Add($setting,$ruleStatus)
            } else {
                LogInfo (-join("FW cannot save error code for ",$loggingName,", it already exists with result:",$rulesToSetErrors[$setting]['errCode']))
            }
        } 
        else
        {
            $settingErrorCode = [HResult]::S_OK.value__
            $settingErrorMessage = ""
        }
    }
}

# function sets (Set/New) each rules that is both requested (based on $rulesToSet) AND marked as changed (based on $script:changedRules, 
# meaning, requested values for it differs from the rule's current state on the machine).
function AddRulesIfRequested()
{
    foreach($ruleName in $rulesToSet.keys)
    {
        $ruleErrorCode = [HResult]::S_OK.value__
        $ruleErrorString = ""
        $singleRulestartTimeMilli = GetCurrentTimeInMillis

        try 
        {
            if($ruleName -in $script:changedRules)
            {
                
                AddSingleRule $ruleName
                LogInfo (-join("Creation end:", ((GetCurrentTimeInMillis) - $singleRulestartTimeMilli) , "ms"))
            } 
            else
            {
                LogInfo (-join("FW: R ", $ruleName, "- No change"))
            }
        } 
        catch 
        {
            #exception on New-NetFirewallRule - save the error for VerifyAssignment
            LogWarning (-join("FW: error while updating rule (",$ruleName,"). HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
            $ruleErrorCode = $_.Exception.HResult
            $ruleErrorString = $_.Exception.Message
            # go on to set each setting failure (error code is per setting)
        }

        #rule was recreated/exists  - now set the rule's settings result (to allow full report on each one)
        SetSettingsResult $ruleErrorCode $ruleErrorString
        
        LogInfo (-join("FW: Single Rule Application (",$ruleName,") took ",((GetCurrentTimeInMillis) - $singleRulestartTimeMilli)," ms"))
    }
}

# function sets current values in script structs (only if there was a change requested in the relevant values).
# those structs are later used for verifying assignment: 
# (1) $script:fwRulesCurrentStatus - for fw rules that were requested (based on $rulesToSet)
# (2) $script:fwSettingsCurrentStatus - for 'global' fw settings
# (3) $script:fwProfileCurrentStatus - for each profile's settings
function SaveCurrentStatus()
{
    #save current status of the device to avoid expesinve repeating calls to cmdlets (for each setting)
    if($rulesToSet.keys.Count -gt 0) #needed only for values that were 'set'. values that were removed are validated through registry.
    {
        $script:fwRulesCurrentStatus = Get-NetFirewallRule -Name $rulesToSet.keys
    }

    if($script:anyGlobalValueSet)
    {
        $script:fwSettingsCurrentStatus = Get-NetFirewallSetting
        # TBD: activate log when verbosity level for scripts is tested and default is changed to info.
        #LogDebug (-join("FW: fwSettingsCurrentStatus:",(Out-String -InputObject $script:fwSettingsCurrentStatus)))
    }

    #foreach($profile in  @('Private', 'Domain', 'Public'))
    foreach($settingScope in $script:profileScopeValues)
    {
        if($settingScope -in $script:anyProfileValueSet)
        {
            $script:fwProfileCurrentStatus.Add($settingScope, (Get-NetFirewallProfile -Profile (GetProfileNameFromScope($settingScope))))
            # TBD: activate log when verbosity level for scripts is tested and default is changed to info.
            #LogDebug (-join("FW: fwProfileCurrentStatus[",$settingScope,"]=",(Out-String -InputObject $script:fwProfileCurrentStatus[$settingScope])))
        }
    }
}

# function adds values to array $script:changedRules. A rule name is added if any of it's requested settings'value is different than
# the machines' current state (for this rule).
# Notes: 
# (1) function calls Get-NetFirewallRule one time for all the rules that were requested (uses $rulesToSet struct).
# (2) function sets the result of Get-NetFirewallRule in script struct $script:fwRulesCurrentStatus
function markRulesThatChanged()
{
    $script:fwRulesCurrentStatus = Get-NetFirewallRule -Name $rulesToSet.keys

    foreach($ruleName in $rulesToSet.keys)
    {
        foreach($settingName in $rulesToSet.$ruleName.keys)
        {
            $workPath = GetRuleSettingWithoutRuleName -fullPath $settingName -ruleNameToRmove $ruleName
            $namePathArr = $workPath.Split($delimiter)
            $settingSuffixName = GetSuffixName($namePathArr)

            $currentValue, $expectedValue = GetCompareValuesForFWRuleSetting -ruleName $ruleName -settingSuffixName $settingSuffixName -value $rulesToSet.$ruleName.$settingName
            if(($currentValue -eq $null) -or ($expectedValue -eq $null) -or ($null -ne (Compare-Object $currentValue $expectedValue)))
            {
                $script:changedRules += $ruleName
                LogInfo "FW: R ($ruleName) marked for change ($settingSuffixName)"
                break #once a single difference detected - rule will need an update. no need to check other settings.
            }
        }     
    }
}

 # function returns a pair of values:
 # currentValue - that is active on the device, based on the cmdlets used to retrieve the status, at the end of CompletionActions()
 # expectedValue - the value recived in the policy, converted to a format comparable with currentValue.
 function GetCompareValuesForFWRuleSetting([string] $ruleName, [string] $settingSuffixName, [string] $value)
{
    #at this point rule exists for sure (otherwise we return before)
    $rule = $script:fwRulesCurrentStatus | Where {$_.Name -eq $ruleName}

    # TBD: activate log when verbosity level for scripts is tested and default is changed to info.
    #LogDebug (-join("GetCompareValues rule:$ruleName result:",(Out-String -InputObject $rule)))

    $currentValue = $rule.($settingsNamesMapping.$settingSuffixName)
    $expectedValue = $value

    switch ($settingsNamesMapping.$settingSuffixName)
    {
        #rule settings:
        $settingsNamesMapping.'Action.Type'
        {
            $expectedValue = $ActionTypeValuesMapping.$value 
            break;
        }
        $settingsNamesMapping.Profiles
        {
            try {
                $expectedValue = ConvertRuleProfileToCmdletFormat $value
            } catch {
                $expectedValue = $value #will fail in comparison because conversion failed 
            }
            break;
        }
        {$_ -in $RuleFilterSettings} { 
            $currentValue = GetCurrentValueFromRuleFilter -settingNameFWSuffix $_ -ruleName $ruleName
            if($settingsNamesMapping.$SettingSuffixName -eq $settingsNamesMapping.InterfaceTypes)
            {
                $expectedValue = ConvertInterfaceTypeValue($expectedValue)
            }
            
            if ($settingsNamesMapping.$SettingSuffixName -in $RuleAddressFilterSettings)
            {
                $expectedValue = @() #resetting $expectedValue as array before adding values to it
                $valuesArray = $value.split(",")
                foreach($address in $valuesArray)
                {
                    $splitted = $address.split("/")
                    if(($null -ne $splitted[1]) -and (Confirm-IPv4 $splitted[0])) #Check if value is in "IPv4\cidr format"
                    {
                        $address = Resolve-IPRange $address
                    }
                    $expectedValue += $address
                }
            }

                if($settingsNamesMapping.$SettingSuffixName -in $RulePortRanges)
                {
                    $expectedValue = $value.split(",")
                }

                if ($settingsNamesMapping.$SettingSuffixName -eq $settingsNamesMapping.Protocol) {
                    $currentValue = ConvertProtocolTypeValue($currentValue)
                }
                
                break;
        }
        {$_ -in $SimpleComparisonValues} {break;}
    }

    return $currentValue, $expectedValue
}

function GetCompareValuesForFWGeneralSetting([string] $settingSuffixName, [settingScope] $settingScope, [string] $value) # [string] $scopeValue, 
{
    # TBD: enable log when debug verbosity level is tested
    #LogDebug (-join("GetCompareValues general/profile. setting:", $settingSuffixName," settingScope:",$settingScope," expected:" ,$value))

    if($settingScope -in $script:profileScopeValues) {
        $mpPref = $script:fwProfileCurrentStatus[$settingScope]
    } else { 
        $mpPref = $script:fwSettingsCurrentStatus
    }

    $currentValue = $mpPref.($settingsNamesMapping.$settingSuffixName)
    $expectedValue = $value

    switch ($settingsNamesMapping.$settingSuffixName)
    {
        {$_ -in $ReversedGPOBooleanValues} #must be before GPObooleanValues, becuase current value must be toggled.
        {
            $currentValue = ToggleGPOBoolean(ConvertStrValuesToGPOBool($mpPref.$_))
            $expectedValue = ConvertStrValuesToGPOBool($value)
            break; 
        }
        $settingsNamesMapping.EnablePacketQueue
        {
            $expectedValue = $PacketQueuingValuesMapping.$expectedValue
            break;
        }
        $settingsNamesMapping.DefaultInboundAction
        {
            $expectedValue = $DefaultActionValuesMapping.$expectedValue
            break;
        }
        $settingsNamesMapping.DefaultOutboundAction
        {
            $expectedValue = $DefaultActionValuesMapping.$expectedValue
            break;
        }
        {$_ -in $GPObooleanValues}
        {
            $currentValue = ConvertStrValuesToGPOBool($mpPref.$_)
            $expectedValue = ConvertStrValuesToGPOBool($value)
            break;
        }
        $settingsNamesMapping.SaIdleTime 
        {
            $currentValue = $mpPref.$_ 
            $expectedValue = ConvertValueToGPOBoolOrUInt($value)
            break;
        }
        {$_ -in $SimpleComparisonValues} {break;}
    }

    return $currentValue, $expectedValue
}


 function GetRuleFilterFuncBySetting([string] $settingNameFWSuffix)
{ 
    switch($settingNameFWSuffix) {
        {$_ -in $RuleApplicationFilterSettings} { return (Get-Item "Function:\Get-NetFirewallApplicationFilter") }
        $settingsNamesMapping.'App.ServiceName' { return (Get-Item "Function:\Get-NetFirewallServiceFilter") }
        {$_ -in $RulePortFilterSettings} { return (Get-Item "Function:\Get-NetFirewallPortFilter") }
        $settingsNamesMapping.InterfaceTypes { (Get-Item "Function:\Get-NetFirewallInterfaceTypeFilter") }
        $settingsNamesMapping.LocalUserAuthorizedList { return (Get-Item "Function:\Get-NetFirewallSecurityFilter")}
        {$_ -in $RuleAddressFilterSettings} { return (Get-Item "Function:\Get-NetFirewallAddressFilter") }
        default { return $null }
    }
}

function GetCurrentValueFromRuleFilter([string] $settingNameFWSuffix,[string] $ruleName)
{
    $getFilterFunc = GetRuleFilterFuncBySetting($settingNameFWSuffix)

    if($getFilterFunc -eq $null) 
    { 
        LogInfo (-join("FW: cannot get Get Rule Filter Func for ",$settingNameFWSuffix))
        return $null
    }

    $filterItem = $script:fwRulesCurrentStatus | Where {$_.Name -eq $ruleName} | & $getFilterFunc #Get-NetFirewallRule -Name $ruleName
    if($filterItem -eq $null) 
    { 
        LogInfo (-join("FW: failed to get filter object for rule (",$ruleName,"), settingNameSuffix:",$settingNameSuffix))
        return $null
    }

    return $filterItem.($settingNameFWSuffix)

}

function ConvertInterfaceTypeValue([string] $value)
{
    #in powershell 'All' and 'Lan' are not valid as 'Any' and 'Wired'
    return ($value.Replace("All","Any")).Replace("Lan","Wired")
}

function ConvertProtocolTypeValue([string] $value)
{
    if($value -in $ProtocolNamesMapping.Keys)
    {
        return $ProtocolNamesMapping.$value
    }
    return $value
}

 function SetValue([string] $workPath, [string] $value, [settingScope] $settingScope, [string] $ruleName, [ref] $errorCode, [ref] $errorStr)
{
    $namePathArr = $workPath.Split($delimiter)
    
    #get relevant cmdlet:
    $setFunc = GetFirewallCmd $settingScope Set
        
    if($null -eq $setFunc) 
    {
        $errorStr.Value = -join("FW: cannot get function to set value ",$value)
        LogInfo $errorStr
        $errorCode.Value = [HResult]::E_UNEXPECTED.value__
        return 
    }

    $SettingSuffixName = GetSuffixName($namePathArr)
    $errorCode.Value = [HResult]::S_OK.value__
    try 
    {
        switch ($settingsNamesMapping.$SettingSuffixName)
        {
            #global scope settings:
            $settingsNamesMapping.DisableStatefulFtp { & $setFunc -EnableStatefulFtp (ToggleGPOBoolean($value)) -ErrorAction Stop; break }
            $settingsNamesMapping.SaIdleTime 
            {
                $convertedValue = ConvertValueToGPOBoolOrUInt($value)
                & $setFunc -MaxSAIdleTimeSeconds  $convertedValue -ErrorAction Stop
                break
            }
            $settingsNamesMapping.PresharedKeyEncoding #uses NotConfigured as string
            {
                & $setFunc -KeyEncoding $value -ErrorAction Stop
                break
            }
            $settingsNamesMapping.RequireFullAuthSupport {& $setFunc -RequireFullAuthSupport $value -ErrorAction Stop; break}   
            $settingsNamesMapping.Exemptions {& $setFunc -Exemptions $value -ErrorAction Stop; break}
            $settingsNamesMapping.CertValidationLevel {& $setFunc -CertValidationLevel $value -ErrorAction Stop; break}
            $settingsNamesMapping.EnablePacketQueue {& $setFunc -EnablePacketQueuing $PacketQueuingValuesMapping.$value -ErrorAction Stop; break}       
            #profile scope settings:
            $settingsNamesMapping.DisableUnicastResponsesToMulticastBroadcast {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -AllowUnicastResponseToMulticast  (ToggleGPOBoolean($value)) -ErrorAction Stop; break }
            $settingsNamesMapping.DefaultInboundAction {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -DefaultInboundAction  $DefaultActionValuesMapping.$value -ErrorAction Stop; break } 
            $settingsNamesMapping.DefaultOutboundAction {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -DefaultOutboundAction  $DefaultActionValuesMapping.$value -ErrorAction Stop; break }
            $settingsNamesMapping.AllowLocalPolicyMerge {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -AllowLocalFirewallRules $value -ErrorAction Stop; break }
            $settingsNamesMapping.DisableInboundNotifications {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -NotifyOnListen (ToggleGPOBoolean($value)) -ErrorAction Stop; break }
            $settingsNamesMapping.AuthAppsAllowUserPrefMerge  {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -AllowUserApps $value -ErrorAction Stop; break }
            $settingsNamesMapping.GlobalPortsAllowUserPrefMerge {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -AllowUserPorts $value -ErrorAction Stop; break }
            $settingsNamesMapping.AllowLocalIpsecPolicyMerge {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -AllowLocalIPsecRules $value -ErrorAction Stop; break }
            $settingsNamesMapping.DisableStealthModeIpsecSecuredPacketExemption {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -EnableStealthModeForIPsec  (ToggleGPOBoolean($value)) -ErrorAction Stop; break }
            $settingsNamesMapping.Shielded {& $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -AllowInboundRules (ToggleGPOBoolean($value)) -ErrorAction Stop; break }
            #rule scope settings:
            $settingsNamesMapping.Description {& $setFunc -Name $ruleName -Description $value -ErrorAction Stop; break}
            $settingsNamesMapping.Direction {& $setFunc -Name $ruleName -Direction $value -ErrorAction Stop; break }
            $settingsNamesMapping.'Action.Type' {& $setFunc -Name $ruleName -Action $ActionTypeValuesMapping.$value -ErrorAction Stop; break }
            $settingsNamesMapping.Profiles {& $setFunc -Name $ruleName -Profile (ConvertRuleProfileToCmdletFormat $value) -ErrorAction Stop; break }
            $settingsNamesMapping.'App.PackageFamilyName' {& $setFunc -Name $ruleName -Package $value -ErrorAction Stop; break }
            $settingsNamesMapping.'App.FilePath' {& $setFunc -Name $ruleName -Program $value -ErrorAction Stop; break }
            $settingsNamesMapping.'App.ServiceName' {& $setFunc -Name $ruleName -Service $value -ErrorAction Stop; break }
            $settingsNamesMapping.Protocol {& $setFunc -Name $ruleName -Protocol $value -ErrorAction Stop; break }
            $settingsNamesMapping.InterfaceTypes {
                $validatedValue = ConvertInterfaceTypeValue($value) #some values allowed by the UI are not supported by FW, and must be converted for PS support
                & $setFunc -Name $ruleName -InterfaceType $validatedValue -ErrorAction Stop; 
                break 
            }
            $settingsNamesMapping.LocalUserAuthorizedList {& $setFunc -Name $ruleName -LocalUser $value -ErrorAction Stop; break }
            $settingsNamesMapping.LocalAddressRanges {$valueArr = $value -split(","); & $setFunc -Name $ruleName -LocalAddress $valueArr -ErrorAction Stop; break }
            $settingsNamesMapping.RemoteAddressRanges {$valueArr = $value -split(","); & $setFunc -Name $ruleName -RemoteAddress $valueArr -ErrorAction Stop; break }
            $settingsNamesMapping.LocalPortRanges {$valueArr = $value -split(","); & $setFunc -Name $ruleName -LocalPort $valueArr -ErrorAction Stop; break }
            $settingsNamesMapping.RemotePortRanges {$valueArr = $value -split(","); & $setFunc -Name $ruleName -RemotePort $valueArr -ErrorAction Stop; break }
            $settingsNamesMapping.EdgeTraversal {& $setFunc -Name $ruleName -EdgeTraversalPolicy $value -ErrorAction Stop; break }
            $settingsNamesMapping.'App.Fqbn' {& $setFunc -Name $ruleName -PolicyStore $value -ErrorAction Stop; break }
            #same FW name settings (horrible)
            $settingsNamesMapping.Enabled {
                $value = ConvertStrValuesToGPOBool $value 
                if($settingScope -eq [settingScope]::RuleSetting) #rule setting
                {
                    & $setFunc -Name $ruleName -Enabled $value -ErrorAction Stop
                }    
                else #profile setting
                {
                    & $setFunc -Profile (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])) -Enabled $value -ErrorAction Stop
                }
                break
            }
            Default 
            {
                $errorStr.Value = -join("FW set value of input key is not implemented: ", $SettingSuffixName)
                $errorCode.Value = [HResult]::E_NOINTERFACE.value__
                if($SettingSuffixName -ne $settingsNamesMapping.Name)
                {
                    LogWarning $errorStr
                }
            }
        }
    }  
    catch 
    { 
        $errorCode.Value = $_.Exception.HResult
        $errorStr.Value = $_.Exception.Message
        if($null -eq $ruleName) {$ruleNameMsg = "-"} else {$ruleNameMsg = $ruleName}
        LogWarning (-join("FW::SetValue fail for ",$workPath,", rule:",$ruleNameMsg,", profile:", (GetProfileNameFromProfileStr($namePathArr[[settingNameSegment]::Scope])), 
        ", value:",$value, ", func: ", $setFunc, ". HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
        if($_.Exception.Message -match "Invalid class")
        {
            LogDebug (-join("Exception details: ", $_ | Format-List * -Force | Out-String))
            LogDebug (-join("Invocation info: ", $_.InvocationInfo | Format-List * -Force | Out-String))
            LogDebug (-join("Inner exception: ", $_.InnerException | Format-List * -Force | Out-String))
        }
    }
}

 function SaveSettingForNewRule([string]$ruleName, [string]$fullPath, [string]$value)  #function assumes valid input (called after IsSupportedParam & Set)
 {
    if(!$rulesToSet.ContainsKey($ruleName))
    {
        LogInfo (-join("SaveSettingForNewRule added (",$ruleName,")"))
        $rulesToSet.Add($ruleName,@{})
        $script:rulesToSetSplatHash.Add($ruleName,@{'DisplayName' = $ruleName})
    }

    #add setting to database 
    $rulesToSet.$ruleName.Add($fullPath,$value)

    $settingNameAsArr = (GetRuleSettingWithoutRuleName -fullPath $name -ruleNameToRmove $ruleName).Split($delimiter) 
    $settingSuffixName = GetSuffixName($settingNameAsArr)
    $script:rulesToSetSplatHash.$ruleName.Add($settingsNamesMapping.$settingSuffixName,(SettingValueToSplatValue $settingSuffixName $value))
 }

 # function converts value recieved in policy to a value that can be passed to a cmdlet
function SettingValueToSplatValue([string]$settingSuffixName, [string]$value)
{
    switch ($settingsNamesMapping.$settingSuffixName)
    {
        #rule scope settings:
        $settingsNamesMapping.'Action.Type' {return $ActionTypeValuesMapping.$value }
        $settingsNamesMapping.Profiles {return (ConvertRuleProfileToCmdletFormat $value) }
        #some values allowed by the UI are not supported by FW, and must be converted for PS support
        $settingsNamesMapping.InterfaceTypes {return (ConvertInterfaceTypeValue $value) }
        $settingsNamesMapping.LocalAddressRanges {$valueArr = $value -split(","); return $valueArr }
        $settingsNamesMapping.RemoteAddressRanges {$valueArr = $value -split(","); return $valueArr }
        $settingsNamesMapping.LocalPortRanges {$valueArr = $value -split(","); return $valueArr  }
        $settingsNamesMapping.RemotePortRanges {$valueArr = $value -split(","); return $valueArr}
        default {return $value}
    }
}

 function GetRuleSettingWithoutRuleName([string] $fullPath, [string] $ruleNameToRmove) #function assumes valid input (called after IsSupportedParam)
 {
     return $fullPath.Replace("."+$ruleNameToRmove,"")
 }

 function GetSettingRuleName([string] $fullPath) #function assumes valid input (called after IsSupportedParam)
 {
    $settingScopeAndName = $fullPath.Replace($componentPrefix,"") #remove module prefix
    $settingScopeStr = $settingScopeAndName.Substring(0,$settingScopeAndName.IndexOf($delimiter))   
    $restOfNameStr = $settingScopeAndName.Replace($settingScopeStr+".","") #remove scope

    #find the suffix name of the setting
    $settingFound = $settingsNamesMapping.Keys | where { 
        $restOfNameStr.EndsWith("."+$_) 
     }
    
    #remove name suffix
    $nameIndex = $restOfNameStr.LastIndexOf("."+$settingFound)
    $ruleName = $restOfNameStr.Remove($nameIndex)

    return $ruleName
 }

 function SaveFWRuleForRemoval([string] $ruleName) #function assumes valid input (called after IsSupportedParam)
 {

    if(!$rulesToRmove.ContainsKey($ruleName))
    {
        LogInfo (-join("SaveFWRuleForRemoval added (",$ruleName,")"))
        #New rule status - value of removal result is set to false, after removal validation value will be changed to success. Error string is empty.
        $ruleStatus = @{
            errCode = [HResult]::E_FAIL.value__
            errStr = ""
        }  

        $rulesToRmove.Add($ruleName,$ruleStatus) 
    }
 }

 function GetSettingScopeFromFullPath([string] $fullPath) #function assumes valid input (called after IsSupportedParam)
{
    $settingScopeAndName = $fullPath.Replace($componentPrefix,"") #remove module prefix
    $settingScopeStr = $settingScopeAndName.Substring(0,$settingScopeAndName.IndexOf($delimiter))
    return GetSettingScope($settingScopeStr)
}
 
 function validateSettingCleanName([string] $settingCleanName)
 {
    return $settingsNamesMapping.ContainsKey( $settingCleanName ) 
 }


 function validateRuleSettingName([string] $settingRuleAndSettingName)
 {
    $settingFound = $settingsNamesMapping.Keys | where { 
        $settingRuleAndSettingName.EndsWith("."+$_) 
     }
    if($settingFound -ne $null)
    {
        return ($settingsNamesMapping.$settingFound -in $RuleSettings)
    }

    return $false;
 }

 function GetSuffixName([string[]] $namePathArr)
 {
    # setting name IS valid, so it has to be either last cell (for a "regular" setting name)
    # or the last 2 cells (for a setting name with a dot, like Action.Type).
    $optionalSuffix = $namePathArr[$namePathArr.count - 1]
    if($null -eq $settingsNamesMapping.$optionalSuffix)
    {
        $optionalSuffix = ($namePathArr[$namePathArr.count - 2] + "." + $optionalSuffix)

    }
    
    if($null -eq $settingsNamesMapping.$optionalSuffix)
    {
        LogWarning (-join("FW: GetSuffixName invalid:", ($namePathArr -join "|")))
    }

    return $optionalSuffix 
 }

 function GetProfileNameFromProfileStr([string]$scopeName) #assumes possible values: PrivateProfile, DomainProfile, PublicProfile
 {
    return $scopeName.Replace("Profile","") 
 }

 function GetProfileNameFromScope([settingScope]$scope) 
 {
    switch($scope)
    {
        ([settingScope]::PrivateProfileSetting) {return 'Private'}
        ([settingScope]::PublicProfileSetting) {return 'Public'}
        ([settingScope]::DomainProfileSetting) {return 'Domain'}
    }
 }


 function ToggleGPOBoolean([Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean] $value)
 {
    switch ($value) {
        ([Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::True) { return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::False }
        ([Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::False) { return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::True }
        default { return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::NotConfigured}
    }
 }

function GetFirewallCmd([settingScope]$scope, [settingAction]$action)
{
    switch($scope)
        {
            GlobalSetting
            { 
                if($action -eq [settingAction]::Get) { 
                    return (Get-Item "Function:\Get-NetFirewallSetting")  
                } elseif($action -eq [settingAction]::Set) { 
                    return (Get-Item "Function:\Set-NetFirewallSetting")  
                }
            }
            {$_ -in $script:profileScopeValues}
            { 
                if($action -eq [settingAction]::Get) { 
                    return (Get-Item "Function:\Get-NetFirewallProfile")  
                } elseif($action -eq [settingAction]::Set) {
                    return (Get-Item "Function:\Set-NetFirewallProfile") 
                } 
            }
            RuleSetting
            { 
                if($action -eq [settingAction]::Get) {
                    return (Get-Item "Function:\Get-NetFirewallRule")  
                } elseif ($action -eq [settingAction]::Set) {
                    return (Get-Item "Function:\Set-NetFirewallRule") 
                } 
            }
            Default {
                return $null
            }
        }
}

function ConvertToGPOBooleanNumericValue([string] $value)
{
    #GpoBoolean has an additional value besides true/false (Notconfigured)
    if($value -is [int])
    {
        return $value
    }

    return ([Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::$value).value__
}

function ConvertStringToGPOBoolean([string] $strValue)
{
    #GpoBoolean has an additional value besides true/false (Notconfigured)
    return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::$strValue
}

function ConvertValueToGPOBoolOrUInt([string] $value)
{
    if($value -ieq "NotConfigured")  
    {
        return ConvertToGPOBooleanNumericValue($value)
    } else {
        return [System.Convert]::ToUInt32($value)
    }
}

function GetSettingScope([string] $scopeStr)
{
    switch ($scopeStr) {
        "Global" { return [settingScope]::GlobalSetting }
        "PrivateProfile" { return  [settingScope]::PrivateProfileSetting }
        "PublicProfile" { return  [settingScope]::PublicProfileSetting }
        "DomainProfile" { return  [settingScope]::DomainProfileSetting }
        "FirewallRules" { return  [settingScope]::RuleSetting }
        # a valid setting name (validated in IsSupportedParam) no need for default
   }
}

function ConvertStrValuesToGPOBool([string] $value)
{
    switch ($value) 
    {
        "0" { return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::False }
        "1" { return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::True }
        "False" { return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::False }
        "True" { return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::True }
        "Notconfigured" { return [Microsoft.PowerShell.Cmdletization.GeneratedTypes.NetSecurity.GpoBoolean]::NotConfigured }
    }
}

function Resolve-IPRange([string] $value)
{
    $split = $value.split("/")
    [ipaddress]$baseAddress = $split[0]
    $cidr = $split[1]
    if($cidr -eq 0) #if bit musk is 0 FW doesn't apply it
    {
        return $split[0]
    }
    #calculate subnet using xor between (2^32 - 1) and (2^(32-cidr) - 1) giving a binary with (32-cidr) leading bits ON
    $subnet = [ipaddress]([math]::pow(2, 32) -1 -bxor [math]::pow(2, (32 - $cidr))-1)
    #calculate start address using bitwise and between original address and subnet
    [ipaddress]$start = $baseAddress.Address -band $subnet.Address
    return "$start/$subnet"
}

function Confirm-IPv4([string] $value)
{
    #Validate that value is in IPv4 format. Not checking if value is legal!
    return ($value -match '([0-9]{1,3}\.){3}[0-9]{1,3}')
}

function IsValidRuleProfileValue([string] $profileValueStr)
{
    #convert value to uint32 (might throw, will catch this on Set).
    try
    {
        $profileValue = [System.Convert]::ToUInt32($profileValueStr)
    } 
    catch 
    {
        return $false
    }

    #If the input has other values that 'Any' or a bitmast of the other supported types - then it is invalid.
    # for details, see Icm https://portal.microsofticm.com/imp/v3/incidents/details/353114611
    if($profileValue -ne $script:RuleProfileTypeAny)
    {
        if((($profileValue -band (-bnot $script:RuleProfileTypesSupportedCombinations)) -ne 0))
        {
            #other bits where active in the requested value, that do not match any of the supported values
            return $false
        }
    }

    return $true
}


function GetCmdletRuleProfileValueAsStr([uint32] $profileValue)
{
    #get profileValue types that are included in the bitmask:
    if($profileValue -eq $script:RuleProfileTypeAny)
    {
        # if 'Any' is the bitmask, don't care about other types, just use 'Any'
        return "Any"
    }

    $convertedValue = ""
    foreach($profileTypeStr in $script:RuleProfileTypesInfoArr.Keys)
    {
        # other profile types can be combined in a bitmask, need to create the cmdlet string with all those types:
        $profileBitwiseValue = $script:RuleProfileTypesInfoArr.$profileTypeStr
        if(($profileValue -band $profileBitwiseValue) -eq $profileBitwiseValue) #bitmask contains this profile type
        {
            #add profile type to string 
            if(![string]::IsNullOrEmpty($convertedValue))
            {
                $convertedValue += ","
            }
            $convertedValue += $profileTypeStr 
        }
    }

    return $convertedValue
}

# see Icm https://portal.microsofticm.com/imp/v3/incidents/details/353114611 for further explanation of rule profile types.
# see also documentation: https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-fasp/7704e238-174d-4a5e-b809-5f3787dd8acc
function ConvertRuleProfileToCmdletFormat([string] $profileValueStr)
{
    # validate input. if value is not valid, throw here.
    if(!(IsValidRuleProfileValue $profileValueStr))
    {
        Throw [System.ArgumentException] "Invalid Argument:$profileValueStr"
    }

    # get uint format (we get a string that represents a uint from MEM. example: "5").
    # will not throw at this point, because validated in prev step
    $profileValue = [System.Convert]::ToUInt32($profileValueStr)

    # get cmdlet string format from uint (comma separated string. example: "Domain,Public")
    $convertedStr = GetCmdletRuleProfileValueAsStr $profileValueStr

    LogInfo (-join("FW: Convert R Profile: ", $profileValueStr, "->", $convertedStr)) 
    return $convertedStr
}

#RetrieveTestInfo - returns module info for the TestRunner. Returning here any variable and struct needed for running the tests for this module
function RetrieveTestInfo() {
    return $settingsNamesMapping, $RuleFilterSettings, $script:profileScopeValues, $componentPrefix, $regKey
}
# SIG # Begin signature block
# MIIntAYJKoZIhvcNAQcCoIInpTCCJ6ECAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCB0S78kpkTKxZKR
# nS0/pnP+p9aCpmH+mPyE8Qmrm4uV06CCDZcwggYVMIID/aADAgECAhMzAAADcaWg
# nFyRHuruAAAAAANxMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMwNTExMTg0MTEzWhcNMjQwNTA4MTg0MTEzWjCBlDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjE+MDwGA1UEAxM1TWlj
# cm9zb2Z0IFdpbmRvd3MgRGVmZW5kZXIgQWR2YW5jZWQgVGhyZWF0IFByb3RlY3Rp
# b24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDQl+fOHFGkNfDHrcv3
# g8shIsUp58WmSbtNm4uu7Sw75Wgsgo5p2yYGGqiRWbu0eIKhAJ1bRp9Qa7vAjMEw
# d0wf9wlLjDNAITzDsokwoCl0rOgWxzYuGSusZq4e/liClIpY5fuxrpsEnsssR1o6
# jITkdJQ6ZkWsTFETxLLOEGxe9BkTkXZaBOHlhTt5QIq6UB6j8zNN/Vo4rFmq751X
# lQLA1jOnNi5NOS3/11TW16ZCo9XKG22zdNsJCUckT/QH2eeQNgpuyfIvP06o+RpP
# Uo7P4xdc4c74S3HKfRiNuqbs3+xvb22GtXIz3cKFDhILpNKN/SC8qLjizq0PxwL1
# lr1ZAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBggrBgEFBQcDAwYKKwYBBAGCN0wv
# ATAdBgNVHQ4EFgQUcR5wDY+Mh5A/I7scm54A8ODeSVcwRQYDVR0RBD4wPKQ6MDgx
# HjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEWMBQGA1UEBRMNNDUxODk0
# KzUwMTAwNTAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8E
# TTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9N
# aWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBR
# BggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0
# cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAw
# DQYJKoZIhvcNAQELBQADggIBAKYhzPfo4pWE2MYg2W1VnHJ7NZ/nwEpGYEz7Qhqc
# bPj2sJFFhnUzYHOAyXz4vP0+ERhFw9lrqLvoyj8s6xQU11QPQVItopahH65USGWh
# UcyrXFBVSx7MG3Dwi+NT6rxq9MjY3KLB1hZPF7hgtOjWgtjXgeJOsy3IjPiRkDwX
# 5b5Pg6o8Xpyt84mpvAPsDKO7yJGJq8V528deIYh+0DcELvUZTtl4Kz8ElT4BxF6i
# XxXkuqpCWO/rwEa6HrLNV2udND/AImVio09bbZXOWBcb9uTTNezYX5fZvinIX7hl
# UHHIDaVUqqwTFtac8q7BLE02T0P7NAeBtayujIQFzGNuwLN302UmUWAyc4LynqZa
# h2OmZWcMODUCuHIMP3pk54XjWt0YYEWPFAbA7T7j8wjby/vPXdG93LybID/yjGDb
# U1t5xMoEpN5rvSUM1U42iJ1tF268/nE2/8nKqUpdJXH/MaXxYU3FnbshmIe1xfxF
# NjeMYN4jg+5gjz5SBzHAhT1X9QWXNOhAXxib3wcchlsxFYtSNihHLEJ2GuckRuPi
# 47VWi/3hTTnl88S8r3jsbWLatGbQxcbrsF4sglAvFD9SQSXLGUsE5wz/1p/Gb1Fh
# qfNBc6E4AWVj9+A1KezSnMBAw8gJk3fiRkdxfSfpLOkHvgA7qsOmQ7qwjs05kWZf
# 055CMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWlj
# cm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4
# MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3Y
# bqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUB
# FDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnbo
# MlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT
# +OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuy
# e4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEh
# NSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2
# z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3
# s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78Ic
# V9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E
# 11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5P
# M4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcV
# AQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAf
# BgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBL
# hklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggr
# BgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsG
# AQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwA
# ZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0G
# CSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDB
# ZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc
# 8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYq
# wooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu
# 5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWI
# UUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXh
# j38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yH
# PgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtI
# EJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA9Z74v2u3S5fi63V4Guzq
# N5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgR
# MiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQ
# zTGCGXMwghlvAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
# dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9y
# YXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEC
# EzMAAANxpaCcXJEe6u4AAAAAA3EwDQYJYIZIAWUDBAIBBQCgga4wGQYJKoZIhvcN
# AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUw
# LwYJKoZIhvcNAQkEMSIEIEfnd5H4DAEBxaKTkYkweCkGd3wG0pO838aRzkNlIMis
# MEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8AcwBvAGYAdKEagBhodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEASArsjjlF/rg+
# +hxrFv7OfwwfKwQdoMOI0TunqsF9zfn7oWAfN1NnI2VTxYyD1BiOw5LsbHlQZJCD
# OYqfQdMo+IhLset5/xVaXaBGiZuy2vX5qaxdeDDpSI658kaU/vZoBTwmken0M7ZT
# PON6+UdLfH6a5cKdbYrbdcrWruDYoYxJUghH5/Z33PSBkKOESUlNTmlWCE6MVUkf
# 8kR/z1VOwITEYoFqB+vx4frtNngxugtJcrc04VD3n+Y77B25UhrYOd0qe8o3n6fN
# So0hWOGolfldSQ4cALthPy2Y07LcKsN/a+JvcRuQ2NuesrcKZzefeIsVyHsWvqTx
# dJHxQxyef6GCFv0wghb5BgorBgEEAYI3AwMBMYIW6TCCFuUGCSqGSIb3DQEHAqCC
# FtYwghbSAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFRBgsqhkiG9w0BCRABBKCCAUAE
# ggE8MIIBOAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCAmvrEnX6La
# 3Q0+qFf21BQZhEZ5Ki2VnmaHrfkRajGG3wIGZItRBCxvGBMyMDIzMDcxMTE3NDY1
# OS45MDZaMASAAgH0oIHQpIHNMIHKMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBPcGVyYXRpb25z
# MSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjpFNUE2LUUyN0MtNTkyRTElMCMGA1UE
# AxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaCCEVQwggcMMIIE9KADAgEC
# AhMzAAABvvQgou6W1iDWAAEAAAG+MA0GCSqGSIb3DQEBCwUAMHwxCzAJBgNVBAYT
# AlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYD
# VQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBU
# aW1lLVN0YW1wIFBDQSAyMDEwMB4XDTIyMTEwNDE5MDEyMloXDTI0MDIwMjE5MDEy
# MlowgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
# EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAjBgNV
# BAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJjAkBgNVBAsTHVRoYWxl
# cyBUU1MgRVNOOkU1QTYtRTI3Qy01OTJFMSUwIwYDVQQDExxNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBTZXJ2aWNlMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA
# pV/y2z7Da7nMu0tykLY8olh7Z03EqFNz3iFlMp9gOfVmZABmheCc87RuPdQ2P+OH
# UJiqCQAWNuNSoI/Q1ixEw9AA657ldD8Z3/EktpmxKHHavOhwQSFPcpTGFVXIxKCw
# oO824giyHPG84dfhdi6WU7f7D+85LaPB0dOsPHKKMGlC9p66Lv9yQzvAhZGFmFhl
# usCy/egrz6JX/OHOT9qCwughrL0IPf47ULe1pQSEEihy438JwS+rZU4AVyvQczlM
# C26XsTuDPgEQKlzx9ru7EvNV99l/KCU9bbFf5SnkN1moUgKUq09NWlKxzLGEvhke
# 2QNMopn86Jh1fl/PVevN/xrZSpV23rM4lB7lh7XSsCPeFslTYojKN2ioOC6p3By7
# kEmvZCh6rAsPKsARJISdzKQCMq+mqDuiP6mr/LvuWKinP+2ZGmK/C1/skvlTjtIe
# hu50yoXNDlh1CN9B3QLglQY+UCBEqJog/BxAn3pWdR01o/66XIacgXI/d0wG2/x0
# OtbjEGAkacfQlmw0bDc02dhQFki/1Q9Vbwh4kC7VgAiJA8bC5zEIYWHNU7C+He69
# B4/2dZpRjgd5pEpHbF9OYiAf7s5MnYEnHN/5o/bGO0ajAb7VI4f9av62sC6xvhKT
# B5R4lhxEMWF0z4v7BQ5CHyMNkL+oTnzJLqnLVdXnuM0CAwEAAaOCATYwggEyMB0G
# A1UdDgQWBBTrKiAWoYRBoPGtbwvbhhX6a2+iqjAfBgNVHSMEGDAWgBSfpxVdAF5i
# XYP05dJlpxtTNRnpcjBfBgNVHR8EWDBWMFSgUqBQhk5odHRwOi8vd3d3Lm1pY3Jv
# c29mdC5jb20vcGtpb3BzL2NybC9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENB
# JTIwMjAxMCgxKS5jcmwwbAYIKwYBBQUHAQEEYDBeMFwGCCsGAQUFBzAChlBodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY3Jvc29mdCUyMFRp
# bWUtU3RhbXAlMjBQQ0ElMjAyMDEwKDEpLmNydDAMBgNVHRMBAf8EAjAAMBMGA1Ud
# JQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBCwUAA4ICAQDHlfu9c0ImhdBis1yj
# 56bBvOSyGpC/rSSty+1F49Tf6fmFEeqxhwTTHVHOeIRNd8gcDLSz0d79mXCqq8yn
# q6gJgy2u4LyyAr2LmwzFVuuxoGVR8YuUnRtvsDH5J+unye/nMkwHiC+G82h3uQ8f
# cGj+2H0nKPmUpUtfQruMUXvzLjV5NyRjDiCL5c/f5ecmz01dnpnCvE6kIz/FTpkv
# OeVJk22I2akFZhPz24D6OT6KkTtwBRpSEHDYqCQ4cZ+7SXx7jzzd7b+0p9vDboqC
# y7SwWgKpGQG+wVbKrTm4hKkZDzcdAEgYqehXz78G00mYILiDTyUikwQpoZ7am9pA
# 6BdTPY+o1v6CRzcneIOnJYanHWz0R+KER/ZRFtLCyBMvLzSHEn0sR0+0kLklncKj
# GdA1YA42zOb611UeIGytZ9VhNwn4ws5GJ6n6PJmMPO+yPEkOy2f8OBiuhaqlipiW
# hzGtt5UsC0geG0sW9qwa4QAW1sQWIrhSl24MOOVwNl/Am9/ZqvLRWr1x4nupeR8G
# 7+DNyn4MTg28yFZRU1ktSvyBMUSvN2K99BO6p1gSx/wvSsR45dG33PDG5fKqHOgD
# xctjBU5bX49eJqjNL7S/UndLF7S0OWL9mdk/jPVHP2I6XtN0K4VjdRwvIgr3jNib
# 3GZyGJnORp/ZMbY2Dv1mKcx7dTCCB3EwggVZoAMCAQICEzMAAAAVxedrngKbSZkA
# AAAAABUwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRl
# IEF1dGhvcml0eSAyMDEwMB4XDTIxMDkzMDE4MjIyNVoXDTMwMDkzMDE4MzIyNVow
# fDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMd
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwggIiMA0GCSqGSIb3DQEBAQUA
# A4ICDwAwggIKAoICAQDk4aZM57RyIQt5osvXJHm9DtWC0/3unAcH0qlsTnXIyjVX
# 9gF/bErg4r25PhdgM/9cT8dm95VTcVrifkpa/rg2Z4VGIwy1jRPPdzLAEBjoYH1q
# UoNEt6aORmsHFPPFdvWGUNzBRMhxXFExN6AKOG6N7dcP2CZTfDlhAnrEqv1yaa8d
# q6z2Nr41JmTamDu6GnszrYBbfowQHJ1S/rboYiXcag/PXfT+jlPP1uyFVk3v3byN
# pOORj7I5LFGc6XBpDco2LXCOMcg1KL3jtIckw+DJj361VI/c+gVVmG1oO5pGve2k
# rnopN6zL64NF50ZuyjLVwIYwXE8s4mKyzbnijYjklqwBSru+cakXW2dg3viSkR4d
# Pf0gz3N9QZpGdc3EXzTdEonW/aUgfX782Z5F37ZyL9t9X4C626p+Nuw2TPYrbqgS
# Uei/BQOj0XOmTTd0lBw0gg/wEPK3Rxjtp+iZfD9M269ewvPV2HM9Q07BMzlMjgK8
# QmguEOqEUUbi0b1qGFphAXPKZ6Je1yh2AuIzGHLXpyDwwvoSCtdjbwzJNmSLW6Cm
# gyFdXzB0kZSU2LlQ+QuJYfM2BjUYhEfb3BvR/bLUHMVr9lxSUV0S2yW6r1AFemzF
# ER1y7435UsSFF5PAPBXbGjfHCBUYP3irRbb1Hode2o+eFnJpxq57t7c+auIurQID
# AQABo4IB3TCCAdkwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEEAYI3FQIEFgQU
# KqdS/mTEmr6CkTxGNSnPEP8vBO4wHQYDVR0OBBYEFJ+nFV0AXmJdg/Tl0mWnG1M1
# GelyMFwGA1UdIARVMFMwUQYMKwYBBAGCN0yDfQEBMEEwPwYIKwYBBQUHAgEWM2h0
# dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvRG9jcy9SZXBvc2l0b3J5Lmh0
# bTATBgNVHSUEDDAKBggrBgEFBQcDCDAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMA
# QTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBTV9lbL
# j+iiXGJo0T2UkFvXzpoYxDBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1p
# Y3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXRfMjAxMC0w
# Ni0yMy5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIz
# LmNydDANBgkqhkiG9w0BAQsFAAOCAgEAnVV9/Cqt4SwfZwExJFvhnnJL/Klv6lwU
# tj5OR2R4sQaTlz0xM7U518JxNj/aZGx80HU5bbsPMeTCj/ts0aGUGCLu6WZnOlNN
# 3Zi6th542DYunKmCVgADsAW+iehp4LoJ7nvfam++Kctu2D9IdQHZGN5tggz1bSNU
# 5HhTdSRXud2f8449xvNo32X2pFaq95W2KFUn0CS9QKC/GbYSEhFdPSfgQJY4rPf5
# KYnDvBewVIVCs/wMnosZiefwC2qBwoEZQhlSdYo2wh3DYXMuLGt7bj8sCXgU6ZGy
# qVvfSaN0DLzskYDSPeZKPmY7T7uG+jIa2Zb0j/aRAfbOxnT99kxybxCrdTDFNLB6
# 2FD+CljdQDzHVG2dY3RILLFORy3BFARxv2T5JL5zbcqOCb2zAVdJVGTZc9d/HltE
# AY5aGZFrDZ+kKNxnGSgkujhLmm77IVRrakURR6nxt67I6IleT53S0Ex2tVdUCbFp
# AUR+fKFhbHP+CrvsQWY9af3LwUFJfn6Tvsv4O+S3Fb+0zj6lMVGEvL8CwYKiexcd
# FYmNcP7ntdAoGokLjzbaukz5m/8K6TT4JDVnK+ANuOaMmdbhIurwJ0I9JZTmdHRb
# atGePu1+oDEzfbzL6Xu/OHBE0ZDxyKs6ijoIYn/ZcGNTTY3ugm2lBRDBcQZqELQd
# VTNYs6FwZvKhggLLMIICNAIBATCB+KGB0KSBzTCByjELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMcTWljcm9zb2Z0IEFtZXJpY2Eg
# T3BlcmF0aW9uczEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046RTVBNi1FMjdDLTU5
# MkUxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2WiIwoBATAH
# BgUrDgMCGgMVAGitWlL3vPu8ENOAe+i2+4wfTMB7oIGDMIGApH4wfDELMAkGA1UE
# BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAc
# BgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0
# IFRpbWUtU3RhbXAgUENBIDIwMTAwDQYJKoZIhvcNAQEFBQACBQDoV2xYMCIYDzIw
# MjMwNzExMTM1MTIwWhgPMjAyMzA3MTIxMzUxMjBaMHQwOgYKKwYBBAGEWQoEATEs
# MCowCgIFAOhXbFgCAQAwBwIBAAICGq0wBwIBAAICEVYwCgIFAOhYvdgCAQAwNgYK
# KwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAqAKMAgCAQACAwehIKEKMAgCAQAC
# AwGGoDANBgkqhkiG9w0BAQUFAAOBgQBJrPjs5lr1EOB7Td9PT0T7JREOpH/aH715
# EtFBmzffPQmpgE+xfXru+DSpcFrKYtQavLXukW032vBhdwc/11NBqLnj3YdmDQDi
# /7pGBYkLgFBo5f03xwpOUnRrCBJHpnMzb4J/TE5oJv5s0QW18EifXxMl0e2IlP/c
# 7uByOUPjcjGCBA0wggQJAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAy
# MDEwAhMzAAABvvQgou6W1iDWAAEAAAG+MA0GCWCGSAFlAwQCAQUAoIIBSjAaBgkq
# hkiG9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEINzBcyV3dJS/
# N/0iKf0mec8I7VKfTyklJzVWTjdmmwc2MIH6BgsqhkiG9w0BCRACLzGB6jCB5zCB
# 5DCBvQQglO6Kr632/Oy9ZbXPrhEPidNAg/Fef7K3SZg+DxjhDC8wgZgwgYCkfjB8
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1N
# aWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAb70IKLultYg1gABAAAB
# vjAiBCA68emIaGVihT7leXqMdfTFaG9YBXC9y8FhTbDUbjLDITANBgkqhkiG9w0B
# AQsFAASCAgBhJdHR30rhludWc54zpwBkJtYSI3dIe41y5kUu1A0i7vpCyLRZCjGB
# v9i1WtPDJE8rjaD9bzdRqeU+jXRvbhIcUo9c9TdzGDmIO3QaVTY2uK9QwEk/kyIv
# lq9RhwJWSOpEySpgcaoapPkTVmJUiTjMIfdJDbW209zXLqjwO5qNOoRYxY78bEXj
# ibO4RpuQnhXHjzRfvx/ZQ+vmoC0buy2kYEf1IQZLFsG6/TStYtNhAvULC4JrvunM
# dPC8lZm+EBqPBpcZrVGPtgpO0THrJzcqYKKtXMP7UZk5MxIw5aPgfwrOr4bNAUb6
# cA6wQe+7Js/SgtWcMV95CuNIlLlqQMhpJfXZn8aEWqUmNJ0IpL9cubOB2Hi3XHkK
# 2vDxUTnirld0TXmaYqod8XIhK32Y7EZ+Sw1JNDfqjThNNns0ePxpzfx50EFEBkOR
# TQd+bOZ+70XyD68gzIE4iC5UYp+QqhQzAIDglFYFXqCwAi4H/n3Egm6YWgvlNVX2
# XQ9XxpIF//+FO0zq+gZ1J9tcbUjcjV5btlNDykvK6tEbKsF1x0Ky4+tQRnqZf9aY
# LbCVZguLYfVQrXlE2CmWt6ggmjEU8vuAdFFR5qtwjqQCBU9SgoE+MNWl4CPZVZ4e
# WVnWPzwNjs5UXdO7J4XRGQuatwFk/SiZ+3b03q5G+tcH9lmY0va5oQ==
# SIG # End signature block
