# Group Policy Object Utils module:
#
# File used to allow Group Policy updates on machine.
# Usage - 
# (1) use $module = Import-Module for GroupPolicyObject.psm1
# (2) call $module.GenerateGroupPolicyObject() to compile the c# source code
# (3) call $gpoObj = 'New-Object GPO.ComputerGroupPolicyObject' to create an object
# (4) call $gpoObj.SetPolicySetting to make changes to the GPO
# (5) call $gpoObj.Save() to save the changes done. 
# After the save, RegistryExtension change will trigger the actual change accoring to the snap-in model, by another process.

$script:sleepIntervalmilli = 200
$script:maxSleepTimeMilli = 20000

function GenerateGroupPolicyObject()
{ 

$GroupPolicyObjectCode = @"
namespace GPO
{
using System;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using Microsoft.Win32;
using Microsoft.Win32.SafeHandles;

[ComImport]
[GuidAttribute("EA502722-A23D-11D1-A7D3-0000F87571E3")] 
internal class GPClass {}

[ComImport]
[GuidAttribute("EA502723-A23D-11D1-A7D3-0000F87571E3")] //GUID that identifies the MMC snap-in used to edit this policy (IGroupPolicyObject)
[InterfaceType(ComInterfaceType.InterfaceIsIUnknown)] 
public interface IGroupPolicyObject
{
    uint New(
        [MarshalAs(UnmanagedType.LPWStr)] string domainName,
        [MarshalAs(UnmanagedType.LPWStr)] string displayName, uint flags);

    uint OpenDSGPO([MarshalAs(UnmanagedType.LPWStr)] string path, uint flags);

    uint OpenLocalMachineGPO(uint flags);

    uint OpenRemoteMachineGPO([MarshalAs(UnmanagedType.LPWStr)] string computerName, uint flags);

    uint Save(
        [MarshalAs(UnmanagedType.Bool)] bool machine, [MarshalAs(UnmanagedType.Bool)] bool add,
        [MarshalAs(UnmanagedType.LPStruct)] Guid extension, [MarshalAs(UnmanagedType.LPStruct)] Guid app);

    uint Delete();

    uint GetName([MarshalAs(UnmanagedType.LPWStr)] StringBuilder name, int maxLength);

    uint GetDisplayName([MarshalAs(UnmanagedType.LPWStr)] StringBuilder name, int maxLength);

    uint SetDisplayName([MarshalAs(UnmanagedType.LPWStr)] string name);

    uint GetPath([MarshalAs(UnmanagedType.LPWStr)] StringBuilder path, int maxPath);

    uint GetDSPath(uint section, [MarshalAs(UnmanagedType.LPWStr)] StringBuilder path, int maxPath);

    uint GetFileSysPath(uint section, [MarshalAs(UnmanagedType.LPWStr)] StringBuilder path, int maxPath);

    uint GetRegistryKey(uint section, out IntPtr key);

    uint GetOptions(out uint options);

    uint SetOptions(uint options, uint mask);

    uint GetType(out IntPtr gpoType);

    uint GetMachineName([MarshalAs(UnmanagedType.LPWStr)] StringBuilder name, int maxLength);

    uint GetPropertySheetPages(out IntPtr pages);
}

public enum GroupPolicySection
{
    Root = 0,
    User = 1,
    Machine = 2,
}

public abstract class GroupPolicyObject
{
    protected const int MaxLength = 1024;

    /// <summary>
    /// The snap-in that processes .pol files (GUID of the registry snap-in extension that will process the policy)
    // "35378EAC-683F-11D2-A89A-00C04FBBCFA2"
    /// </summary>
    private static readonly Guid RegistryExtension = new Guid(0x35378EAC,0x683F,0x11D2,0xA8,0x9A,0x00,0xC0,0x4F,0xBB,0xCF,0xA2);  

    /// <summary>
    /// This application's GUID {FD7F7960-3815-49DC-BEDE-C67CC936F7A4} 
    /// </summary>
    private static readonly Guid LocalGuid = new Guid(0xfd7f7960, 0x3815, 0x49dc, 0xbe, 0xde, 0xc6, 0x7c, 0xc9, 0x36, 0xf7, 0xa4); 

    protected IGroupPolicyObject Instance = null;

    internal GroupPolicyObject()
    {
        Instance = GetInstance();
    }

    static public string GetGPOGuid() 
    {
        return RegistryExtension.ToString();
    }

    static public string GetLocalGuid() 
    {
        return LocalGuid.ToString();
    }

    public void Save()
    {
        Console.WriteLine("INFO: GPO Saving policy for local machine ({0},{1})",RegistryExtension.ToString(),LocalGuid.ToString());

        var result = Instance.Save(true, true, RegistryExtension, LocalGuid);
        if (result != 0)
        {
            throw new Exception("Error saving machine settings");
        }
    }

    public void Delete()
    {
        var result = Instance.Delete();
        if (result != 0)
        {
            throw new Exception("Error deleting the GPO");
        }

        Instance = null;
    }

    public RegistryKey GetRootRegistryKey(GroupPolicySection section)
    {
        IntPtr key;
        var result = Instance.GetRegistryKey((uint)section, out key);
        if (result != 0)
        {
            throw new Exception(
                string.Format(
                    "Unable to get section '{0}'",
                    Enum.GetName(typeof(GroupPolicySection), section)));
        }

        var handle = new SafeRegistryHandle(key, true);
        return RegistryKey.FromHandle(handle, RegistryView.Default);
    }

    protected static IGroupPolicyObject GetInstance()
    {
        var concrete = new GPClass();
        return (IGroupPolicyObject)concrete;
    }
}

public class GroupPolicyObjectSettings
{
    public readonly bool LoadRegistryInformation;
    public readonly bool Readonly;

    public GroupPolicyObjectSettings(bool loadRegistryInfo = true, bool readOnly = false)
    {
        LoadRegistryInformation = loadRegistryInfo;
        Readonly = readOnly;
    }

    private const uint RegistryFlag = 0x00000001;
    private const uint ReadonlyFlag = 0x00000002;

    internal uint Flag
    {
        get
        {
            uint flag = 0x00000000;
            if (LoadRegistryInformation)
            {
                flag |= RegistryFlag;
            }

            if (Readonly)
            {
                flag |= ReadonlyFlag;
            }

            return flag;
        }
    }
}

public class ComputerGroupPolicyObject : GroupPolicyObject
{
    public ComputerGroupPolicyObject()
    {
        var options = new GroupPolicyObjectSettings();
        var result = Instance.OpenLocalMachineGPO(options.Flag);
        if (result != 0)
        {
            throw new Exception("Unable to open local machine GPO");
        }
    }

    public void SetPolicySetting(string registryInformation, string settingValue, RegistryValueKind registryValueKind) 
    {
        string valueName;
        GroupPolicySection section;
        string key = Key(registryInformation, out valueName, out section);
        
        Console.WriteLine("INFO: GPO SetPolicySetting to valueName:{0}, settingValue:{1}",valueName,settingValue);
        Exception exception = null;
        try
        {
            using (RegistryKey rootRegistryKey = GetRootRegistryKey(section))
            {
                // Data can't be null so we can use this value to indicate key must be delete
                if (String.IsNullOrEmpty(settingValue))
                {
                    using (RegistryKey subKey = rootRegistryKey.OpenSubKey(key, true))
                    {
                        if (subKey != null)
                        {
                            Console.WriteLine("INFO: GPO SetPolicySetting deleting value:{0}",valueName);
                            subKey.DeleteValue(valueName);
                        }
                    }
                }
                else
                {
                    Console.WriteLine("INFO: GPO SetPolicySetting setting value:{0}, valueName:{1}",settingValue,valueName);
                    using (RegistryKey subKey = rootRegistryKey.CreateSubKey(key))
                    {
                        subKey.SetValue(valueName, settingValue, registryValueKind);
                    }
                }
            }
        }
        catch (Exception ex)
        {
            exception = ex;
            throw exception;
        }            
    }

    public object GetPolicySetting(string registryInformation) 
    {
        string valueName;
        GroupPolicySection section;
        string key = Key(registryInformation, out valueName, out section);

        object result = null;
        using (RegistryKey rootRegistryKey = GetRootRegistryKey(section))
        {
            using (RegistryKey subKey = rootRegistryKey.OpenSubKey(key, true))
            {
                if (subKey == null)
                {
                    result = null;
                }
                else
                {
                    result = subKey.GetValue(valueName);
                    subKey.Close();
                }
            }
        }

        return result;
    }

    private string Key(string registryInformation, out string value, out GroupPolicySection section) 
    {
        // Parse parameter of format HKCU\Software\Policies\Microsoft\Windows\Personalization!NoChangingSoundScheme
        string[] split = registryInformation.Split('!');
        string key = split[0];
        string hive = key.Substring(0, key.IndexOf('\\'));
        key = key.Substring(key.IndexOf('\\') + 1);

        value = split[1];

        if (hive.Equals(@"HKLM", StringComparison.OrdinalIgnoreCase)
            || hive.Equals(@"HKLM:", StringComparison.OrdinalIgnoreCase)
            || hive.Equals(@"HKEY_LOCAL_MACHINE", StringComparison.OrdinalIgnoreCase)
            || hive.Equals(@"HKEY_LOCAL_MACHINE:", StringComparison.OrdinalIgnoreCase))
        {
            section = GroupPolicySection.Machine;
        }
        else
        {
            section = GroupPolicySection.User;
        }

        return key;
    }
}
}
"@
    Add-Type -TypeDefinition $GroupPolicyObjectCode
    LogInfo (-join("GPO added type, max sleep time:",$script:maxSleepTimeMilli))
}

#function used to debug issues in group policy creation. will be removed.
function DumpGPRegistryContent()
{
    # reg export dumps into a file (only), so we must get the file's content and remove it when we are done:
    $inputFileName = "GPHistoryRegistryDump.log"
    $res = (reg export "HKLM\SOFTWARE\Microsoft\Windows\CurrentVersion\Group Policy" $local:inputFileName /y)
    #have to use Write-Host here to allow larger log
    Write-Host "Group Policy registry dump:" (Get-Content $local:inputFileName -Raw)
    Remove-Item $local:inputFileName
}

# function returns Version from Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Group Policy\History\{REGISTY_EXTENTION_Guid}\X
# where 'X' is the index that has this application's Guid in the 'Extesions' registry value. If none is found, 0 is returned
function GetGPOCurrentVersion()
{
	$GPOGuid = [GPO.GroupPolicyObject]::GetGPOGuid()
	$localGuid = [GPO.GroupPolicyObject]::GetLocalGuid()
    $GPHistoryRegPath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Group Policy\History"
	$GPRegExtHistoryRegPath =  ( Join-Path $local:GPHistoryRegPath (-join("{",$local:GPOGuid,"}")) )

	# get relevant GPO index, by verifying its extension is our application Guid. There should be only one like that
	$GPOIndices = Get-ChildItem $local:GPRegExtHistoryRegPath -Name | Where-Object { $GPOFullPath = -join($local:GPRegExtHistoryRegPath,"\",$_); $GPOExtensions = Get-ItemProperty -Path $GPOFullPath -Name "Extensions"; ($GPOExtensions -match $local:localGuid) }
    
    if($GPOIndices.Count -eq 0)
    {
        #LogInfo "GPO no indices of GPO found, setting version to 0"
        $GPOLatestVersion = 0
    } 
    else 
    {
    	$GPOLatestVersion = (Get-ItemProperty -Path (-join($local:GPRegExtHistoryRegPath,"\",$local:GPOIndices[0])) -Name "Version").Version
        #LogInfo (-join("GPO indices of GPO found, getting version from index 0:",$GPOLatestVersion))
    }
    
    return $GPOLatestVersion
}

# returns true if new version was created. otherwise - false.
# sets currently detected version in newGPOVersion reference.
# uses WasGPOChangeApplied, see relevant documentation for more details.
function WaitForGroupPolicyUpdate([int32]$prevVersion,[ref]$newGPOVersion)
{
    $milliSlept = 0

    do {
        Start-Sleep -Milliseconds $script:sleepIntervalmilli 
        $local:milliSlept += $script:sleepIntervalmilli
    } until ((WasGPOChangeApplied $local:prevVersion $newGPOVersion) -or ($local:milliSlept -ge $script:maxSleepTimeMilli))

    LogInfo (-join("GPO waited $local:milliSlept ms (",($newGPOVersion.Value -ne $local:prevVersion),",old:",$local:prevVersion,",new:",$newGPOVersion.Value,")"))
    return ($newGPOVersion.Value -ne $local:prevVersion)
}

# function returns true if the GPO save took affect by checking if the Version changed 
# in Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Group Policy\History\{REGISTY_EXTENTION_Guid}\X
# where 'X' is the index that has this application's Guid in the 'Extesions' registry value
# (in comparison to old GPO version provided in $prevVersion)
# NOTE: functions receives a reference to $newGPOVersion and fills it with the current GPO version found (or 0, if none is found).
function WasGPOChangeApplied([int32]$prevVersion,[ref]$newGPOVersion)
{
    $existingNewVersion = $newGPOVersion.Value
    #get new GPO version
    $newGPOVersion.Value = GetGPOCurrentVersion

    if($newGPOVersion.Value -ne $local:existingNewVersion)
    {
        LogInfo (-join("GPO detected different version:",$newGPOVersion.Value," (before:",$local:existingNewVersion,")"))
    }

    if($local:prevVersion -eq $newGPOVersion.Value) 
    {
        return $false
    } 

    return $true
}

# Note: this fuction sets the value in the GPO but does not save.
# To save, call function SaveAllChangesInPolicy()
function SetPolicySetting([GPO.ComputerGroupPolicyObject]$currentObj ,[string]$GPRegistryPathToKey, [string]$settingName, [string]$settingValue, [Microsoft.Win32.RegistryValueKind]$valueType)
{
    if($null -eq $local:currentObj) 
    {
        LogWarning "GPO: SetPolicySetting($settingName) currentObj is null. Cannot set."
        return
    }

    $registryInfo = -join($local:GPRegistryPathToKey,"!",$local:settingName)
    $local:currentObj.SetPolicySetting($local:registryInfo,$local:settingValue,$local:valueType) 
}

# Note: this fuction clears the value in the GPO but does not save.
# To save, call function SaveAllChangesInPolicy()
function ClearPolicySetting([GPO.ComputerGroupPolicyObject]$currentObj, [string]$GPRegistryPathToKey, [string]$settingName)
{
    if($null -eq $local:currentObj)  
    {
        LogWarning "GPO: ClearPolicySetting($local:settingName) currentObj is null. Cannot clear."
        return
    }

    if((GetFuturePolicySetting $local:currentObj $local:GPRegistryPathToKey $local:settingName) -ne $null) #if value already about to be remove don't attempt to remove again (will result in error)
    {
        SetPolicySetting -currentObj $local:currentObj -GPRegistryPathToKey $local:GPRegistryPathToKey -settingName $local:settingName -settingValue $null -valueType ([Microsoft.Win32.RegistryValueKind]::Unknown)
    }
}

# Gets the value of a setting, as currently in GPO (not necessarily saved).
# If no such value - returns $null
function GetFuturePolicySetting([GPO.ComputerGroupPolicyObject]$currentObj, [string]$GPRegistryPathToKey, [string]$settingName)
{
    if($null -eq $local:currentObj) 
    {
        LogWarning "GPO: GetFuturePolicySetting($local:settingName) currentObj is null. Cannot set."
        return
    }

    $registryInfo = -join($local:GPRegistryPathToKey,"!",$local:settingName)
    $newValue = $local:currentObj.GetPolicySetting($local:registryInfo)
    return $local:newValue
}

# Function saves changes done in the GPO, which will finalize the changes and trigger change in the GP hive as well.
# Function also sets values into referenced arguments:  
# prevGPOVer - value from Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Group Policy\History\{REGISTRY_EXTENSION_GUID} before the save,
# newGPOVer - value from Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Group Policy\History\{REGISTRY_EXTENSION_GUID} after the save & wait,
# IMPORTANT: function contains a sleep until GPO version update or 1 second 
function SaveAllChangesInPolicy([GPO.ComputerGroupPolicyObject]$currentObj,[ref]$prevGPOVer,[ref]$newGPOVer)
{
    if($local:currentObj -eq $null) 
    {
        LogWarning "GPO: SaveAllChangesInPolicy currentObj is null. Cannot save."
        return
    }

    #save current GPO version before save, for update validation
    $prevGPOVer.Value = GetGPOCurrentVersion 
    LogInfo (-join("GPO version before save:",$prevGPOVer.Value))

    $local:currentObj.Save() 
    if((WaitForGroupPolicyUpdate $prevGPOVer.Value $newGPOVer))
    {
         LogInfo ("GPO saved and updated")
    }
    else
    {
         LogInfo ("GPO saved but not updated yet")
    }
}





# SIG # Begin signature block
# MIIntwYJKoZIhvcNAQcCoIInqDCCJ6QCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCAjqEsNd3iWzmSj
# JRqxtUvmMuAOiE10YEcvPv/O5DJDraCCDZcwggYVMIID/aADAgECAhMzAAADcaWg
# nFyRHuruAAAAAANxMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMwNTExMTg0MTEzWhcNMjQwNTA4MTg0MTEzWjCBlDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjE+MDwGA1UEAxM1TWlj
# cm9zb2Z0IFdpbmRvd3MgRGVmZW5kZXIgQWR2YW5jZWQgVGhyZWF0IFByb3RlY3Rp
# b24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDQl+fOHFGkNfDHrcv3
# g8shIsUp58WmSbtNm4uu7Sw75Wgsgo5p2yYGGqiRWbu0eIKhAJ1bRp9Qa7vAjMEw
# d0wf9wlLjDNAITzDsokwoCl0rOgWxzYuGSusZq4e/liClIpY5fuxrpsEnsssR1o6
# jITkdJQ6ZkWsTFETxLLOEGxe9BkTkXZaBOHlhTt5QIq6UB6j8zNN/Vo4rFmq751X
# lQLA1jOnNi5NOS3/11TW16ZCo9XKG22zdNsJCUckT/QH2eeQNgpuyfIvP06o+RpP
# Uo7P4xdc4c74S3HKfRiNuqbs3+xvb22GtXIz3cKFDhILpNKN/SC8qLjizq0PxwL1
# lr1ZAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBggrBgEFBQcDAwYKKwYBBAGCN0wv
# ATAdBgNVHQ4EFgQUcR5wDY+Mh5A/I7scm54A8ODeSVcwRQYDVR0RBD4wPKQ6MDgx
# HjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEWMBQGA1UEBRMNNDUxODk0
# KzUwMTAwNTAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8E
# TTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9N
# aWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBR
# BggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0
# cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAw
# DQYJKoZIhvcNAQELBQADggIBAKYhzPfo4pWE2MYg2W1VnHJ7NZ/nwEpGYEz7Qhqc
# bPj2sJFFhnUzYHOAyXz4vP0+ERhFw9lrqLvoyj8s6xQU11QPQVItopahH65USGWh
# UcyrXFBVSx7MG3Dwi+NT6rxq9MjY3KLB1hZPF7hgtOjWgtjXgeJOsy3IjPiRkDwX
# 5b5Pg6o8Xpyt84mpvAPsDKO7yJGJq8V528deIYh+0DcELvUZTtl4Kz8ElT4BxF6i
# XxXkuqpCWO/rwEa6HrLNV2udND/AImVio09bbZXOWBcb9uTTNezYX5fZvinIX7hl
# UHHIDaVUqqwTFtac8q7BLE02T0P7NAeBtayujIQFzGNuwLN302UmUWAyc4LynqZa
# h2OmZWcMODUCuHIMP3pk54XjWt0YYEWPFAbA7T7j8wjby/vPXdG93LybID/yjGDb
# U1t5xMoEpN5rvSUM1U42iJ1tF268/nE2/8nKqUpdJXH/MaXxYU3FnbshmIe1xfxF
# NjeMYN4jg+5gjz5SBzHAhT1X9QWXNOhAXxib3wcchlsxFYtSNihHLEJ2GuckRuPi
# 47VWi/3hTTnl88S8r3jsbWLatGbQxcbrsF4sglAvFD9SQSXLGUsE5wz/1p/Gb1Fh
# qfNBc6E4AWVj9+A1KezSnMBAw8gJk3fiRkdxfSfpLOkHvgA7qsOmQ7qwjs05kWZf
# 055CMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWlj
# cm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4
# MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3Y
# bqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUB
# FDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnbo
# MlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT
# +OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuy
# e4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEh
# NSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2
# z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3
# s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78Ic
# V9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E
# 11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5P
# M4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcV
# AQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAf
# BgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBL
# hklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggr
# BgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsG
# AQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwA
# ZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0G
# CSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDB
# ZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc
# 8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYq
# wooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu
# 5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWI
# UUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXh
# j38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yH
# PgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtI
# EJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA9Z74v2u3S5fi63V4Guzq
# N5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgR
# MiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQ
# zTGCGXYwghlyAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
# dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9y
# YXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEC
# EzMAAANxpaCcXJEe6u4AAAAAA3EwDQYJYIZIAWUDBAIBBQCgga4wGQYJKoZIhvcN
# AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUw
# LwYJKoZIhvcNAQkEMSIEIPZkF04z8Y/n6Lo+CsZVBg1ZcsZ+h8ig0EDS6HE1pBxO
# MEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8AcwBvAGYAdKEagBhodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEAs3NrA/Jgkx4C
# Jf5F1hXNPOmG3X9vnhbUtbI7fEL+qAC2zum+FPqvoRfNSKenaTXlGaqTo4sH954+
# WTwGshSKBeKyHn/NmwDg/Iu18WIJTcGsn270qF2KeR8VrLLvQeJpLEqVt9LjqPlP
# a68hgCtAv5TVc5JxDLmG4SMVIXcvaTRX7X7Fzl+5PzuA7HJFFDWzcwFweZxfOyjV
# Za/VW38a6Y2I+C5XdYXTI0xvpZSlJj9qnl6SRL5G8wbrf+jafUq1a2AoFmMqRoN5
# 3QmrsayznnR2dpMKI0kZylCu8DYKU0DV1gtByvUuRwrkKyNanv/63pUgOWJuT0/4
# sI/cgQsPb6GCFwAwghb8BgorBgEEAYI3AwMBMYIW7DCCFugGCSqGSIb3DQEHAqCC
# FtkwghbVAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFRBgsqhkiG9w0BCRABBKCCAUAE
# ggE8MIIBOAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCDXPTBDnn4l
# xta1yzsGP6PlH1QtEOS9NWm3Z5aEPwnGWwIGZIt2mi5GGBMyMDIzMDcxMTE3NDcw
# MC45NTRaMASAAgH0oIHQpIHNMIHKMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBPcGVyYXRpb25z
# MSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjo4QTgyLUUzNEYtOUREQTElMCMGA1UE
# AxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaCCEVcwggcMMIIE9KADAgEC
# AhMzAAABwvp9hw5UU0ckAAEAAAHCMA0GCSqGSIb3DQEBCwUAMHwxCzAJBgNVBAYT
# AlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYD
# VQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBU
# aW1lLVN0YW1wIFBDQSAyMDEwMB4XDTIyMTEwNDE5MDEyOFoXDTI0MDIwMjE5MDEy
# OFowgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
# EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAjBgNV
# BAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJjAkBgNVBAsTHVRoYWxl
# cyBUU1MgRVNOOjhBODItRTM0Ri05RERBMSUwIwYDVQQDExxNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBTZXJ2aWNlMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA
# tfEJvPKOSFn3petp9wco29/UoJmDDyHpmmpRruRVWBF37By0nvrszScOV/K+LvHW
# WWC4S9cme4P63EmNhxTN/k2CgPnIt/sDepyACSkya4ukqc1sT2I+0Uod0xjy9K2+
# jLH8UNb9vM3yH/vCYnaJSUqgtqZUly82pgYSB6tDeZIYcQoOhTI+M1HhRxmxt8Ra
# AKZnDnXgLdkhnIYDJrRkQBpIgahtExtTuOkmVp2y8YCoFPaUhUD2JT6hPiDD7qD7
# A77PLpFzD2QFmNezT8aHHhKsVBuJMLPXZO1k14j0/k68DZGts1YBtGegXNkyvkXS
# gCCxt3Q8WF8laBXbDnhHaDLBhCOBaZQ8jqcFUx8ZJSXQ8sbvEnmWFZmgM93B9P/J
# TFTF6qBVFMDd/V0PBbRQC2TctZH4bfv+jyWvZOeFz5yltPLRxUqBjv4KHIaJgBhU
# 2ntMw4H0hpm4B7s6LLxkTsjLsajjCJI8PiKi/mPKYERdmRyvFL8/YA/PdqkIwWWg
# 2Tj5tyutGFtfVR+6GbcCVhijjy7l7otxa/wYVSX66Lo0alaThjc+uojVwH4psL+A
# 1qvbWDB9swoKla20eZubw7fzCpFe6qs++G01sst1SaA0GGmzuQCd04Ue1eH3DFRD
# ZPsN+aWvA455Qmd9ZJLGXuqnBo4BXwVxdWZNj6+b4P8CAwEAAaOCATYwggEyMB0G
# A1UdDgQWBBRGsYh76V41aUCRXE9WvD++sIfGajAfBgNVHSMEGDAWgBSfpxVdAF5i
# XYP05dJlpxtTNRnpcjBfBgNVHR8EWDBWMFSgUqBQhk5odHRwOi8vd3d3Lm1pY3Jv
# c29mdC5jb20vcGtpb3BzL2NybC9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENB
# JTIwMjAxMCgxKS5jcmwwbAYIKwYBBQUHAQEEYDBeMFwGCCsGAQUFBzAChlBodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY3Jvc29mdCUyMFRp
# bWUtU3RhbXAlMjBQQ0ElMjAyMDEwKDEpLmNydDAMBgNVHRMBAf8EAjAAMBMGA1Ud
# JQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBCwUAA4ICAQARdu3dCkcLLPfaJ3rR
# 1M7D9jWHvneffkmXvFIJtqxHGWM1oqAh+bqxpI7HZz2MeNhh1Co+E9AabOgj94Sp
# 1seXxdWISJ9lRGaAAWzA873aTB3/SjwuGqbqQuAvUzBFCO40UJ9anpavkpq/0nDq
# Lb7XI5H+nsmjFyu8yqX1PMmnb4s1fbc/F30ijaASzqJ+p5rrgYWwDoMihM5bF0Y0
# riXihwE7eTShak/EwcxRmG3h+OT+Ox8KOLuLqwFFl1siTeQCp+YSt4J1tWXapqGJ
# DlCbYr3Rz8+ryTS8CoZAU0vSHCOQcq12Th81p7QlHZv9cTRDhZg2TVyg8Gx3X6mk
# pNOXb56QUohI3Sn39WQJwjDn74J0aVYMai8mY6/WOurKMKEuSNhCiei0TK68vOY7
# sH0XEBWnRSbVefeStDo94UIUVTwd2HmBEfY8kfryp3RlA9A4FvfUvDHMaF9BtvU/
# pK6d1CdKG29V0WN3uVzfYETJoRpjLYFGq0MvK6QVMmuNxk3bCRfj1acSWee14UGj
# glxWwvyOfNJe3pxcNFOd8Hhyp9d4AlQGVLNotaFvopgPLeJwUT3dl5VaAAhMwvIF
# mqwsffQy93morrprcnv74r5g3ejC39NYpFEoy+qmzLW1jFa1aXE2Xb/KZw2yawql
# dSp0Hu4VEkjGxFNc+AztIUWwmTCCB3EwggVZoAMCAQICEzMAAAAVxedrngKbSZkA
# AAAAABUwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRl
# IEF1dGhvcml0eSAyMDEwMB4XDTIxMDkzMDE4MjIyNVoXDTMwMDkzMDE4MzIyNVow
# fDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMd
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwggIiMA0GCSqGSIb3DQEBAQUA
# A4ICDwAwggIKAoICAQDk4aZM57RyIQt5osvXJHm9DtWC0/3unAcH0qlsTnXIyjVX
# 9gF/bErg4r25PhdgM/9cT8dm95VTcVrifkpa/rg2Z4VGIwy1jRPPdzLAEBjoYH1q
# UoNEt6aORmsHFPPFdvWGUNzBRMhxXFExN6AKOG6N7dcP2CZTfDlhAnrEqv1yaa8d
# q6z2Nr41JmTamDu6GnszrYBbfowQHJ1S/rboYiXcag/PXfT+jlPP1uyFVk3v3byN
# pOORj7I5LFGc6XBpDco2LXCOMcg1KL3jtIckw+DJj361VI/c+gVVmG1oO5pGve2k
# rnopN6zL64NF50ZuyjLVwIYwXE8s4mKyzbnijYjklqwBSru+cakXW2dg3viSkR4d
# Pf0gz3N9QZpGdc3EXzTdEonW/aUgfX782Z5F37ZyL9t9X4C626p+Nuw2TPYrbqgS
# Uei/BQOj0XOmTTd0lBw0gg/wEPK3Rxjtp+iZfD9M269ewvPV2HM9Q07BMzlMjgK8
# QmguEOqEUUbi0b1qGFphAXPKZ6Je1yh2AuIzGHLXpyDwwvoSCtdjbwzJNmSLW6Cm
# gyFdXzB0kZSU2LlQ+QuJYfM2BjUYhEfb3BvR/bLUHMVr9lxSUV0S2yW6r1AFemzF
# ER1y7435UsSFF5PAPBXbGjfHCBUYP3irRbb1Hode2o+eFnJpxq57t7c+auIurQID
# AQABo4IB3TCCAdkwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEEAYI3FQIEFgQU
# KqdS/mTEmr6CkTxGNSnPEP8vBO4wHQYDVR0OBBYEFJ+nFV0AXmJdg/Tl0mWnG1M1
# GelyMFwGA1UdIARVMFMwUQYMKwYBBAGCN0yDfQEBMEEwPwYIKwYBBQUHAgEWM2h0
# dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvRG9jcy9SZXBvc2l0b3J5Lmh0
# bTATBgNVHSUEDDAKBggrBgEFBQcDCDAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMA
# QTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBTV9lbL
# j+iiXGJo0T2UkFvXzpoYxDBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1p
# Y3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXRfMjAxMC0w
# Ni0yMy5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIz
# LmNydDANBgkqhkiG9w0BAQsFAAOCAgEAnVV9/Cqt4SwfZwExJFvhnnJL/Klv6lwU
# tj5OR2R4sQaTlz0xM7U518JxNj/aZGx80HU5bbsPMeTCj/ts0aGUGCLu6WZnOlNN
# 3Zi6th542DYunKmCVgADsAW+iehp4LoJ7nvfam++Kctu2D9IdQHZGN5tggz1bSNU
# 5HhTdSRXud2f8449xvNo32X2pFaq95W2KFUn0CS9QKC/GbYSEhFdPSfgQJY4rPf5
# KYnDvBewVIVCs/wMnosZiefwC2qBwoEZQhlSdYo2wh3DYXMuLGt7bj8sCXgU6ZGy
# qVvfSaN0DLzskYDSPeZKPmY7T7uG+jIa2Zb0j/aRAfbOxnT99kxybxCrdTDFNLB6
# 2FD+CljdQDzHVG2dY3RILLFORy3BFARxv2T5JL5zbcqOCb2zAVdJVGTZc9d/HltE
# AY5aGZFrDZ+kKNxnGSgkujhLmm77IVRrakURR6nxt67I6IleT53S0Ex2tVdUCbFp
# AUR+fKFhbHP+CrvsQWY9af3LwUFJfn6Tvsv4O+S3Fb+0zj6lMVGEvL8CwYKiexcd
# FYmNcP7ntdAoGokLjzbaukz5m/8K6TT4JDVnK+ANuOaMmdbhIurwJ0I9JZTmdHRb
# atGePu1+oDEzfbzL6Xu/OHBE0ZDxyKs6ijoIYn/ZcGNTTY3ugm2lBRDBcQZqELQd
# VTNYs6FwZvKhggLOMIICNwIBATCB+KGB0KSBzTCByjELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMcTWljcm9zb2Z0IEFtZXJpY2Eg
# T3BlcmF0aW9uczEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046OEE4Mi1FMzRGLTlE
# REExJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2WiIwoBATAH
# BgUrDgMCGgMVAMp1N1VLhPMvWXEoZfmF4apZlnRUoIGDMIGApH4wfDELMAkGA1UE
# BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAc
# BgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0
# IFRpbWUtU3RhbXAgUENBIDIwMTAwDQYJKoZIhvcNAQEFBQACBQDoV5H7MCIYDzIw
# MjMwNzExMTYzMTU1WhgPMjAyMzA3MTIxNjMxNTVaMHcwPQYKKwYBBAGEWQoEATEv
# MC0wCgIFAOhXkfsCAQAwCgIBAAICAbgCAf8wBwIBAAICEecwCgIFAOhY43sCAQAw
# NgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAqAKMAgCAQACAwehIKEKMAgC
# AQACAwGGoDANBgkqhkiG9w0BAQUFAAOBgQBO4z01k6Iqe7OI0KcPkDaDfpY+WTCH
# XBQuyJdaOZMcByGhNri8Ovy0yK27mriDK7l2H9jN3D4DN7mXUEw+PIIY+8fu7iuz
# hs6a0vNtr6eIcUMsM3lUk0YUSxGouhphkYH1hyZhmWef0uOtOfVGsj7GchaWHb7R
# P8TDfdJ+B9uZ/DGCBA0wggQJAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQI
# EwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3Nv
# ZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBD
# QSAyMDEwAhMzAAABwvp9hw5UU0ckAAEAAAHCMA0GCWCGSAFlAwQCAQUAoIIBSjAa
# BgkqhkiG9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEIEGpNaEo
# xfgR+gG3eIzSCEk8TwlCLIQ+5BlzcMlbZlnvMIH6BgsqhkiG9w0BCRACLzGB6jCB
# 5zCB5DCBvQQgypNgW8fpsMV57r0F5beUuiEVOVe4BdmaO+e28mGDUBYwgZgwgYCk
# fjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
# UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQD
# Ex1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAcL6fYcOVFNHJAAB
# AAABwjAiBCAy40hidc2mgQT9XJzRRwEgNwmzhT3aFW8L0Q8xzT4uOTANBgkqhkiG
# 9w0BAQsFAASCAgCU+oN2kf/yKNeBvTRVNkDTrF/wvUsc6MozgbwAop+4FErJcv/s
# r1/KyjSD5Jy5LTv3PCbv/cDpoYV8VrENqiEShqTfqxJlE2uLnlknREJTwcckj99W
# zCfa6Jq/A33wVvgKoc/6iwV3ZR3nx8hgvroybqFRSZLWJgC9J+8XlZtqUGgN6lBL
# PBwZqx0j0N78KfjOLpAIT15J0/XPXoawarXQrZ7RhpUWMACRZQCaQ4Z4izp1KRDA
# sg1uzZOR/GlfeMkJsyizkYGHumlALIJisf/LxNLAkGA06nDkVcPRxavkAxqWghdK
# vmlo+MLNJeTz1YpV6+GRW82iaUHPrA34K4jCbOw2ZBeOwLYjBuwDjbjtWOIu0Fzo
# MIkFdpq9495Spz00++qQoPZ6jGX7CjYV3RCYF39DMSH/sX8Wyvr1BiaigDClcn90
# g5TT1kIh/DuL2Qf6AdzXSh1umMnZ61n4SBX9cXtd6XvHobmhFzf2QBaBbahMTHK4
# QzInYocLRwYOquvMS/dWtsoCG7fpM3Y1hIzJOxR3Za1H4kv9KzNIODBs6c6LbtyJ
# UtkapDpMWuyQUuZRK1PaO+DXpWc1BKPNyF9yPVZmgIXdcHDSmRDqgdjDvkXffRIP
# COJVcJ28v+1JaSjah80cK7PPt9Ta00lHh2CgeEKLQtJjCLOMcnJGYTATbQ==
# SIG # End signature block
