########################################################################
### This script is responsible for enforcing the security policies 
### persisted on the machine and to create a report file.
### The script uses a PS module for each supported security component
### that exposes:
###     Setup, Set, VerifyAssignment, ResetToDefault, VerifyRemoval, IsApplicable, IsSupportedParam, IsMergeSupported, CompletionActions cmdlets 
########################################################################

param($logLevel)
$script:versionId = 1.0021
Write-Host "PolicyEnforcer setup ($script:versionId)"

Write-Host "Running on Powershell version" $PSversionTable.PSVersion

try { # general try/catch for policyEnforcer

if(!("verbosityLevel" -as [type]))
{
Add-Type -TypeDefinition @"
   public enum verbosityLevel
   {
      verbosity_warn = 0, //log only warnings
      verbosity_info = 1, //log info and up
      verbosity_debug = 2 //log debug and up, current default verbosity level
   }
"@
}

### constants 
# verbosity level of script (default is debug level). can be set in SenseConfigurations.
$global:ScriptsVerbosityLevel = [verbosityLevel]::verbosity_debug

#Time format string for Get-Date
$global:TimeFormat = "HH\:mm\:ss\.fff"

# output file size considerations:
$global:OutputSize = 0 # holds the amount of bytes written so far by LogMessage to output file, to make sure we don't exceed max file size
$global:ErrorsOutputSize = 0 # holds the amount of bytes written so far by LogMessage to errors file, to make sure we don't exceed max file size
$global:OutputMaxSizeInBytes = 60000 # 55KB - if ETW event is over 64KB IN TOTAL, it is dropped. this size was set to consider additional event header etc and prints that are not from LogMessage.

$logLevelToStr = @{ #keep all in same length (used for calcs of log size)
    [verbosityLevel]::verbosity_debug = "DBG: "
    [verbosityLevel]::verbosity_info = "INF: "
    [verbosityLevel]::verbosity_warn = "WRN: "
}

$eventLogName = "Microsoft-Windows-SENSE/Operational"
$eventLogSource = "SenseCM"
$messageLogHeader = "SenseCM: "
$totalLogSize = 1024
$timeStampFormat = "MM-dd-yyyy:HH:mm:ss.fff"
$timeStampSize = $timeStampFormat.Length
$maxLogMsgSize = $totalLogSize - $messageLogHeader.Length - ($logLevelToStr.([verbosityLevel]::verbosity_debug)).Length - $timeStampSize - 1 #total of 1024, the 1 is for space between t.s. and content
$validationStrErrHeader = "Validation Error: cannot validate action result for setting"

#registry key to get curently installed defender version:
$script:DefenderEngineVersionPath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows Defender\Signature Updates"

#get machine OS and build version, define build variables for easy usage in modules:
$global:OSVersion = (get-itemproperty -Path "HKLM:\SOFTWARE\Microsoft\Windows NT\CurrentVersion" -Name ProductName).ProductName
$global:OSBuild = (get-itemproperty -Path "HKLM:\SOFTWARE\Microsoft\Windows NT\CurrentVersion" -Name CurrentBuild).CurrentBuild
$global:2012BuildStr = "Windows Server 2012" #use regex to catch all versions
$global:2016BuildStr = "Windows Server 2016" #use regex to catch all versions 
$global:2019BuildStr = "Windows Server 2019" #use regex to catch all versions
$global:2022BuildStr = "Windows Server 2022" #use regex to catch all versions
$global:Win10Str = "Windows 10" #use regex to catch all versions for win 10
$global:Win11Str = "Windows 11" #use regex to catch all versions for win 11

$global:ServersStrs = @(
    $global:2012BuildStr,
    $global:2016BuildStr,
    $global:2019BuildStr,
    $global:2022BuildStr
)

#Set verbosity level according to argument received on activation of script
function ParseLogLevel
{
    switch ($logLevel) {
        info { $global:ScriptsVerbosityLevel = [verbosityLevel]::verbosity_info; break }
        warn { $global:ScriptsVerbosityLevel = [verbosityLevel]::verbosity_warn; break }
        #else logLevel stays at default level
    }
    Write-Host (-join("PE: logLevel is set to ", $global:ScriptsVerbosityLevel))    
}

#Parse scripts logLevel parameter before starting. Need to add verbosityLevel enum first for logLevel parameter
ParseLogLevel

function global:LogMessage([verbosityLevel] $logLevel, [string] $message)
{
    if($logLevel -le $global:ScriptsVerbosityLevel)  #less/equal - need to log message
    {
        if($message.Length -gt $maxLogMsgSize) #add consideration of log level
        {
            $lengthValidatedMsg = $message.Remove($maxLogMsgSize)
        } 
        else 
        {
            $lengthValidatedMsg = $message
        }

        $formattedMsg = -join($logLevelToStr.$logLevel,$lengthValidatedMsg) # contains log level & actuall msg

        $newLogAdditionLength = $formattedMsg.Length
        if($logLevel -eq [verbosityLevel]::verbosity_warn)
        {
            $newLogAdditionLength = 2 * $formattedMsg.Length # will be written both in output file and in errors file.
        }

        if($global:OutputSize + $global:ErrorsOutputSize + $newLogAdditionLength -gt $global:OutputMaxSizeInBytes) # don't log if output file + erros file is too big (issued in same ETW event)
        {
            return
        }

        $global:OutputSize += $formattedMsg.Length # aggregate output bytes logged so far to output file .
        Write-Host $formattedMsg
        
        if($logLevel -eq [verbosityLevel]::verbosity_warn)
        {
            $eventLogMsg = -join($messageLogHeader,$formattedMsg)  #contains header & log level & avctual msg
            $global:ErrorsOutputSize += $formattedMsg.Length   # aggregate output bytes logged to errors file so far.
            # To write to EXTERNAL stderr (directed to file then sent as a warning ETW event). 
            # Note - Error.WriteLine only works with the PowerShell console host (not shown in ISE)
            [Console]::Error.WriteLine($formattedMsg) 
            #To add to eventlog:
            # Event ID chosen so it won't match existing Sense EvenIDs (from here: https://microsoft.visualstudio.com/DefaultCollection/OS/_git/os.2020?path=%2Famcore%2Fwcd%2FSource%2FSetup%2FManifest%2FSenseClient-Events.man)
            Write-EventLog -LogName $eventLogName -Source $eventLogSource -EventID 2001 -EntryType Warning -Message $eventLogMsg -ErrorAction SilentlyContinue
        }
    } #else - don't log message
}

function global:LogDebug([string] $message)
{
    LogMessage ([verbosityLevel]::verbosity_debug) ($message)
}

function global:LogInfo([string] $message)
{
    LogMessage ([verbosityLevel]::verbosity_info) ($message)
}

function global:LogWarning([string] $message)
{
    LogMessage ([verbosityLevel]::verbosity_warn) ($message)
}

function global:GetCurrentTimeInMillis()
{
    return ([TimeSpan](Get-Date -Format $global:TimeFormat)).TotalMilliseconds
}

if(!("HResult" -as [type]))
{$HResult = @"
public enum HResult : long
{
    S_OK = 0,
    E_INTERNAL_CONFLICT = 1,
    E_UNEXPECTED = -2147418113,
    E_FAIL = -2147467259,
    E_NOINTERFACE = -2147500034,
    E_ACCESSDENIED = -2147024891,
    E_INVALIDARG = -2147024809,
    E_HANDLE = -2147024890,
    E_FAIL_NO_SOURCE = -2016407289,
    MK_E_UNAVAILABLE = -2147221021,
    E_NOTIMPL = -2147467263,
    COR_E_SYSTEM = -2146233087
}
"@

#happens here to allow ValidateParentProcess and exitWithCode to send informative errors
Add-Type -TypeDefinition $HResult 
} #closing the HResult definition. Checking if type was already added to avoid an Exception (in TestsRunner)

#exit function to make exit code pass to application runnig this script (PS exit code are context scoped):
function exitWithCode ([HResult] $exitCode)
{
    LogInfo (-join("PE: exitWithCode - ",$exitCode))
    $Host.SetShouldExit($exitCode)
    exit
}
try{
    $scriptTotalStartTimeMilli = GetCurrentTimeInMillis
}
catch
{
    Write-Host "Error initializing scriptTotalStartTimeMilli. errorCode: " $_.Exception.HResult "errorMessage: " $_.Exception.Message
    $scriptTotalStartTimeMilli = 0
}

$global:WorkingFolder = (Join-Path (Get-CimInstance Win32_OperatingSystem -ErrorAction Stop).SystemDrive "ProgramData\Microsoft\Windows Defender Advanced Threat Protection\SenseCM")

Write-Host "Working folder is set to " $global:WorkingFolder

$global:InstallationPathRegPath = "Registry::HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows Advanced Threat Protection"
$global:InstallationPathRegKey = "InstallLocation"

$global:PoliciesFileName = "policies.json"

#All supported components scripts and their release status:
$script:SupportedComponents = @{
    "Firewall" = @{Ring=1};
    "AntiVirus" = @{Ring=1};
    "EDR" = @{Ring=1};
    "WindowsSecurityExperience" = @{Ring=0};
    "ASR" = @{Ring=1};
}

#All supported utils scripts and their release status and the module object itself we import:
$script:UtilsModules = @{
    "ErrorHandlingUtils" = @{Ring=1; Module=$null};
    "ConflictResolutionUtils" = @{Ring=1; Module=$null};
}

# will be filled after we import the module, holds all settings we support
$global:AllSupportedSettings = @()

# will hold the setting per component reviewed in effective policy function 
$global:SupportedComponentSettings = @{}

## FileUtilities
function GetFileContent
{
    param(
        [Parameter(Mandatory=$true)]
        [String]$filePath
    )
    
    return Get-Content -Path $filePath
}

function WriteToFile
{
    param(
        [Parameter(AttributeValues)]
        [String]$filePath,
        [String]$content
    )

    Out-File -FilePath $filePath -InputObject $content
}

function GetPoliciesObj
{
    # todo - validate the policy signature 
    $policiesFilePath = Join-Path -Path $global:WorkingFolder -ChildPath $global:PoliciesFileName
    [String]$policiesRaw = GetFileContent($policiesFilePath)
    return ConvertFrom-Json $policiesRaw
} 

# policies as object
$global:policiesObj = GetPoliciesObj 

$source = @"
using System;
using System.Collections.Generic;
using System.Globalization;

public class PolicyIdentifier
{
    public string id { get; set; }
    public string intent { get; set; }
    public string checksum { get; set; }

    public PolicyIdentifier(string id, string intent, string checksum)
    {
        this.id = id;
        this.intent = intent;
        this.checksum = checksum;
    }
}


public class Setting
{
    public string name { get; set; }
    public string value { get; set; }
    public string intent {get; set; }
    public List<PolicyIdentifier> policies { get; set; }

    public Setting(string name, string value, string settingIntent, string policyId, string policyIntent, string policyChecksum)
    {
        this.name = name;
        this.value = value;
        this.intent = settingIntent;
        this.policies = new List<PolicyIdentifier>();
        this.policies.Add(new PolicyIdentifier(policyId, policyIntent, policyChecksum));
    }
}

public class SettingResult
{
    public string name { get; set; }
    public string value { get; set; }  

    public SettingResult(string name, string value)
    {
        this.name = name;
        this.value = value;
    }

    public override bool Equals(Object obj)
    {
       //Check for null and compare run-time types.
       if ((obj == null) || ! this.GetType().Equals(obj.GetType()))
       {
          return false;
       }
       else {
          SettingResult res = (SettingResult) obj;
          return (name == res.name) && (value == res.value);
       }
    }

    //Must override also GetHashCode 
    public override int GetHashCode()
    {
        unchecked // Overflow is fine, just wrap
        {
            int hash = 17;
            // Suitable nullity checks etc, of course :)
            hash = hash * 23 + name.GetHashCode();
            hash = hash * 23 + value.GetHashCode();


            return hash;
        }
    }
}

public class SettingFailResult : SettingResult
{
    public string errorCode { get; set; }
    public string errorMessage { get; set; }

    public SettingFailResult(string name, string value, string errorCode, string errorMessage) : base(name,value)
    {
        this.errorCode = errorCode;
        this.errorMessage = errorMessage;
    }

    //Must override also GetHashCode 
    public override int GetHashCode()
    {
        unchecked // Overflow is fine, just wrap
        {
            int hash = 17;
            // Suitable nullity checks etc, of course :)
            hash = hash * 23 + name.GetHashCode();
            hash = hash * 23 + value.GetHashCode();
            hash = hash * 23 + errorCode.GetHashCode();
            hash = hash * 23 + errorMessage.GetHashCode();
            return hash;
        }
    }
}

// The order of the values is also the 
// order of what we report if there is more
// than one correct result 

// Compre returns 0 on success, -1 on failure
public enum SettingEnforceResults : ushort
{
    Success,
    Fail,
    Conflict,
    NotApplicable,
    ValidationFailure,
    NotRecognized
}

public enum IntentType : ushort
{
    Remove,
    Apply,
    Report
}

public class PolicyReport
{
    public string policyId { get; set; }
    public string reportChecksum { get; set; } 
    public string policyChecksum { get; set; } 
    public string intentType { get; set; }
    public List<SettingResult> settings { get; set; } 

    public PolicyReport(string policyId, string intentType, string policyChecksum)
    {
        this.policyId = policyId;
        this.intentType = intentType;
        this.policyChecksum = policyChecksum;
        this.settings = new List<SettingResult>();
    }
}

public class Report
{
    public List<PolicyReport> policyInfo { get; set; } 

    public Report()
    {
        this.policyInfo = new List<PolicyReport>();
    }

    public void ReportSettingResult(string policyId, string policyIntent, string policyChecksum, string name, SettingEnforceResults result, bool overrideIfExists = false, string errorCode = "0", string errorMessage = "")
    {
        PolicyReport report = this.policyInfo.Find(
        delegate(PolicyReport pr)
        {
            return pr.policyId == policyId;
        });

        if (report == null)
        {
            report = new PolicyReport(policyId, policyIntent, policyChecksum);
            this.policyInfo.Add(report);
        }


        SettingResult setResult = report.settings.Find(
            delegate(SettingResult res)
            {
                return res.name == name;
            }
        );

        SettingResult newSettingsResult;
        if (String.IsNullOrEmpty(errorCode) || errorCode == "0")
        {
            newSettingsResult = new SettingResult(name, result.ToString());
        }
        else
        {
            try
            {
                errorCode = errorCode.Replace("0x", ""); //remove 0x prefix if exists
                //converting errorCode to Hex
                long dec;
                bool isHex = Int64.TryParse(errorCode, NumberStyles.HexNumber, CultureInfo.InvariantCulture, out dec);
                if(!isHex)
                {
                    errorCode = Convert.ToString(long.Parse(errorCode), 16).ToUpper(); //converting to Hex
                }
                newSettingsResult = new SettingFailResult(name, result.ToString(),errorCode,errorMessage);
            }
            catch
            {
                Console.WriteLine("PE: ReportSettingResult ignoring bad error value:{0}",errorCode);
                newSettingsResult = new SettingResult(name, result.ToString());
            }
        }

        if(setResult == null)
        {
           
            report.settings.Add(newSettingsResult);
        } else if (overrideIfExists) { //setting result already in report
            report.settings.Remove(setResult); 
            report.settings.Add(newSettingsResult);

        }
    }
}

"@

function Succeeded
{
    param(
        [Parameter(Mandatory=$true)]
        [HResult]$hr
    )
    
    return $hr -ge 0;
}

# should add a checksum to each policy report (and add to the policy report json, after checksum calculation)
# required for comparison between reports
function CalcChecksum
{
    param(
        [Parameter(Mandatory=$true)]
        [String]$settingsString
    )
    
    $hasher = new-object System.Security.Cryptography.SHA256CryptoServiceProvider
    $toHash = [System.Text.Encoding]::UTF8.GetBytes($settingsString)
    $hashByteArray = $hasher.ComputeHash($toHash)
    foreach($byte in $hashByteArray)
    {
      $result += "{0:X2}" -f $byte
    }
    return $result;
}

function IsApplyOrReport([String] $intent)
{
    return ($intent -eq [IntentType]::Report.ToString()) -or ($intent -eq [IntentType]::Apply.ToString())
}

function IsSameIntent([String] $intent1, [String] $intent2)
{
    if ($intent1 -eq $intent2)
    {
        return $true
    }
    if ((IsApplyOrReport($intent1)) -and (IsApplyOrReport($intent2)))
    {
        return $true
    }
    if (($intent1 -eq [IntentType]::Remove.ToString()) -and ($intent2 -eq [IntentType]::Remove.ToString()))
    {
        return $true
    }
    return $false
}

function GetPoliciesToIgnore($policiesPayload, [String] $componentName)
{
    $policiesToIgnore = [System.Collections.Generic.HashSet[string]]@()
    $policies = @()
    # save all policies ids and intents
    foreach($policyNode in $policiesPayload.Policies)
    {
        $policyPayload = $policyNode.payload
        if ($policyPayload.component -eq $componentName)
        {
            $policies += New-Object PolicyIdentifier -ArgumentList $policyPayload.policyId, $policyPayload.intentType
        }
    }
    # now check for bugs and report conflict if found any
    foreach($policyIdentifier in $policies)
    {
        $shouldReportConflict = $false
        if ($policiesToIgnore.Contains($policyIdentifier.id))
        {
            # in this case we report conflict only for the first policy
            $shouldReportConflict = $false
        } 
        else
        {
            $hasDifferentIntent = $policies | Where-Object { $_.id -eq $policyIdentifier.id} | Where-Object { $_.intent -ne $policyIdentifier.intent}
            if ($hasDifferentIntent)
            {
                $shouldReportConflict = $true
                $policiesToIgnore.Add($policyIdentifier.id)
            } 
        }

        if ($shouldReportConflict)
        {
            $policyNode = $policiesPayload.Policies |  Where-Object { $_.payload.policyId -eq $policyIdentifier.id} | Where-Object { $_.payload.intentType -eq $policyIdentifier.intent}
            foreach($setting in $policyNode.payload.settings)
            {
                $global:report.ReportSettingResult($policyIdentifier.id, $policyIdentifier.intent, $policyIdentifier.checksum, $setting.name, [SettingEnforceResults]::Conflict)
            }
        }
    }
    return @($policiesToIgnore)
}

function EffectivePolicy 
{
    param(
        [Parameter(Mandatory=$true)][String]$componentName,
        [Parameter(Mandatory=$true)][System.Object]$componentModule
    )

    $startTimeMilli = GetCurrentTimeInMillis

    $effectivePolicySettings = @()

    # first - handle bugs of the same policies with different intents
    $policiesToIgnore = GetPoliciesToIgnore $global:policiesObj $componentName

    # we save the conflicted settings names just to be able to detect future conflicts 
    $conflictedSettings = @()
    foreach ($policyNode in $global:policiesObj.policies)
    {
        $policy = $policyNode.payload
        if ($policiesToIgnore.Count -ne 0 -and $policiesToIgnore.Contains($policy.policyId))
        {
            # do nothing - there's no simple continue in PowerShell, only ugly hacks
        }
        else
        {
            foreach ($setting in $policy.settings) 
            {
                $settingIntent = $policy.intentType
                if ($setting.type -eq [IntentType]::Remove.ToString()) #setting type is remove if the setting was removed from an existing policy in MEM
                {
                    $settingIntent = $setting.type
                }
                if ($componentModule."IsSupportedParam"($setting.name))
                {
                    # this setting will not be marked as "NotRecognized"
                    $global:AllSupportedSettings += $setting.name
                    $mergeSupported = $componentModule."IsMergeSupported"($setting.name)
                    # we might already found a conflict in this setting 
                    $settingIsAlreadyConflicted = $conflictedSettings | Where-Object { $_ -eq $setting.name}
                    if ($settingIsAlreadyConflicted)
                    {
                        $global:report.ReportSettingResult($policy.policyId, $policy.intentType, $policy.policyChecksum, $setting.name, [SettingEnforceResults]::Conflict)
                    } else 
                    {
                        # no conflict detected yet - check if this setting introduces a conflict 
                        if($mergeSupported)
						{
							$alreadyExistingSetting = $effectivePolicySettings | Where-Object { (($_.name -eq $setting.name) -and ($_.value -eq $setting.value)) }    
						}
						else
						{
							$alreadyExistingSetting = $effectivePolicySettings | Where-Object { $_.name -eq $setting.name }    
						}
						
                        if ($alreadyExistingSetting)
                        {
                            if (!(IsSameIntent $settingIntent $alreadyExistingSetting.intent))
                            {
                                # note - these are different intents for different policies
                                if ($alreadyExistingSetting.intent -eq [IntentType]::Remove.ToString())
                                { 
                                    # apply and report intent wins. 
                                    # 1. need to report success for the remove and update the existing setting:
                                    #   a. delete the "remove" policies (all of the current policies)
                                    #   b. add the new policy (report\apply) and update existing setting intent
                                    foreach($removepolicy in $alreadyExistingSetting.policies)
                                    {
                                        #unless the policy intent is not remove (only the setting intent is)
                                        if ($removepolicy.intent -eq [IntentType]::Remove.ToString())
                                        {
                                        $global:report.ReportSettingResult($removepolicy.id, [IntentType]::Remove.ToString(), $removepolicy.checksum, $setting.name, [SettingEnforceResults]::Success)
                                        }    
                                    } 
                                    
                                    $alreadyExistingSetting.intent = $settingIntent
                                    $alreadyExistingSetting.value = $setting.value
                                    $alreadyExistingSetting.policies.Clear()
                                    $policyIdentifier = New-Object PolicyIdentifier -ArgumentList $policy.policyId, $policy.intentType, $policy.policyChecksum
                                    $alreadyExistingSetting.policies.Add($policyIdentifier)                    
                                } 
                                else #already existing settings intent is apply or report but the new one is remove - report success for the remove and continue
                                {
                                    #unless the policy intent is not remove (only the setting intent is)
                                    if ($policy.intentType -eq [IntentType]::Remove.ToString())
                                    {
                                    $global:report.ReportSettingResult($policy.policyId, $policy.intentType, $policy.policyChecksum, $setting.name, [SettingEnforceResults]::Success)
                                    }                                
                                }
                            }
                            elseif ((IsSameIntent $settingIntent $alreadyExistingSetting.intent) -and ($setting.value -ne $alreadyExistingSetting.value))
                            {
                                # first time a conflict is detected - 
                                # 1. should report conflict for the current setting and for the already existing one
                                # 2. remove the setting from the settings list
                                # 3. save the conflicted setting name
                                $global:report.ReportSettingResult($policy.policyId, $policy.intentType, $policy.policyChecksum, $setting.name, [SettingEnforceResults]::Conflict)
                                foreach ($policyIdentifier in $alreadyExistingSetting.policies) 
                                {
                                    $global:report.ReportSettingResult($policyIdentifier.id, $policyIdentifier.intent, $policyIdentifier.checksum, $alreadyExistingSetting.name, [SettingEnforceResults]::Conflict)
                                }
                                $conflictedSettings += $setting.name
                                $effectivePolicySettings = $effectivePolicySettings | Where-Object { $_ -ne $alreadyExistingSetting }
                                # if all effective was conflicted, the list type doesn't remain
                                if (!$effectivePolicySettings) {  $effectivePolicySettings = @() } 

                            } else
                            {
                                # no conflict - just add the policy identifier
                                $policyIdentifier = New-Object PolicyIdentifier -ArgumentList $policy.policyId, $policy.intentType, $policy.policyChecksum
                                $alreadyExistingSetting.policies.Add($policyIdentifier)
                            }
                        } else 
                        {
                            # The first time we encounter this setting - add it to the effective policy
                            # todo: validate config format to schema
                            [array]$effectivePolicySettings += New-Object Setting -ArgumentList $setting.name, $setting.value, $settingIntent, $policy.policyId, $policy.intentType, $policy.policyChecksum
                        } 
                    }    
                }
            }
        }
    }

    $endTimeMilli = GetCurrentTimeInMillis
    LogMessage ([verbosityLevel]::verbosity_info) (-join ("PE: EffectivePolicy for ", $componentName ," took ", ($endTimeMilli - $startTimeMilli)," ms."))

    return @($effectivePolicySettings)
}

function GetRing([string] $componentName)
{
    if($componentName -in $script:SupportedComponents.Keys)
    {
        return $script:SupportedComponents[$componentName].Ring
    }
    elseif ($componentName -in $script:UtilsModules.Keys) {
        return $script:UtilsModules[$componentName].Ring
    }   

    LogWarning (-join ("PE: Component ", $componentName ," is not supported."))
}

function ImportModule([String] $componentName)
{
    # Loads utils and security components modules, each exposes the following functions:
    # 1. ResetToDefault(settingName, value, [ref] errorCode, [ref] errorString)
    # 2. Set(settingName, value, [ref] errorCode, [ref] errorString)
    # 3. VerifyAssignment(settingName, value, [ref] enforceSucceeded, [ref] errorCode, [ref] errorString) -> HResult
    # 4. IsApplicable(settingName) -> Boolean
    # 5. VerifyRemoval(settingName, value, [ref] enforceSucceeded, [ref] errorCode, [ref] errorString) -> HResult
    # 6. IsSupportedParam([string] $name)  -> Boolean
    # 7. IsMergeSupported([string] $name) -> Boolean
    # 7. Setup() -> Void
    # 8. CompletionActions() -> Void

    $startTimeMilli = GetCurrentTimeInMillis

    $moduleName = $componentName + ".psm1"
    $modulePath = Join-Path $global:WorkingFolder -ChildPath $moduleName
    try
    {
        $componentModule = Import-Module -Name $modulePath -Force -AsCustomObject -PassThru -ErrorAction Stop  #-Verbose 
        if($componentName -in $script:SupportedComponents.Keys)
        {
            $componentModule.Setup()      
        }
    } 
    catch
    {
        $errorMessage = $_.Exception.Message
        $errorCode = $_.Exception.HResult
        $msg = -join("Import/Setup error on ", $moduleName, ", HResult: ", $errorCode, ".")
        if((GetRing $componentName) -lt 1)
        {
            $msg += "This warning is only relevant for private preview customers."
        }
        $msg += " Error message:",$errorMessage
        LogWarning $msg
        $componentModule = $null
    }

    #todo - check that all of the expected function were imported 
    $endTimeMilli = GetCurrentTimeInMillis
    LogInfo (-join ("PE: Import for ", $componentName ," took ", ($endTimeMilli - $startTimeMilli)," ms."))
    return $componentModule
}

function SafeInvokeModuleFunction([string] $moduleName,[string] $functionName, [object[]] $functionParameters)
{
    if($null -eq $script:UtilsModules[$moduleName].Module)
    {
        LogWarning (-join("PE: SafeInvokeModuleFunction: Module ", $moduleName, " was not loaded."))
    }
    else {
        try {
            $module = $script:UtilsModules[$moduleName].Module
            $module.$functionName.Invoke($functionParameters)
        }
        catch {
            LogWarning (-join("PE: SafeInvokeModuleFunction: Exception calling function:", $functionName, " from module:", $moduleName, ". HResult:" ,$_.Exception.HResult," error message:" ,$_.Exception.Message))
        }
    }
}

function GetProcessByID([Uint32] $id)
{
    return (Get-CimInstance win32_process -Filter "processid='$id'")
}

function GetParentProcess([CimInstance] $process)
{
    try{
        return (GetProcessByID $process.ParentProcessId)
    }
    catch
    {
        $errorMessage = $_.Exception.Message
	    LogWarning (-join("GetParentProcess: Cannot get parent process. Error message: ", ,$errorMessage))
        exitWithCode -exitCode ([HResult]::E_UNEXPECTED)
    }
}

function GetCurrentProcessParent()
{
    try{
        $currentProcess = GetProcessByID $PID
    }
    catch
    {
        $errorMessage = $_.Exception.Message
	    LogWarning (-join("GetCurrentProcessParent: Cannot get current process. Error message: " ,$errorMessage))
        exitWithCode -exitCode ([HResult]::E_UNEXPECTED)
    }
    return GetParentProcess($currentProcess)
}

function LoadProcessUtilsCode
{           
$queryProcessInfoSource = @"

	// has the following layout that is not supported by csharp
    // struct {
    //    UCHAR Type : 3;
    //    UCHAR Audit : 1;
    //    UCHAR Signer : 4;
    // } DUMMYSTRUCTNAME;
    

    
    [StructLayout(LayoutKind.Sequential)]
    public struct ProcessInfo
    {
        internal byte rawData;
    }

    [DllImport("ntdll.dll")]
    private static extern int NtQueryInformationProcess(
    IntPtr processHandle, 
    int processInformationClass, 
    ref ProcessInfo processInformation, 
    int processInformationLength, 
    out int returnLength);

    public static bool IsProcessProtected(int processId)
    {
        IntPtr hprocess = OpenProcess(ProcessAccessFlags.QueryLimitedInformation, false, processId);
        if (hprocess != IntPtr.Zero)
        {
            try 
            {
                ProcessInfo processInfo = new ProcessInfo();
                int returnLength;
				// 61 to query ProcessProtectionInformation
                int status = NtQueryInformationProcess(hprocess, 61, ref processInfo, Marshal.SizeOf(processInfo), out returnLength);
                if (status != 0)
                {
					Console.WriteLine("NT query failed: " + status);
					return false;
                }
                var type = (processInfo.rawData >> 0) & 7;  // bits 0-2 are storing the type
                var signer = (processInfo.rawData >> 4) & 7; // bits 4-7 are storing the signer 
                return  (type == 1) && 
                        (signer == 5); // type is process protected light and signer is windows
            }             
            finally
            {
                CloseHandle(hprocess);
            }
        }
        Console.WriteLine("Unable to open process");
		return false;
    }

	[DllImport("kernel32.dll")]
    private static extern bool QueryFullProcessImageName(IntPtr hprocess, int dwFlags, System.Text.StringBuilder lpExeName, out int size);

    [DllImport("kernel32.dll")]
    private static extern IntPtr OpenProcess(ProcessAccessFlags dwDesiredAccess, bool bInheritHandle, int dwProcessId);

    [DllImport("kernel32.dll", SetLastError = true)]
    private static extern bool CloseHandle(IntPtr hHandle);

    [Flags]
    public enum ProcessAccessFlags : uint
    {
        All = 0x001F0FFF,
        Terminate = 0x00000001,
        CreateThread = 0x00000002,
        VirtualMemoryOperation = 0x00000008,
        VirtualMemoryRead = 0x00000010,
        VirtualMemoryWrite = 0x00000020,
        DuplicateHandle = 0x00000040,
        CreateProcess = 0x000000080,
        SetQuota = 0x00000100,
        SetInformation = 0x00000200,
        QueryInformation = 0x00000400,
        QueryLimitedInformation = 0x00001000,
        Synchronize = 0x00100000
    }  

    public static string GetExecutablePath(int processId)
    {
        var buffer = new System.Text.StringBuilder(1024);
        IntPtr hprocess = OpenProcess(ProcessAccessFlags.QueryLimitedInformation,
                                    false, processId);
        if (hprocess != IntPtr.Zero)
        {
            try
            {
                int size = buffer.Capacity;
                if (QueryFullProcessImageName(hprocess, 0, buffer, out size))
                {
                    return buffer.ToString().Replace(@"\", @"\\");
                }
            }
            finally
            {
                CloseHandle(hprocess);
            }
        }
		Console.WriteLine("Unable to get executable path: " + Marshal.GetLastWin32Error());
		return "";
    }
"@
    Add-Type -MemberDefinition $queryProcessInfoSource -Name "Utils" -Namespace "QueryProcessInfo" 
}

function ValidateField([string] $expected, [string] $actual, [string] $fieldName, [string] $errorMessage)
{
    if (!($expected -eq $actual))
    {
        $parnetValidationErrorMsg = -join($errorMessage," field:",$fieldName," expected:",$expected," actual:",$actual," - Exiting script.")
        LogWarning  $parnetValidationErrorMsg 
        exitWithCode -exitCode ([HResult]::E_HANDLE)
    }
    else 
    {
        LogDebug (-join("field: ",$fieldName," is as expected: ",$expected))
    }
}

#NOTE: This function calls ValidateField which exits the script in case of failure.
function ValidateParentProcess()
{
    $startTimeMilli = GetCurrentTimeInMillis
    LoadProcessUtilsCode 
    $parent = GetCurrentProcessParent
    $binaryFolder = (Join-Path (Get-CimInstance Win32_OperatingSystem -ErrorAction Stop).SystemDrive "\\Program Files\\Windows Defender Advanced Threat Protection\\" ) #parent path from original installation
    $currentRegItem = $null
    try
    {
        # note: we do not use Get-ItemPropertyValue directly to get value because if InstallLocation does not exist
        # it with throw, and we want to know if the failure is because the value doesn't exist or an unexpected crash
        $currentRegItem = Get-ItemProperty $global:InstallationPathRegPath -ErrorAction SilentlyContinue
        if ($null -ne $currentRegItem) #install location reg key found (update was performed) - follow path to find valid parent
        {
            #get entry value. Example: C:\ProgramData\Microsoft\Windows Defender Advanced Threat Protection\Platform\10.8044.22439.1001\
            $newInstallationPath = $currentRegItem.$global:InstallationPathRegKey
            if($null -ne $newInstallationPath) #InstallLocation exists
            {
                $binaryFolder = $newInstallationPath.Replace("\","\\")
            }
        }
    } catch {
        $parnetValidationErrorMsg = -join("ValidateParentProcess: Cannot get installation path for parent process validation err msg:",$_.Exception.Message)
        LogWarning $parnetValidationErrorMsg 
        exitWithCode -exitCode ([HResult]::E_UNEXPECTED)
    }

    #validate parent process:
    $expectedParentImage = (Join-Path $binaryFolder "SenseCM.exe") 
    $actualParentImage = [QueryProcessInfo.Utils]::GetExecutablePath($parent.ProcessId)
    ValidateField $expectedParentImage $actualParentImage "parent image file name" -errorMessage "Parent validation failed."

    #validate parent's parent process:
    LogDebug "ValidateParentProcess: attempting to get parent of parent..."
    $parentOfParent = GetParentProcess($parent)  
    $actualParentOfParentImage = [QueryProcessInfo.Utils]::GetExecutablePath($parentOfParent.ProcessId) 
    $expectedParentOfParentImage = (Join-Path $binaryFolder "MsSense.exe")
    ValidateField $expectedParentOfParentImage $actualParentOfParentImage "parent of parent image file name" -errorMessage "Parent of parent validation failed."

    #validate parent of parent is PPL:
    $isProtected = [QueryProcessInfo.Utils]::IsProcessProtected($parentOfParent.ProcessId)
    ValidateField $true $isProtected "parent of parent protection level" -errorMessage "Parent of parent validation failed."

    $endTimeMilli = GetCurrentTimeInMillis
    LogInfo (-join ("PE: ValidateParentProcess took ", ($endTimeMilli - $startTimeMilli)," ms."))
}

function GetDefenderVersion()
{
    $defenderVersionItem = Get-ItemProperty $script:DefenderEngineVersionPath -ErrorAction SilentlyContinue
    return ($local:defenderVersionItem.EngineVersion)
}

# finally - this is where the magic happens 
LogInfo "PolicyEnforcer starting..."
# function ValidateParentProcess should be called first before any logic is happening
try
{
    ValidateParentProcess
} catch {
    LogWarning (-join("Exception on ValidateParentProcess HResult:",$_.Exception.HResult,". Error message:",$_.Exception.Message))
    exitWithCode -exitCode ($_.Exception.HResult)
}

LogInfo (-join("Defender Version:[",(GetDefenderVersion),"]"))

Add-Type -TypeDefinition $source
New-EventLog -LogName $EventLogName -Source $EventLogSource -ErrorAction SilentlyContinue
$global:report = New-Object Report

foreach ($componentName in $script:UtilsModules.Keys) 
{
    $script:UtilsModules[$componentName].Module = ImportModule($componentName)
}

foreach ($componentName in $script:SupportedComponents.Keys) 
{
    try 
    {
        $componentModule = ImportModule($componentName)

        if($null -eq $componentModule)
        {
            LogInfo ("PE: Skipping $componentName")
            continue
        }

        # creating an effective policy along with conflicts detections. 
        # The conflicts are added to $global:report
        $effectivePolicySettings = EffectivePolicy -componentName $componentName -componentModule $componentModule
       
        $startTotalComponentApplicationTimeMilli = 0 #accumulate the time spent on this component Set/Remove
        foreach ($setting in $effectivePolicySettings) 
        {
            try
            {
                $isApplicable = $componentModule."IsApplicable"($setting.name, $setting.value)
            }
            catch
            {
                LogWarning (-join("Exception on IsApplicable for ",$setting.name," - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
                $isApplicable = $false #don't continue to set/remove if couldn't  validate applicability
            }

            
            if ($isApplicable)
            {
                $actionHRes = [HResult]::S_OK.value__ #error code is as an int (not HResult enum) because when receiving error value from PS it is not the enum.
                $resErrorStr = ""
                $startTimeMilli = GetCurrentTimeInMillis

                if ($setting.intent -eq [IntentType]::Remove.ToString())
                {
                    #keep reset to default first. So that we first remove whatever is needed and then apply the new values (so we won't delete values we need to leave applied).
                    try
                    {
                        $componentModule."ResetToDefault"($setting.name, $setting.value, [ref] $actionHRes, [ref] $resErrorStr)
                    }
                    catch
                    {
                        LogWarning (-join("Exception on ResetToDefault for ",$setting.name," - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
                    }
                }
                else ## (intent -eq "Apply" -or intent -eq "Report")
                {
                    try
                    {
                        $componentModule."Set"($setting.name, $setting.value, [ref] $actionHRes, [ref] $resErrorStr)                        
                    }
                    catch
                    {
                        LogWarning (-join("Exception on Set for ",$setting.name," - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
                    }
                }

                $endTimeMilli = GetCurrentTimeInMillis
                $startTotalComponentApplicationTimeMilli += ($endTimeMilli - $startTimeMilli)

                if (! (Succeeded $actionHRes)) #if we failed at this point, we have probably failed entirely for this setting - save the HResult
                {
                    foreach ($policyIdentifier in $setting.policies) 
                    {
                        SafeInvokeModuleFunction "ErrorHandlingUtils" "HandleMpPreferenceError" @(([ref] $actionHRes), $resErrorStr)
                        $global:report.ReportSettingResult($policyIdentifier.id, $policyIdentifier.intent, $policyIdentifier.checksum, $setting.name, [SettingEnforceResults]::Fail, $true, $actionHRes, $resErrorStr) 
                    }
                }

            }       
        }
        LogInfo (-join ("PE: applications for ", $componentName ," took ", $startTotalComponentApplicationTimeMilli ," ms."))

        $startTimeMilli = GetCurrentTimeInMillis
        try
        {
            $componentModule."CompletionActions"()
        }
        catch
        {
            LogWarning (-join("Exception on CompletionActions for ",$componentName," - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
        }
        $endTimeMilli = GetCurrentTimeInMillis
        LogInfo (-join ("PE: completion for ", $componentName ," took ", ($endTimeMilli - $startTimeMilli)," ms."))

        
        $startTotalComponentAssertionTimeMilli = 0 #accumulate the time spent on this component VerifyAssignment/VerifyRemoval
        foreach ($setting in $effectivePolicySettings) 
        {
            $enforceSucceeded = $true
            $validationHRes = [HResult]::S_OK #validation res is HResult enum because the functions in the scripts return the value (our code), no PS received values
            $errorCode = [HResult]::S_OK.value__ #error code is as an int (not HResult enum) because when receiving error value from PS it is not the enum.
            $resErrorStr = ""
            try
            {
                $isApplicable = $componentModule."IsApplicable"($setting.name, $setting.value)
            }
            catch
            {
                LogMessage ([verbosityLevel]::verbosity_warn) (-join("Exception on IsApplicable for ",$setting.name," - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
                $isApplicable = $false #don't continue to set/remove if couldn't  validate applicability
            }

            if ($isApplicable)
            {
                $startTimeMilli = GetCurrentTimeInMillis
                if ($setting.intent -eq [IntentType]::Remove.ToString())
                {
                    try
                    {
                        $validationHRes = $componentModule."VerifyRemoval"($setting.name, $setting.value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $resErrorStr)
                    }
                    catch
                    {
                        LogWarning (-join("Exception on VerifyRemoval for ",$setting.name," - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
                        $validationHRes = [HResult]::E_Fail
                    }
                }
                else ## (intent -eq "Apply" -or intent -eq "Report")
                {
                    try
                    {
                        $validationHRes = $componentModule."VerifyAssignment"($setting.name, $setting.value, [ref] $enforceSucceeded, [ref] $errorCode, [ref] $resErrorStr)
                    }
                    catch
                    {
                        LogWarning (-join("Exception on VerifyAssignment for ",$setting.name," - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
                        $validationHRes = [HResult]::E_Fail
                    }
                }
                $endTimeMilli = GetCurrentTimeInMillis
                $startTotalComponentAssertionTimeMilli += ($endTimeMilli - $startTimeMilli)
            }

            # a setting can be a part of a few policies 
            foreach ($policyIdentifier in $setting.policies) 
            {
                if ($setting.intent -eq $policyIdentifier.intent)
                {
                    # setting intent can be different only if the setting intent is remove but the policy intent is not 
                    # in this case we do not report this setting
                    if (-not $isApplicable)
                    {
                        $global:report.ReportSettingResult($policyIdentifier.id, $policyIdentifier.intent, $policyIdentifier.checksum, $setting.name, [SettingEnforceResults]::NotApplicable)
                    } 
                    else
                    {
                        if (Succeeded($validationHRes))
                        {
                            if ($enforceSucceeded)
                            {
                                #if value matched expectations - that's what should be reported.
                                $global:report.ReportSettingResult($policyIdentifier.id, $policyIdentifier.intent, $policyIdentifier.checksum, $setting.name, [SettingEnforceResults]::Success, $true, [HResult]::S_OK.value__ , $null)
                            }
                            else
                            {
                                if($errorCode -eq [HResult]::E_INTERNAL_CONFLICT) #conflict detected by module (possible only for mergeable settings where we cannot detect conflict in effectivePolicy().
                                {
                                    $global:report.ReportSettingResult($policyIdentifier.id, $policyIdentifier.intent, $policyIdentifier.checksum, $setting.name, [SettingEnforceResults]::Conflict)
                                }
                                else
                                {
                                    SafeInvokeModuleFunction "ErrorHandlingUtils" "HandleMpPreferenceError" @(([ref] $errorCode), $resErrorStr)
                                    #"plain old" failure - if earlier failure documented, don't override
                                    $global:report.ReportSettingResult($policyIdentifier.id, $policyIdentifier.intent, $policyIdentifier.checksum, $setting.name, [SettingEnforceResults]::Fail, $false, $errorCode, $resErrorStr) 
                                }
                            }                      
                        }
                        else 
                        {
                            if($resErrorStr -eq "")
                            {
                                $validationErrString = -join($validationStrErrHeader," ",$setting.name)
                            } else {
                                $validationErrString = $resErrorStr
                            }
                            #could not validate - report this over previous fail
                            SafeInvokeModuleFunction "ErrorHandlingUtils" "HandleMpPreferenceError" @(([ref] $errorCode), $validationErrString)
                            $global:report.ReportSettingResult($policyIdentifier.id, $policyIdentifier.intent, $policyIdentifier.checksum, $setting.name, [SettingEnforceResults]::ValidationFailure, $true, $errorCode, $validationErrString)
                        }   
                    }
                }
            }
        }
        LogInfo (-join ("PE: assertion for ", $componentName ," took ", $startTotalComponentAssertionTimeMilli ," ms."))

    } catch 
    {
        LogWarning (-join("Exception on enforcing effective policy of component:",$componentName," - HResult:",$_.Exception.HResult," error message:",$_.Exception.Message))
    }

}

# now mark setting that didn't belong to any component
$startTimeMilli = GetCurrentTimeInMillis
foreach ($policyNode in $global:policiesObj.policies)
{
    $policy = $policyNode.payload
    foreach ($setting in $policy.settings) 
    {
        if (!$global:AllSupportedSettings.Contains($setting.name))
        {
            $global:report.ReportSettingResult($policy.policyId, $policy.intentType, $policy.policyChecksum, $setting.name, [SettingEnforceResults]::NotRecognized)
        }
    }
}
$endTimeMilli = GetCurrentTimeInMillis
LogInfo (-join ("PE: Reporting NotRecognized settings took ", ($endTimeMilli - $startTimeMilli)," ms."))

$startTimeMilli = GetCurrentTimeInMillis
# after settings are set, calc checksum on settings values
foreach ($policyReport in $global:report.policyInfo)
{
    $settingsString = Out-String -InputObject $policyReport.settings
    $policyReport.reportChecksum = CalcChecksum $settingsString
}
$endTimeMilli = GetCurrentTimeInMillis
LogInfo (-join ("PE: checksum calculations took ", ($endTimeMilli - $startTimeMilli)," ms."))


$startTimeMilli = GetCurrentTimeInMillis

$reportJson = $global:report | ConvertTo-Json -Depth 4
$timestamp =  Get-Date  -UFormat "%Y%m%d%R%s" | ForEach-Object { $_ -replace ":", "."}
$reportFilePath =  Join-Path -Path $global:WorkingFolder -ChildPath "report_${timestamp}.json"
$reportJson | Out-File -FilePath $reportFilePath -Encoding ascii

$endTimeMilli = GetCurrentTimeInMillis
LogInfo (-join ("PE: Generating report file took ", ($endTimeMilli - $startTimeMilli)," ms."))

#override max log size - we need to know PE finished gracefully even if logging was cut due to size.
Write-Host (-join ("PE: Done in ",((GetCurrentTimeInMillis) - $scriptTotalStartTimeMilli)," ms. Goodbye."))

} catch { # general try/catch for policyEnforcer
    Write-Host (-join("Exception on PolicyEnforcer. HResult: ", $_.Exception.HResult, ". Error message: ", $_.Exception.Message))
    $Host.SetShouldExit($_.Exception.HResult)
    exit
}
# SIG # Begin signature block
# MIIntwYJKoZIhvcNAQcCoIInqDCCJ6QCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCACGavLz4bCB0aS
# 9WvgYGciypDH30Yhl97mTNp6JBI8vaCCDZcwggYVMIID/aADAgECAhMzAAADcaWg
# nFyRHuruAAAAAANxMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjMwNTExMTg0MTEzWhcNMjQwNTA4MTg0MTEzWjCBlDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjE+MDwGA1UEAxM1TWlj
# cm9zb2Z0IFdpbmRvd3MgRGVmZW5kZXIgQWR2YW5jZWQgVGhyZWF0IFByb3RlY3Rp
# b24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDQl+fOHFGkNfDHrcv3
# g8shIsUp58WmSbtNm4uu7Sw75Wgsgo5p2yYGGqiRWbu0eIKhAJ1bRp9Qa7vAjMEw
# d0wf9wlLjDNAITzDsokwoCl0rOgWxzYuGSusZq4e/liClIpY5fuxrpsEnsssR1o6
# jITkdJQ6ZkWsTFETxLLOEGxe9BkTkXZaBOHlhTt5QIq6UB6j8zNN/Vo4rFmq751X
# lQLA1jOnNi5NOS3/11TW16ZCo9XKG22zdNsJCUckT/QH2eeQNgpuyfIvP06o+RpP
# Uo7P4xdc4c74S3HKfRiNuqbs3+xvb22GtXIz3cKFDhILpNKN/SC8qLjizq0PxwL1
# lr1ZAgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBggrBgEFBQcDAwYKKwYBBAGCN0wv
# ATAdBgNVHQ4EFgQUcR5wDY+Mh5A/I7scm54A8ODeSVcwRQYDVR0RBD4wPKQ6MDgx
# HjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEWMBQGA1UEBRMNNDUxODk0
# KzUwMTAwNTAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8E
# TTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9N
# aWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBR
# BggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0
# cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAw
# DQYJKoZIhvcNAQELBQADggIBAKYhzPfo4pWE2MYg2W1VnHJ7NZ/nwEpGYEz7Qhqc
# bPj2sJFFhnUzYHOAyXz4vP0+ERhFw9lrqLvoyj8s6xQU11QPQVItopahH65USGWh
# UcyrXFBVSx7MG3Dwi+NT6rxq9MjY3KLB1hZPF7hgtOjWgtjXgeJOsy3IjPiRkDwX
# 5b5Pg6o8Xpyt84mpvAPsDKO7yJGJq8V528deIYh+0DcELvUZTtl4Kz8ElT4BxF6i
# XxXkuqpCWO/rwEa6HrLNV2udND/AImVio09bbZXOWBcb9uTTNezYX5fZvinIX7hl
# UHHIDaVUqqwTFtac8q7BLE02T0P7NAeBtayujIQFzGNuwLN302UmUWAyc4LynqZa
# h2OmZWcMODUCuHIMP3pk54XjWt0YYEWPFAbA7T7j8wjby/vPXdG93LybID/yjGDb
# U1t5xMoEpN5rvSUM1U42iJ1tF268/nE2/8nKqUpdJXH/MaXxYU3FnbshmIe1xfxF
# NjeMYN4jg+5gjz5SBzHAhT1X9QWXNOhAXxib3wcchlsxFYtSNihHLEJ2GuckRuPi
# 47VWi/3hTTnl88S8r3jsbWLatGbQxcbrsF4sglAvFD9SQSXLGUsE5wz/1p/Gb1Fh
# qfNBc6E4AWVj9+A1KezSnMBAw8gJk3fiRkdxfSfpLOkHvgA7qsOmQ7qwjs05kWZf
# 055CMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDEL
# MAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1v
# bmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWlj
# cm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4
# MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3Y
# bqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUB
# FDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnbo
# MlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT
# +OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuy
# e4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEh
# NSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2
# z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3
# s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78Ic
# V9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E
# 11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5P
# M4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcV
# AQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAf
# BgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBL
# hklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggr
# BgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNS
# b29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsG
# AQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwA
# ZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0G
# CSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDB
# ZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc
# 8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYq
# wooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu
# 5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWI
# UUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXh
# j38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yH
# PgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtI
# EJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA9Z74v2u3S5fi63V4Guzq
# N5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgR
# MiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQ
# zTGCGXYwghlyAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5n
# dG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9y
# YXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTEC
# EzMAAANxpaCcXJEe6u4AAAAAA3EwDQYJYIZIAWUDBAIBBQCgga4wGQYJKoZIhvcN
# AQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUw
# LwYJKoZIhvcNAQkEMSIEIDoulyV0idiryAqrRxdXiyiksUnTRJCWgoMlis9X/fRz
# MEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8AcwBvAGYAdKEagBhodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEAVWep9nScGW21
# jIu3tzIwPufYtCttoVXbW7cOv3vrVNzcJnaq1ZD+WBYj8C6dK/S6oQzChMixkYmR
# Qi21TjoN4NQGO+n8DnS1h8SG1231Msh7GECzZsGvCPKFYSbQIuKMD1a+VhNQFec+
# WsD6DBW9gxx0iaRYBetwz04VfNFv5Ghlo+vzOg2//kjJJWx82dhYbHG2KuQB2Czw
# 0iKeZ3FlA3ly4fVQobMqd78R/Y+rmm1mzTVwCa6mCEolMEO/p1CHL+Pmo/HWr1o9
# WS9grmDmC9/rOb7gWp6gThf+D7jMSeUpe3BbuJCXdbWCf31dbuFXSmyJOx2tsdkW
# iaZIN3tED6GCFwAwghb8BgorBgEEAYI3AwMBMYIW7DCCFugGCSqGSIb3DQEHAqCC
# FtkwghbVAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFRBgsqhkiG9w0BCRABBKCCAUAE
# ggE8MIIBOAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCCTUk+v122m
# sZwDjIVB5fOA2RzZQ1a/7ihbUO30ao1mwwIGZIsBgWGeGBMyMDIzMDcxMTE3NDcw
# Mi44MjdaMASAAgH0oIHQpIHNMIHKMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2Fz
# aGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENv
# cnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1lcmljYSBPcGVyYXRpb25z
# MSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjozQkJELUUzMzgtRTlBMTElMCMGA1UE
# AxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaCCEVcwggcMMIIE9KADAgEC
# AhMzAAABxjDNLtbTocD0AAEAAAHGMA0GCSqGSIb3DQEBCwUAMHwxCzAJBgNVBAYT
# AlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYD
# VQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBU
# aW1lLVN0YW1wIFBDQSAyMDEwMB4XDTIyMTEwNDE5MDEzNFoXDTI0MDIwMjE5MDEz
# NFowgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
# EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJTAjBgNV
# BAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJhdGlvbnMxJjAkBgNVBAsTHVRoYWxl
# cyBUU1MgRVNOOjNCQkQtRTMzOC1FOUExMSUwIwYDVQQDExxNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBTZXJ2aWNlMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA
# 770iOr6v4Hk1m3SZj+1BR/iuofv+f6eVb7Hc21YxzAzro4G6kKXF47YAsEgrWWT1
# ogvp0IroFm8CdRZTf/DlQ0fbNNO9pCA01KJ03zH82Clmk9ze9r1jPJ1ZJaMnsZmA
# y7VpY9mNqX9dhPvnW1/ZxbbiHv7qwwgw9U2ST5mfcpPutsI/Qr/gLC6aTI3UCYzi
# VPZ/Qfag8NQhKkpHZO3Kr5r83cy7jz4OWPy5M2WitWv5bJJ5rBTW518QPEzFwzq8
# e8P722CWKZJFjN8etBgsK05gHeHaN9kmlpYJJL84v9JiaX7NFJkORhApEFZiUIaZ
# oLxJt4pcBDzf+WD9UAjRKCrAseJ/ckzQvOn95X4Ot4asnOuNhcCdcQWcrZoykFmE
# aoYkrsD7n/4nFFHwJDKUaBYZZLwPj7ux48S1Ye+cMccMxdRSjuoG4rqJqpEd6gzf
# z239v36L+LtOlQhfL5cnclhNSWdmKw1THyekH96RNtOpIE7c+9Tnsv1aE9hphejL
# OJxfsXTkyw0FIdilc0CP7zzUsqaCGF2mFXRwL4jfX1RyV7QGKEeOpvGZqQKLHVZb
# LD32ztW8Lfv99xQ24d/TIxO9LReuHsnwotn+8BsCrzu+/24vOcA9Xcpd7kIeW4zm
# UxhFsv4KCfiqMptZhhMAA0SqDuj23cj10smXROLUnhUCAwEAAaOCATYwggEyMB0G
# A1UdDgQWBBRxX/lHiShECp1n2lMa6G1uLvNglDAfBgNVHSMEGDAWgBSfpxVdAF5i
# XYP05dJlpxtTNRnpcjBfBgNVHR8EWDBWMFSgUqBQhk5odHRwOi8vd3d3Lm1pY3Jv
# c29mdC5jb20vcGtpb3BzL2NybC9NaWNyb3NvZnQlMjBUaW1lLVN0YW1wJTIwUENB
# JTIwMjAxMCgxKS5jcmwwbAYIKwYBBQUHAQEEYDBeMFwGCCsGAQUFBzAChlBodHRw
# Oi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NlcnRzL01pY3Jvc29mdCUyMFRp
# bWUtU3RhbXAlMjBQQ0ElMjAyMDEwKDEpLmNydDAMBgNVHRMBAf8EAjAAMBMGA1Ud
# JQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBCwUAA4ICAQALzF724jXugDU486PM
# BYR7Cc6aHr2nJDnGsnkqwZYmjRi28qv6S1Ly772zwm5DI189zgAQ99xhEjW6uYkr
# X5UrtVu7JUQP6bBBBJ98c28FAIPIK3fkQNCv2rxRjfQMJSdcwsJeTK7Ld09OuA5h
# Y4PWCBgJpfY71LcaXz3FR8ANPFX6zcKYbgYOZregtpDub34N0QhR7wc/FcmV+g4I
# 3IdTAoMD2/WI5ZsfKTzBUn/U3ApUhNwbOl5YSC+f9S1LStbZLwPzMS+fLPXJUSe7
# SSvspfSsr/VEe0oQhmaR+5vcq+7MLw861WBVhYpJ7TB5YBS5ORO9XdIbcpbBFwcH
# Pmb8iZqSIqW9JpgG76+5NQULPVzZ75z5W2R5ZiyQktiHpMwjX2OO29Z8+nTw2tOs
# VCcwzH9LoELedv3PjcpbwOyLjtm1T4XHYd3qbd9DXoBjNYkSjdi37pNp58u+rITl
# tLKOjjQCJwj1FpnuBY825B5C0uC/NYESEKsTicEjhS/4ujBXLcNGDhVBl2vHE6qY
# /YW4ky1vcypvUrsG81gpv2+8/ihOwg4wTLO7XqikeIiU3ZWAUAoOpTl14tedQqxb
# HTDveJYR3OU0yKB2xwf87EWCAb0CJimhDmyQaKEvSV0fLW9iVyI0wYcG4V2aVN6T
# rZ4mr+ffaqDQD9F+HpPhP0plAzCCB3EwggVZoAMCAQICEzMAAAAVxedrngKbSZkA
# AAAAABUwDQYJKoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRl
# IEF1dGhvcml0eSAyMDEwMB4XDTIxMDkzMDE4MjIyNVoXDTMwMDkzMDE4MzIyNVow
# fDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMd
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwggIiMA0GCSqGSIb3DQEBAQUA
# A4ICDwAwggIKAoICAQDk4aZM57RyIQt5osvXJHm9DtWC0/3unAcH0qlsTnXIyjVX
# 9gF/bErg4r25PhdgM/9cT8dm95VTcVrifkpa/rg2Z4VGIwy1jRPPdzLAEBjoYH1q
# UoNEt6aORmsHFPPFdvWGUNzBRMhxXFExN6AKOG6N7dcP2CZTfDlhAnrEqv1yaa8d
# q6z2Nr41JmTamDu6GnszrYBbfowQHJ1S/rboYiXcag/PXfT+jlPP1uyFVk3v3byN
# pOORj7I5LFGc6XBpDco2LXCOMcg1KL3jtIckw+DJj361VI/c+gVVmG1oO5pGve2k
# rnopN6zL64NF50ZuyjLVwIYwXE8s4mKyzbnijYjklqwBSru+cakXW2dg3viSkR4d
# Pf0gz3N9QZpGdc3EXzTdEonW/aUgfX782Z5F37ZyL9t9X4C626p+Nuw2TPYrbqgS
# Uei/BQOj0XOmTTd0lBw0gg/wEPK3Rxjtp+iZfD9M269ewvPV2HM9Q07BMzlMjgK8
# QmguEOqEUUbi0b1qGFphAXPKZ6Je1yh2AuIzGHLXpyDwwvoSCtdjbwzJNmSLW6Cm
# gyFdXzB0kZSU2LlQ+QuJYfM2BjUYhEfb3BvR/bLUHMVr9lxSUV0S2yW6r1AFemzF
# ER1y7435UsSFF5PAPBXbGjfHCBUYP3irRbb1Hode2o+eFnJpxq57t7c+auIurQID
# AQABo4IB3TCCAdkwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEEAYI3FQIEFgQU
# KqdS/mTEmr6CkTxGNSnPEP8vBO4wHQYDVR0OBBYEFJ+nFV0AXmJdg/Tl0mWnG1M1
# GelyMFwGA1UdIARVMFMwUQYMKwYBBAGCN0yDfQEBMEEwPwYIKwYBBQUHAgEWM2h0
# dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvRG9jcy9SZXBvc2l0b3J5Lmh0
# bTATBgNVHSUEDDAKBggrBgEFBQcDCDAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMA
# QTALBgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBTV9lbL
# j+iiXGJo0T2UkFvXzpoYxDBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1p
# Y3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXRfMjAxMC0w
# Ni0yMy5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIz
# LmNydDANBgkqhkiG9w0BAQsFAAOCAgEAnVV9/Cqt4SwfZwExJFvhnnJL/Klv6lwU
# tj5OR2R4sQaTlz0xM7U518JxNj/aZGx80HU5bbsPMeTCj/ts0aGUGCLu6WZnOlNN
# 3Zi6th542DYunKmCVgADsAW+iehp4LoJ7nvfam++Kctu2D9IdQHZGN5tggz1bSNU
# 5HhTdSRXud2f8449xvNo32X2pFaq95W2KFUn0CS9QKC/GbYSEhFdPSfgQJY4rPf5
# KYnDvBewVIVCs/wMnosZiefwC2qBwoEZQhlSdYo2wh3DYXMuLGt7bj8sCXgU6ZGy
# qVvfSaN0DLzskYDSPeZKPmY7T7uG+jIa2Zb0j/aRAfbOxnT99kxybxCrdTDFNLB6
# 2FD+CljdQDzHVG2dY3RILLFORy3BFARxv2T5JL5zbcqOCb2zAVdJVGTZc9d/HltE
# AY5aGZFrDZ+kKNxnGSgkujhLmm77IVRrakURR6nxt67I6IleT53S0Ex2tVdUCbFp
# AUR+fKFhbHP+CrvsQWY9af3LwUFJfn6Tvsv4O+S3Fb+0zj6lMVGEvL8CwYKiexcd
# FYmNcP7ntdAoGokLjzbaukz5m/8K6TT4JDVnK+ANuOaMmdbhIurwJ0I9JZTmdHRb
# atGePu1+oDEzfbzL6Xu/OHBE0ZDxyKs6ijoIYn/ZcGNTTY3ugm2lBRDBcQZqELQd
# VTNYs6FwZvKhggLOMIICNwIBATCB+KGB0KSBzTCByjELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjElMCMGA1UECxMcTWljcm9zb2Z0IEFtZXJpY2Eg
# T3BlcmF0aW9uczEmMCQGA1UECxMdVGhhbGVzIFRTUyBFU046M0JCRC1FMzM4LUU5
# QTExJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2WiIwoBATAH
# BgUrDgMCGgMVAC01yuYmIVvsokSacJmWe8Mu2QFtoIGDMIGApH4wfDELMAkGA1UE
# BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAc
# BgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0
# IFRpbWUtU3RhbXAgUENBIDIwMTAwDQYJKoZIhvcNAQEFBQACBQDoV8YmMCIYDzIw
# MjMwNzExMjAxNDMwWhgPMjAyMzA3MTIyMDE0MzBaMHcwPQYKKwYBBAGEWQoEATEv
# MC0wCgIFAOhXxiYCAQAwCgIBAAICEtkCAf8wBwIBAAICEf0wCgIFAOhZF6YCAQAw
# NgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGEWQoDAqAKMAgCAQACAwehIKEKMAgC
# AQACAwGGoDANBgkqhkiG9w0BAQUFAAOBgQCF8Bcr922tUwLLCU8Q2sxARqvtZZvI
# VwgwAD8nWHLUslU9ks6kLdZBf1/UjCr9CrqPvPgNXRGybE4CaxDihQ+oa+YvhihV
# kj4pgGuzDQIlKNJDHotNDUc+w6AtAV3wKUf4fg16nv0U1uuKPUdjs1SfnXQfMVv9
# qJMXYXcnB4AG+zGCBA0wggQJAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQI
# EwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3Nv
# ZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBD
# QSAyMDEwAhMzAAABxjDNLtbTocD0AAEAAAHGMA0GCWCGSAFlAwQCAQUAoIIBSjAa
# BgkqhkiG9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIEIN5iZRbP
# 2XV2CCLhRCgJ45eVEaTp4e2hWgVq9w81rVa0MIH6BgsqhkiG9w0BCRACLzGB6jCB
# 5zCB5DCBvQQgVjETXFXJvZgpxiZqk/BwDWA5Pxw6hehULpkFr9rjry0wgZgwgYCk
# fjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
# UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQD
# Ex1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAcYwzS7W06HA9AAB
# AAABxjAiBCCDj13dpjn27U8pxEo6BuytLiGf+kjKusmJ47kLThPVxjANBgkqhkiG
# 9w0BAQsFAASCAgC96ZsviSP+cM+Ly918ozgggaUxXVjaWagjAX5RZeY7Bx/O0xlj
# IglTMqmgvNDOV1/G5mhoZfQ86hJ/SDYLNvdLyfhx+bnxWmkcmv8uuKcedKfAu/S4
# Gdr5bG0RrRQOU8sqWYg86OWtx/CHnMWaEfjcPp/EXTgWoBEKA9MCPCpvqCV2SJBj
# me0hzvzDkteZsjbQvZtfupJuXjFRvstIFFSAhg4wcGNuidM/lssPt1EzfOkbi5lA
# scbOdZB7D8Sbjiz56YSsfg6z1YZgsjwapvw/IPfZl5IXOu3zsajHwpaGDBB3ZAnP
# +/EfpcWhaKnWd/JAUAzORJGpi2ajUI5bnE81jx5djkOa2G069zqgwYrJdBOrgbvx
# b/sAjl2fxdci7aqeGkhRZHcNUAXFqV3ryVNaUC24UNZtanh97J7hvq1cWKPZQ4aC
# JGNalRvCkerEyTMaWladnrqXy5dbFXULTK2o5oVGKwPjgk+W8AZgFPsVlzOmOcdT
# Y/S7tbHoDX8egegrVdSkyY9Vf6OVdTCVD9srlKQnVT93498QHOGvfb4+T2udX7PS
# TDTyPIpJ77X96APGEWdX+AiqZtdVKPCHx7+vJMAj7vbslk56nurXzdhHk2J9YwD+
# iLosQZ7s6fH6Y1XYxZ2/lOcMfCo5NaawpXcPPaXUIH/Eifmwp2dli32e3Q==
# SIG # End signature block
